unit Expression;

//==============================================================================
interface

//==============================================================================
uses
  Windows, Messages, SysUtils, Classes, MiscSupport, ListSupport, TokenSupport,
  ExpressionTokens, ExpressionLexicon;

//==============================================================================
type
  //============================================================================
  TExpression = class(TTokenList)
  private
    // data
    FParser: TObject;
    FResultant: variant;

    // last error info
    FErrorText: string;
    FErrorAt, FErrorCode: integer;
    FErrors: boolean;
  protected
    // data
    procedure ClearResultant; virtual;

    // last error info
    procedure RememberErrors; virtual;
  public
    // factory floor
    constructor Create(aParser: TObject); virtual;
    destructor Destroy; override;

    // engine
    property Parser: TObject read FParser;
    function Execute: boolean; virtual;
    function Evaluate: variant; virtual;
    property Resultant: variant read FResultant;

    // last error info
    procedure ClearErrors;
    property ErrorAt: integer read FErrorAt;
    property ErrorCode: integer read FErrorCode;
    property ErrorText: string read FErrorText;
    property Errors: boolean read FErrors;
  end;

  //============================================================================
  TExpressionList = class(TObjectList)
  protected
    function GetExpression(Index: integer): TExpression;
  public
    property Expressions[Index: integer]: TExpression read GetExpression;
  end;

  //============================================================================
  TIdentExpressionPairs = class(TStringList)
  protected
    function GetIdent(index: integer): string;
    function GetExpression(index: integer): TExpression;
  public
    property Idents[index: integer]: string read GetIdent;
    property Expressions[index: integer]: TExpression read GetExpression;
    function AddIdent(aIdent: string): integer;
    function AddExpression(aExpr: TExpression): integer;
    function AddIdentExpression(aIdent: string; aExpr: TExpression): integer;
  end;

//==============================================================================
implementation

//==============================================================================
uses
  MiscConstants, ResultSupport, ExpressionStack, ExpressionConstants;

//==============================================================================
// TExpression = class(TObject)
//==============================================================================
constructor TExpression.Create(aParser: TObject);
begin
  inherited Create;
  FParser := aParser;
  ClearResultant;
end;
destructor TExpression.Destroy;
begin
  ClearResultant;
  inherited Destroy;
end;
procedure TExpression.ClearResultant;
begin
  VarClear(FResultant);
end;
function TExpression.Execute: boolean;
var
  Index: integer;
  vStack: TExpressionStack;
begin
  // no mess yet
  ClearResult;
  ClearResultant;

  // make the stack
  if Count > 0 then begin
    vStack := TExpressionStack.Create;
    try

      // execute Tokens
      for Index := 0 to Count-1 do begin

        // try to do it!
        try
          // test our assertions!
          {$ifopt C+}
          if Tokens[Index] is TExpressionOperation then begin
            with TExpressionOperation(Tokens[Index]) do begin
              if ArgCount = cUnlimitedArgs then
                Assert(vStack.Atleast(vStack.Peek + 1), ResultTextFor(DISP_E_BADPARAMCOUNT))
              else
                Assert(vStack.Atleast(ArgCount), ResultTextFor(DISP_E_BADPARAMCOUNT));
            end;
          end;
          {$endif}

          // finally do it!
          TExpressionToken(Tokens[Index]).Execute(Self, vStack);

        // big time errors
        except
          SetErrorException;
          MoveResultAtTo(Tokens[Index].At);
        end;

        // any errors?
        if AnyErrors then begin
          MoveResultAtTo(Tokens[Index].At);
          break;
        end;
      end;

      // remember the result
      if not AnyErrors then begin
        Assert(vStack.Count = 1, ResultTextFor(DISP_E_PARAMNOTFOUND));
        FResultant := vStack.Pop;
      end;

    // clean up
    finally
      Dump(vStack);
    end;
  end;

  // any errors?
  RememberErrors;
  Result := not Errors;
end;
function TExpression.Evaluate: variant;
begin
  Execute;
  Result := Resultant;
end;

// last error info
procedure TExpression.RememberErrors;
begin
  FErrors := AnyErrors;
  FErrorCode := LastResultCode;
  FErrorText := LastResultText;
  FErrorAt := LastResultAt;
end;

procedure TExpression.ClearErrors;
begin
  ClearResult;
  RememberErrors;
end;

//==============================================================================
// TExpressionList = class(TFreeList)
//==============================================================================
function TExpressionList.GetExpression(Index: integer): TExpression;
begin
  Result := TExpression(Objects[Index]);
end;

//==============================================================================
// TIdentExpressionPairs = class(TStringList)
//==============================================================================
function TIdentExpressionPairs.GetIdent(index: integer): string;
begin
  Result := Strings[index];
end;
function TIdentExpressionPairs.GetExpression(index: integer): TExpression;
begin
  Result := TExpression(Objects[index]);
end;
function TIdentExpressionPairs.AddIdent(aIdent: string): integer;
begin
  Result := AddIdentExpression(aIdent, nil);
end;
function TIdentExpressionPairs.AddExpression(aExpr: TExpression): integer;
begin
  Result := AddIdentExpression('', aExpr);
end;
function TIdentExpressionPairs.AddIdentExpression(aIdent: string; aExpr: TExpression): integer;
begin
  Result := AddObject(aIdent, aExpr);
end;

end.
