unit ExpressionObjects;

//==============================================================================
interface

//==============================================================================
uses
  Windows, Messages, SysUtils, Classes, Controls, MiscSupport, SymbolSupport,Variants;

//==============================================================================
type
  //============================================================================
  TExpressionObject = class(TSymbolObject)
  private
    FItemIndex: variant;
  protected
    FCheckDefault: boolean;
    FDefaultProperty: string;
    function CheckName(const aProperty: string): string; virtual;
    function CheckReference: TPersistent; virtual;
  public
    constructor Create(const aSymbol: string; aReference: TPersistent);

    property ItemIndex: variant read FItemIndex write FItemIndex;
    function Indexable: boolean; virtual;
    function IndexableValue(const aProperty: string): boolean; virtual;
    function SelfReference: boolean; virtual;

    function Valid(const aProperty: string): boolean; virtual;
    function ValueOf(const aProperty: string): variant; virtual;
    function ValidIndex: boolean; virtual;
    property DefaultProperty: string read FDefaultProperty;

    function Syntax: string; virtual;
  end;

  //============================================================================
  TExpressionComponent = class(TExpressionObject)
  public
    constructor Create(const aSymbol: string; aReference: TComponent);
  end;

  //============================================================================
  TExpressionIndexedObject = class(TExpressionObject)
  public
    function Indexable: boolean; override;
  end;
  TExpressionIndexedComponent = class(TExpressionComponent)
  public
    function Indexable: boolean; override;
  end;

  //============================================================================
  TExpressionComponentCollection = class(TExpressionIndexedComponent)
  protected
    function CheckReference: TPersistent; override;
  end;

  //============================================================================
  TExpressionParentCollection = class(TExpressionIndexedComponent)
  protected
    function CheckReference: TPersistent; override;
  public
    constructor Create(const aSymbol: string; aReference: TWinControl);
  end;

  //============================================================================
  TExpressionObjectList = class(TSymbolList)
  public
    function Add(aObject: TExpressionObject): integer; virtual;
    function Lookup(const aSymbol: string): TExpressionObject; virtual;
    function LookupReference(aObject: TObject): TExpressionObject; virtual;
    procedure ListSymbols(aList: TStrings); virtual;
    procedure ListSyntaxes(aList: TStrings); virtual;
  end;

//==============================================================================
const
  cValueProperty   = 'Value';
  cNameProperty    = 'Name';
  cCaptionProperty = 'Caption';

//==============================================================================
implementation

//==============================================================================
uses
  ResultSupport, StringSupport, PropertySupport;

//==============================================================================
// TExpressionObject = class(TSymbolObject)
//==============================================================================
constructor TExpressionObject.Create(const aSymbol: string; aReference: TPersistent);
begin
  inherited Create(aSymbol, aReference);
  FCheckDefault := True;
  FDefaultProperty := cValueProperty;
end;

//==============================================================================
function TExpressionObject.CheckName(const aProperty: string): string;
begin
  if FCheckDefault then //begin
    FCheckDefault := False;
    {Assert(Valid(FDefaultProperty),
           ResultTextFor(TYPE_E_AMBIGUOUSNAME) + ': ' + FDefaultProperty);
  end;}

  Result := aProperty;
  if Result = '' then
    Result := DefaultProperty;
end;
function TExpressionObject.CheckReference: TPersistent;
begin
  if SelfReference then
    Result := Self
  else
    Result := TPersistent(Reference);
end;

//==============================================================================
function TExpressionObject.Indexable: boolean;
begin
  Result := False;
end;
function TExpressionObject.IndexableValue(const aProperty: string): boolean;
begin
  Result := Indexable;
end;
function TExpressionObject.ValidIndex: boolean;
begin
  Result := CheckReference <> nil;
end;
function TExpressionObject.SelfReference: boolean;
begin
  Result := False;
end;

//==============================================================================
function TExpressionObject.Valid(const aProperty: string): boolean;
begin
  Result := PropertyExists(CheckReference, CheckName(aProperty));
end;
function TExpressionObject.ValueOf(const aProperty: string): variant;
begin
  // try to get the result
  Result := GetPropertyValue(CheckReference, CheckName(aProperty));

  // prevent a bad situation from happening
  if VarIsEmpty(Result) then
    Result := '';
end;

//==============================================================================
function TExpressionObject.Syntax: string;
begin
  Result := 'value = ' + Symbol;
  if Indexable then
    Result := Result + '[value]';
  Result := Result + '{.';
  if DefaultProperty <> '' then
    Result := Result + DefaultProperty
  else
    Result := Result + 'property';
  Result := Result + '}';
end;

//==============================================================================
// TExpressionComponent = class(TExpressionObject)
//==============================================================================
constructor TExpressionComponent.Create(const aSymbol: string; aReference: TComponent);
begin
  inherited Create(aSymbol, aReference);
  FDefaultProperty := cNameProperty;
end;

//==============================================================================
// TExpressionIndexedObject = class(TExpressionObject)
//==============================================================================
function TExpressionIndexedObject.Indexable: boolean;
begin
  Result := True;
end;

//==============================================================================
// TExpressionIndexedComponent = class(TExpressionComponent)
//==============================================================================
function TExpressionIndexedComponent.Indexable: boolean;
begin
  Result := True;
end;

//==============================================================================
// TExpressionComponentCollection = class(TExpressionIndexedComponent)
//==============================================================================
function TExpressionComponentCollection.CheckReference: TPersistent;
begin
  if VarIsEmpty(ItemIndex) then
    Result := TPersistent(Reference)
  else
    Result := TComponent(Reference).FindComponent(ItemIndex);
  //Assert(Assigned(Result), ResultTextFor(TYPE_E_AMBIGUOUSNAME));
end;

//==============================================================================
// TExpressionParentCollection = class(TExpressionIndexedComponent)
//==============================================================================
constructor TExpressionParentCollection.Create(const aSymbol: string;
  aReference: TWinControl);
begin
  inherited Create(aSymbol, aReference);
  FDefaultProperty := cCaptionProperty;
end;
function TExpressionParentCollection.CheckReference: TPersistent;
var
  vIndex: integer;
begin
  if VarIsEmpty(ItemIndex) then
    Result := TPersistent(Reference)
  else begin
    Result := nil;
    for vIndex := 0 to TWinControl(Reference).ControlCount-1 do
      if AnsiCompare(ItemIndex, TWinControl(Reference).Controls[vIndex].Name) then begin
        Result := TWinControl(Reference).Controls[vIndex];
        break;
      end;
  end;
  //Assert(Assigned(Result), ResultTextFor(TYPE_E_AMBIGUOUSNAME));
end;

//==============================================================================
// TExpressionObjectList = class(TSymbolList)
//==============================================================================
function TExpressionObjectList.Add(aObject: TExpressionObject): integer;
begin
  Result := inherited Add(aObject);
end;
function TExpressionObjectList.Lookup(const aSymbol: string): TExpressionObject;
begin
  Result := TExpressionObject(inherited Lookup(aSymbol));
end;
function TExpressionObjectList.LookupReference(aObject: TObject): TExpressionObject;
begin
  Result := TExpressionObject(inherited LookupReference(aObject));
end;
procedure TExpressionObjectList.ListSymbols(aList: TStrings);
var
  vIndex: integer;
begin
  for vIndex := 0 to Count-1 do
    aList.Add(TExpressionObject(Objects[vIndex]).Symbol);
end;
procedure TExpressionObjectList.ListSyntaxes(aList: TStrings);
var
  vIndex: integer;
begin
  for vIndex := 0 to Count-1 do
    aList.Add(TExpressionObject(Objects[vIndex]).Syntax);
end;

end.
