unit ExpressionParser;

//==============================================================================
interface

//==============================================================================
uses
  Windows, Messages, SysUtils, Classes, MiscSupport, ParserSupport,
  Expression, ExpressionLexicon, ExpressionObjects, TokenSupport,Variants;

//==============================================================================
type
  //============================================================================
  TExpressionLookupValueEvent = procedure(Sender: TObject;
    const aSymbol: string; var aValue: variant) of object;
  TExpressionLookupObjectEvent = procedure(Sender: TObject;
    const aSymbol: string; var aObject: TExpressionObject) of object;
  TExpressionCheckTokenEvent = procedure(Sender: TObject;
    aExpression: TExpression; aToken: TTokenObject;
    var aContinue: boolean) of object;
  TExpressionValidValue = (evvAscii, evvNumber, evvIdent);
  TExpressionValidValues = set of TExpressionValidValue;
  TExpressionSubject = (esExpression, esFormula, esTerm, esUnary, esFunction, esObject);
  TExpressionSubjects = set of TExpressionSubject;

  //============================================================================
  TCustomExpressionParser = class(TComponent)
  private
    // what is a valid value?
    FValidValues: TExpressionValidValues;

    // last cursor used info
    FStoppedAt: integer;

    // last error info
    FErrorText: string;
    FErrorAt, FErrorCode: integer;
    FErrors: boolean;

    // lexicon
    FExpressionLexicon, FFormulaLexicon, FTermLexicon: TExpressionOpers2Lexicon;
    FUnaryLexicon: TExpressionOpers1Lexicon;
    FFunctionLexicon: TExpressionLexicon;
    FObjectLexicon: TExpressionObjectList;

    // late binding
    FOnLateValue: TExpressionLookupValueEvent;
    FOnLateObject: TExpressionLookupObjectEvent;

    // token checking
    FOnCheckToken: TExpressionCheckTokenEvent;
    FLastToken: TTokenObject;
  protected
    // last error info
    procedure RememberErrors; virtual;

    // worker bees
    function ParseSubExpression(aCursor: TParserCursor; aExpression: TExpression): boolean;
    procedure ParseFormula(aCursor: TParserCursor; aExpression: TExpression);
    procedure ParseTerm(aCursor: TParserCursor; aExpression: TExpression);
    procedure ParseFactor(aCursor: TParserCursor; aExpression: TExpression);
    procedure ParseValue(aCursor: TParserCursor; aExpression: TExpression);
    function ParseAscii(aCursor: TParserCursor; aExpression: TExpression): boolean;
    function ParseNumber(aCursor: TParserCursor; aExpression: TExpression): boolean;
    function ParseIdent(aCursor: TParserCursor; aExpression: TExpression): boolean;
    procedure AddToken(aExpression: TExpression; aToken: TTokenObject; aTokenAt: integer);
    property LastToken: TTokenObject read FLastToken;
  public
    // tinker toys
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    // late binding
    function LookupValue(const aSymbol: string): variant; virtual;
    function LookupObject(const aSymbol: string): TExpressionObject; virtual;
    property ValidValues: TExpressionValidValues read FValidValues write FValidValues;

    // meat and potato
    function ParseExpression(aCursor: TParserCursor): TExpression; virtual;
    function ParseToExpression(const aText: string): TExpression; virtual;
    function Evaluate(const aText: string): variant; virtual;
    property StoppedAt: integer read FStoppedAt;

    // last error info
    procedure DeclareErrorAt(aResult: HResult; At: integer); virtual;
    procedure ClearErrors;
    property ErrorAt: integer read FErrorAt;
    property ErrorCode: integer read FErrorCode;
    property ErrorText: string read FErrorText;
    property Errors: boolean read FErrors;

    // documentation
    procedure ListSyntaxes(aSubject: TExpressionSubjects; aList: TStrings); virtual;
    procedure ListSymbols(aSubject: TExpressionSubjects; aList: TStrings); virtual;
    function CountOf(aSubject: TExpressionSubjects): integer; virtual;
    function LexiconCount: integer;

    // lexicon
    property ExpressionLexicon: TExpressionOpers2Lexicon read FExpressionLexicon;
    property FormulaLexicon: TExpressionOpers2Lexicon read FFormulaLexicon;
    property TermLexicon: TExpressionOpers2Lexicon read FTermLexicon;
    property UnaryLexicon: TExpressionOpers1Lexicon read FUnaryLexicon;
    property FunctionLexicon: TExpressionLexicon read FFunctionLexicon;
    property ObjectLexicon: TExpressionObjectList read FObjectLexicon;

    // publishable
    property OnLateValue: TExpressionLookupValueEvent
      read FOnLateValue write FOnLateValue;
    property OnLateObject: TExpressionLookupObjectEvent
      read FOnLateObject write FOnLateObject;
    property OnCheckToken: TExpressionCheckTokenEvent
      read FOnCheckToken write FOnCheckToken;
  end;

  //============================================================================
  TCustomExpressIt = class(TCustomExpressionParser)
  private
    // private bits
    FAutoExecute: boolean;
    FExpression, FResultText: string;
    FResultant: variant;
  protected
    // access to private bits
    procedure SetExpression(aExpression: string); virtual;
    procedure SetResultant(aResultant: variant); virtual;
    procedure SetAutoExecute(aValue: boolean); virtual;
  public
    // alpha
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Value: TPersistent); override;

    // property based evaluation
    function Evaluate(const aText: string): variant; override;
    function Execute: boolean; virtual;
    property Resultant: variant read FResultant write SetResultant;

    // publishable
    property AutoExecute: boolean read FAutoExecute write SetAutoExecute default False;
    property Expression: string read FExpression write SetExpression;
    property ResultText: string read FResultText write FResultText stored False;
  end;

  //============================================================================
  TCustomExpressioneer = class(TComponent)
  private
    // private bits
    FExpressIt: TCustomExpressIt;
    FAutoExecute: boolean;
    FExpression, FResultText: string;
    FResultant: variant;
    FParsedExpression: TExpression;
  protected
    // access to private bits
    procedure SetExpressIt(aExpressIt: TCustomExpressIt); virtual;
    procedure SetExpression(aExpression: string); virtual;
    procedure SetResultant(aResultant: variant); virtual;
    procedure SetAutoExecute(aValue: boolean); virtual;
  public
    // alpha and omega
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Value: TPersistent); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    // work
    function Evaluate: variant; virtual;
    function Execute: boolean; virtual;
    property Resultant: variant read FResultant write SetResultant;
    property ParsedExpression: TExpression read FParsedExpression;

    // last error info
    procedure ClearErrors;
    function ErrorAt: integer;
    function ErrorCode: integer;
    function ErrorText: string;
    function Errors: boolean;

    // publishable
    property ExpressIt: TCustomExpressIt read FExpressIt write SetExpressIt;
    property AutoExecute: boolean read FAutoExecute write SetAutoExecute default False;
    property Expression: string read FExpression write SetExpression;
    property ResultText: string read FResultText write FResultText stored False;
  end;

//==============================================================================
const
  cListAll = [esExpression, esFormula, esTerm, esUnary, esFunction, esObject];
  cListOperators = [esExpression, esFormula, esTerm, esUnary];
  cListFunctions = [esFunction];
  cListObjects = [esObject];

//==============================================================================
implementation

//==============================================================================
uses
  MiscConstants, ResultSupport, StringSupport,
  ExpressionConstants, ExpressionTokens;

//============================================================================
// TCustomExpressionParser = class(TComponent)
//============================================================================
// tinker toys
constructor TCustomExpressionParser.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);

  // support stuff
  FValidValues := [evvAscii, evvNumber, evvIdent];
  FStoppedAt := -1;

  // various lexicon
  FExpressionLexicon := TExpressionOpers2Lexicon.Create;
  FFormulaLexicon := TExpressionOpers2Lexicon.Create;
  FTermLexicon := TExpressionOpers2Lexicon.Create;
  FUnaryLexicon := TExpressionOpers1Lexicon.Create;
  FFunctionLexicon := TExpressionLexicon.Create;
  FObjectLexicon := TExpressionObjectList.Create;
end;
destructor TCustomExpressionParser.Destroy;
begin
  // dump the lexicons
  Dump(FExpressionLexicon);
  Dump(FFormulaLexicon);
  Dump(FTermLexicon);
  Dump(FUnaryLexicon);
  Dump(FFunctionLexicon);
  Dump(FObjectLexicon);

  // free the lists
  inherited Destroy;
end;

//==============================================================================
procedure TCustomExpressionParser.RememberErrors;
begin
  // save and go
  FErrors := AnyErrors;
  FErrorCode := LastResultCode;
  FErrorText := LastResultText;
  FErrorAt := LastResultAt;
end;
procedure TCustomExpressionParser.ClearErrors;
begin
  ClearResult;
  RememberErrors;
end;

//==============================================================================
// late binding
function TCustomExpressionParser.LookupValue(const aSymbol: string): variant;
begin
  VarClear(Result);
  if Assigned(FOnLateValue) then
    FOnLateValue(Self, aSymbol, Result);
end;
function TCustomExpressionParser.LookupObject(const aSymbol: string): TExpressionObject;
begin
  Result := nil;
  if Assigned(FOnLateObject) then
    FOnLateObject(Self, aSymbol, Result);
end;

//==============================================================================
// meat and potato
//============================================================================
// Expression = Formula [EXPRESSIONOP Formula]
function TCustomExpressionParser.ParseSubExpression(aCursor: TParserCursor; aExpression: TExpression): boolean;
var
  vToken: TTokenObject;
  vTokenCount, vTokenAt, vCount: integer;
begin
  // how are there?
  vCount := aExpression.Count;

  // continue parsing?
  vTokenCount := aExpression.Count;
  ParseFormula(aCursor, aExpression);
  while not AnyErrors do begin

    // set up
    vTokenAt := aCursor.At;
    vToken := aCursor.NextTokenOf(ExpressionLexicon);

    // we made a token
    if Assigned(vToken) then begin
      if vTokenCount = aExpression.Count then
        DeclareErrorAt(MK_E_SYNTAX, vTokenAt);
      vTokenCount := aExpression.Count;
      ParseFormula(aCursor, aExpression);
      if vTokenCount = aExpression.Count then
        DeclareErrorAt(MK_E_SYNTAX, vTokenAt);
      if not AnyErrors then
        AddToken(aExpression, vToken, vTokenAt)
      else
        Dump(vToken);
    end else
      break;
  end;

  // did we add stuff?
  Result := vCount <> aExpression.Count;
end;

//============================================================================
// Formula = Term [FORMULAOP Term]
procedure TCustomExpressionParser.ParseFormula(aCursor: TParserCursor; aExpression: TExpression);
var
  vToken: TTokenObject;
  vTokenCount, vTokenAt: integer;
begin
  // continue parsing?
  vTokenCount := aExpression.Count;
  ParseTerm(aCursor, aExpression);
  while not AnyErrors do begin

    // set up
    vTokenAt := aCursor.At;
    vToken := aCursor.NextTokenOf(FormulaLexicon);

    // we made a token
    if Assigned(vToken) then begin
      if vTokenCount = aExpression.Count then
        DeclareErrorAt(MK_E_SYNTAX, vTokenAt);
      vTokenCount := aExpression.Count;
      ParseTerm(aCursor, aExpression);
      if vTokenCount = aExpression.Count then
        DeclareErrorAt(MK_E_SYNTAX, vTokenAt);
      if not AnyErrors then
        AddToken(aExpression, vToken, vTokenAt)
      else
        Dump(vToken);
    end else
      break;
  end;
end;

//============================================================================
// Term = Factor [TERMOP Factor]
procedure TCustomExpressionParser.ParseTerm(aCursor: TParserCursor; aExpression: TExpression);
var
  vToken: TTokenObject;
  vTokenCount, vTokenAt: integer;
begin
  // continue parsing?
  vTokenCount := aExpression.Count;
  ParseFactor(aCursor, aExpression);
  while not AnyErrors do begin

    // set up
    vTokenAt := aCursor.At;
    vToken := aCursor.NextTokenOf(TermLexicon);

    // we made a token
    if Assigned(vToken) then begin
      if vTokenCount = aExpression.Count then
        DeclareErrorAt(MK_E_SYNTAX, vTokenAt);
      vTokenCount := aExpression.Count;
      ParseFactor(aCursor, aExpression);
      if vTokenCount = aExpression.Count then
        DeclareErrorAt(MK_E_SYNTAX, vTokenAt);
      if not AnyErrors then
        AddToken(aExpression, vToken, vTokenAt)
      else
        Dump(vToken);
    end else
      break;
  end;
end;

//============================================================================
// Factor = UNARYOP Factor | (Expression) | Value
procedure TCustomExpressionParser.ParseFactor(aCursor: TParserCursor; aExpression: TExpression);
var
  vToken: TTokenObject;
  vTokenCount, vTokenAt: integer;
begin
  // set up
  vTokenAt := aCursor.At;
  vToken := aCursor.NextTokenOf(UnaryLexicon);

  // token?
  if Assigned(vToken) then begin
    vTokenCount := aExpression.Count;
    ParseFactor(aCursor, aExpression);
    if vTokenCount = aExpression.Count then
      DeclareErrorAt(MK_E_SYNTAX, vTokenAt);
    if not AnyErrors then
      AddToken(aExpression, vToken, vTokenAt)
    else
      Dump(vToken);

  // nested Expression
  end else if aCursor.NextIs('(') then begin

    // do the in between parts
    if not ParseSubExpression(aCursor, aExpression) then
      DeclareErrorAt(MK_E_SYNTAX, aCursor.At);

    // better have a closing para
    if (not AnyErrors) and
       (not aCursor.NextIs(')')) then begin
      aCursor.SetAt(vTokenAt);
      DeclareErrorAt(MK_E_SYNTAX, aCursor.At);
    end;

  // otherwise its just a value
  end else
    ParseValue(aCursor, aExpression);
end;

//============================================================================
// Value = "Ascii" |                                          // quoted string
//         Number |                                          // decimal number
//         Ident                                              // look up value
procedure TCustomExpressionParser.ParseValue(aCursor: TParserCursor; aExpression: TExpression);
var
  vContinue: boolean;
begin
  // do ascii?
  if evvAscii in ValidValues then
    vContinue := ParseAscii(aCursor, aExpression)
  else
    vContinue := True;

  // do numbers?
  if vContinue and
     (evvNumber in ValidValues) then
    vContinue := ParseNumber(aCursor, aExpression);

  // do idents?
  if vContinue and
     (evvIdent in ValidValues) then
    ParseIdent(aCursor, aExpression);
end;

//============================================================================
// Ascii = [Ascii][\"][\\][\000]"   // ascii cannot include an unescaped quote
function TCustomExpressionParser.ParseAscii(aCursor: TParserCursor; aExpression: TExpression): boolean;
var
  vData, vCodeStr: string;
  vToken: TTokenObject;
  vTokenAt, vCode: integer;
begin
  // assume failure
  Result := True;

  // ok to continue?
  if not AnyErrors then begin

    // where did we start?
    vToken := nil;
    vTokenAt := aCursor.At;
    vData := '';

    // must start with a quote!
    if aCursor.NextIs(vQuoteChar) then begin

      // as long as we have valid characters
      while aCursor.TheChar in cValidAscii do begin

        // escaping!
        if aCursor.TheChar = vEscapeChar then begin
          aCursor.NextChar;

          // inline quote
          if aCursor.TheChar = vQuoteChar then
            vData := vData + vQuoteChar

          // inline slash
          else if aCursor.TheChar = vEscapeChar then
            vData := vData + vEscapeChar

          // lets try some other stuff
          else begin

            // 000
            vCodeStr := aCursor.NextNCharsIs(3, cDigits);
            if vCodeStr <> '' then begin
              vCode := StrToIntDef(vCodeStr, -1);

              // out of range, scream!
              if (vCode < 0) or
                 (vCode > 255) then begin
                DeclareErrorAt(OLE_E_CANTCONVERT, aCursor.At);
                Break;

              // encode it please!
              end else
                vData := vData + Chr(vCode);

            // real confused!
            end else begin
              DeclareErrorAt(MK_E_SYNTAX, aCursor.At);
              Break;
            end;
          end;
        end else
          vData := vData + aCursor.TheChar;
        aCursor.NextChar;
      end;

      // we end in a quote!
      if aCursor.NextIs(vQuoteChar) then
        vToken := TExpressionConstant.Create(vData)
      else
        DeclareErrorAt(MK_E_SYNTAX, aCursor.At);
    end;

    // all is swell!
    if Assigned(vToken) then begin
      if not AnyErrors then begin
        AddToken(aExpression, vToken, vTokenAt);
        Result := False;
      end else
        Dump(vToken);
    end;
  end;
end;

//============================================================================
// Number = 0~9[0~9#][.0~9[0~9#]]                           // integer | float
function TCustomExpressionParser.ParseNumber(aCursor: TParserCursor; aExpression: TExpression): boolean;
var
  vData: string;
  vToken: TTokenObject;
  vTokenAt: integer;
  vValue: variant;
begin
  // assume failure
  Result := True;
  vValue := Unassigned;

  // ok to continue?
  if not AnyErrors then begin

    // where did we start?
    vToken := nil;
    vTokenAt := aCursor.At;
    vData := '';

    // keep looking for value
    vData := aCursor.NextDigits;

    // got something?
    if (vData <> '') then begin

      // floating point?
      if aCursor.NextIs('.') then begin

        // do we do reals?
        vData := vData + '.';

        // keep looking for the rest of the value
        if aCursor.TheChar in cDigits then
          vData := vData + aCursor.NextDigits
        else
          DeclareErrorAt(MK_E_SYNTAX, aCursor.At);

        // try to convert it
        try
          vValue := StrToFloat(vData);
        except
          aCursor.SetAt(vTokenAt);
          DeclareErrorAt(OLE_E_CANTCONVERT, aCursor.At);
        end;

      // integer then?
      end else begin
        try
          vValue := StrToInt(vData);
        except
          aCursor.SetAt(vTokenAt);
          DeclareErrorAt(OLE_E_CANTCONVERT, aCursor.At);
        end;
      end;

      // add the value to the stack
      if not VarIsEmpty(vValue) then
        vToken := TExpressionConstant.Create(vValue);
    end;

    // all is swell!
    if Assigned(vToken) then begin
      if not AnyErrors then begin
        AddToken(aExpression, vToken, vTokenAt);
        Result := False;
      end else
        Dump(vToken);
    end;
  end;
end;

//============================================================================
// Ident = Function[(Expression [, Expression#])]        // function reference
//         Object[[Expression]][.Property[.Property#]]     // object reference
function TCustomExpressionParser.ParseIdent(aCursor: TParserCursor; aExpression: TExpression): boolean;
var
  vToken: TTokenObject;
  vFunction: TExpressionOperationClass;
  vObject: TExpressionObject;
  vSymbol, vSubSymbol: string;
  vTokenAt, vSubSymbolAt, vArgCount: integer;
  vIndexed, vValidateIndex, vLateLookup: boolean;
  function ParseProperty: string;
  begin
    // assume nothing
    Result := '';

    // grab the properties
    while aCursor.NextIs('.') do
      Result := Result + '.' + aCursor.NextSymbol;

    // make sure we don't have a preceding period
    Result := Copy(Result, 2, Length(Result));
  end;
begin
  // assume failure!
  Result := True;

  // ok to continue?
  if not AnyErrors then begin

    // where did we start?
    vToken := nil;
    vTokenAt := aCursor.At;
    vSymbol := '';
    vSubSymbol := '';

    // look for the first label
    vSymbol := aCursor.NextSymbol;

    // start of something good?
    if vSymbol <> '' then begin

      // is it a Expression?
      vFunction := FunctionLexicon.Lookup(vSymbol);
      if Assigned(vFunction) then begin

        // is there arguments?
        if aCursor.NextIs('(') then begin

          // too many?
          if vFunction.ArgCount = 0 then
            DeclareErrorAt(MK_E_SYNTAX, aCursor.At)

          // how many then?
          else begin
            vArgCount := 0;

            // keep parsing the arguments
            if not ParseSubExpression(aCursor, aExpression) then
              DeclareErrorAt(MK_E_SYNTAX, aCursor.At);

            // let get happening
            while not AnyErrors do begin

              // we got on
              Inc(vArgCount);

              // another argument found?
              if aCursor.NextIs(',') then begin

                // another Expression?
                if (vFunction.ArgCount = cUnlimitedArgs) or
                   (vArgCount <= vFunction.ArgCount) then begin

                  if not ParseSubExpression(aCursor, aExpression) then
                    DeclareErrorAt(MK_E_SYNTAX, aCursor.At);

                // too many arguments?
                end else
                  DeclareErrorAt(TYPE_E_OUTOFBOUNDS, aCursor.At)

              // found the end of the arguments
              end else if aCursor.NextIs(')') then begin

                // are there the wrong number of arguments?
                if vFunction.ArgCount = cUnlimitedArgs then begin

                  // not enough!
                  if vArgCount < 2 then
                    DeclareErrorAt(TYPE_E_OUTOFBOUNDS, aCursor.At)
                  else
                    AddToken(aExpression, TExpressionConstant.Create(vArgCount), vTokenAt);

                end else if vArgCount < vFunction.ArgCount then
                  DeclareErrorAt(TYPE_E_OUTOFBOUNDS, aCursor.At);

                // all is swell
                break;

              // dammit, that is not right!
              end else begin
                DeclareErrorAt(MK_E_SYNTAX, aCursor.At);
                break;
              end;
            end;
          end;
        end;

        // make the function token?
        if not AnyErrors and
           Assigned(vFunction) then
          vToken := vFunction.Create;

      // check for object reference then
      end else if not AnyErrors then begin

        // does the object exists or can we do a late lookup
        vObject := ObjectLexicon.Lookup(vSymbol);
        vLateLookup := not Assigned(vObject) and
                       (Assigned(FOnLateObject) or
                        Assigned(FOnLateValue));
        vValidateIndex := False;

        // object or late lookup
        if Assigned(vObject) or
           vLateLookup then begin

          // Indexed?
          vIndexed := aCursor.NextIs('[');
          if vIndexed then begin
            vValidateIndex := True;

            // should we do this?
            if (Assigned(vObject) and
                not vObject.Indexable) then
              DeclareErrorAt(DISP_E_NOTACOLLECTION, aCursor.At);

            // keep parsing, its an index
            if (not AnyErrors) and
               (not (ParseSubExpression(aCursor, aExpression) and
                     aCursor.NextIs(']'))) then
              DeclareErrorAt(MK_E_SYNTAX, aCursor.At);
          end;

          // look for properties
          if not AnyErrors then begin
            vSubSymbolAt := aCursor.At;
            vSubSymbol := ParseProperty;

            // its an object
            if not AnyErrors and
               Assigned(vObject) then begin

              // valid index?
              if not vObject.Valid(vSubSymbol) then
                DeclareErrorAt(TYPE_E_ELEMENTNOTFOUND, vSubSymbolAt + 1)

              // symbol exists but is not indexable
              else if vIndexed and
                (not vObject.IndexableValue(vSubSymbol)) then
                DeclareErrorAt(DISP_E_NOTACOLLECTION, aCursor.At);

              // ok if we are indexed
              if (not AnyErrors) and
                 (vIndexed) and
                 (LastToken is TExpressionConstant) then begin
                vObject.ItemIndex := TExpressionConstant(LastToken).Value;
                if not vObject.ValidIndex then
                  DeclareErrorAt(TYPE_E_AMBIGUOUSNAME, LastToken.At)
                else
                  vValidateIndex := False;
              end;
            end;
          end;

          // all is swell?
          if not AnyErrors then begin

            // object, late look or for real?
            if vLateLookup then
              vToken := TExpressionLateProperty.Create(vSymbol, vSubSymbol)
            else
              vToken := TExpressionProperty.Create(vObject, vSubSymbol);

            // remember the index that is on the stack
            TExpressionPropertyToken(vToken).Indexed := vIndexed;
            TExpressionPropertyToken(vToken).ValidateIndex := vValidateIndex;
          end;
        end;

        // no token then?
        if not Assigned(vToken) then
          DeclareErrorAt(TYPE_E_AMBIGUOUSNAME, vTokenAt);
      end;
    end;

    // add the token then
    if Assigned(vToken) then begin
      if not AnyErrors then begin
        AddToken(aExpression, vToken, vTokenAt);
        Result := False;
      end else
        Dump(vToken);
    end;
  end;
end;

//============================================================================
procedure TCustomExpressionParser.AddToken(aExpression: TExpression; aToken: TTokenObject; aTokenAt: integer);
var
  vContinue: boolean;
begin
  // all is swell?
  Assert(Assigned(aToken), ResultTextFor(STG_E_INVALIDHANDLE));
  aToken.At := aTokenAt;

  // let the user check the token out
  vContinue := True;
  if Assigned(FOnCheckToken) then
    FOnCheckToken(Self, aExpression, aToken, vContinue);

  // keep going?
  if vContinue then
    aExpression.Add(aToken)
  else begin
    if not AnyErrors then
      DeclareErrorAt(STG_E_INVALIDHANDLE, aTokenAt);
    Dump(aToken);
  end;

  // remember the last token
  FLastToken := aToken;
end;

//============================================================================
function TCustomExpressionParser.ParseExpression(aCursor: TParserCursor): TExpression;
var
  vExpression: TExpression;
begin
  // assume failure
  Result := nil;
  ClearResult;
  FStoppedAt := -1;

  vExpression := TExpression.Create(Self);
  try

    // lets do it
    if not aCursor.AtEnd then
      ParseSubExpression(aCursor, vExpression);

    // success?
    if not AnyErrors then
      Result := vExpression
    else
      Dump(vExpression);

    // remember where we ended up
    FStoppedAt := aCursor.At;

  // oops an error happened
  except
    Dump(vExpression);
    SetErrorException;
  end;

  // store the error
  RememberErrors;
end;
function TCustomExpressionParser.ParseToExpression(const aText: string): TExpression;
var
  vCursor: TParserCursor;
begin
  // create the token stream
  vCursor := TParserCursor.Create(aText, 1);
  try
    // try to parse
    Result := ParseExpression(vCursor);
  finally
    Dump(vCursor);
  end;
end;
function TCustomExpressionParser.Evaluate(const aText: string): variant;
var
  vExpression: TExpression;
begin
  // assume failure
  VarClear(Result);

  // parse it
  vExpression := ParseToExpression(aText);

  // did it work?
  if Assigned(vExpression) then begin

    // see what happens
    if (not AnyErrors) and
       vExpression.Execute then
      Result := vExpression.Resultant
    else
      RememberErrors;

    // lets get rid of it then
    Dump(vExpression);
  end;
end;
procedure TCustomExpressionParser.DeclareErrorAt(aResult: HResult; At: integer);
begin
  // well if there isn't an error there will be one now
  if not AnyErrors then
    SetResultAt(aResult, At);
end;

//==============================================================================
// documentation
procedure TCustomExpressionParser.ListSyntaxes(aSubject: TExpressionSubjects; aList: TStrings);
begin
  aList.Clear;
  if esExpression in aSubject then
    ExpressionLexicon.ListSyntaxes(aList);
  if esFormula in aSubject then
    FormulaLexicon.ListSyntaxes(aList);
  if esTerm in aSubject then
    TermLexicon.ListSyntaxes(aList);
  if esUnary in aSubject then
    UnaryLexicon.ListSyntaxes(aList);
  if esFunction in aSubject then
    FunctionLexicon.ListSyntaxes(aList);
  if esObject in aSubject then
    ObjectLexicon.ListSyntaxes(aList);
end;
procedure TCustomExpressionParser.ListSymbols(aSubject: TExpressionSubjects; aList: TStrings);
begin
  aList.Clear;
  if esExpression in aSubject then
    ExpressionLexicon.ListSymbols(aList);
  if esFormula in aSubject then
    FormulaLexicon.ListSymbols(aList);
  if esTerm in aSubject then
    TermLexicon.ListSymbols(aList);
  if esUnary in aSubject then
    UnaryLexicon.ListSymbols(aList);
  if esFunction in aSubject then
    FunctionLexicon.ListSymbols(aList);
  if esObject in aSubject then
    ObjectLexicon.ListSymbols(aList);
end;
function TCustomExpressionParser.CountOf(aSubject: TExpressionSubjects): integer;
begin
  Result := 0;
  if esExpression in aSubject then
    Inc(Result, ExpressionLexicon.Count);
  if esFormula in aSubject then
    Inc(Result, FormulaLexicon.Count);
  if esTerm in aSubject then
    Inc(Result, TermLexicon.Count);
  if esUnary in aSubject then
    Inc(Result, UnaryLexicon.Count);
  if esFunction in aSubject then
    Inc(Result, FunctionLexicon.Count);
  if esObject in aSubject then
    Inc(Result, ObjectLexicon.Count);
end;
function TCustomExpressionParser.LexiconCount: integer;
begin
  Result := CountOf(cListAll);
end;

//==============================================================================
// TCustomExpressIt = class(TComponent)
//==============================================================================
constructor TCustomExpressIt.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAutoExecute := False;
  FResultant := Unassigned;
end;
destructor TCustomExpressIt.Destroy;
begin
  FResultant := Unassigned;
  inherited Destroy;
end;
procedure TCustomExpressIt.Assign(Value: TPersistent);
begin
  // one of us?
  if Value is TCustomExpressioneer then begin
    AutoExecute := False;
    Expression := TCustomExpressioneer(Value).Expression;
    AutoExecute := TCustomExpressioneer(Value).AutoExecute;
    Resultant := TCustomExpressioneer(Value).Resultant;

  // one of them?
  end else if Value is TCustomExpressIt then begin
    AutoExecute := False;
    Expression := TCustomExpressIt(Value).Expression;
    AutoExecute := TCustomExpressIt(Value).AutoExecute;
    Resultant := TCustomExpressIt(Value).Resultant;

  // pass the buck
  end else
    inherited Assign(Value);
end;

//==============================================================================
procedure TCustomExpressIt.SetExpression(aExpression: string);
begin
  // any change?
  if aExpression <> Expression then begin
    FExpression := aExpression;

    // do it?
    if AutoExecute then
      Execute;
  end;
end;
procedure TCustomExpressIt.SetResultant(aResultant: variant);
begin
  // store it
  FResultant := aResultant;
  ResultText := VarAsDisplayText(aResultant);
end;
procedure TCustomExpressIt.SetAutoExecute(aValue: boolean);
begin
  // any change?
  if aValue <> AutoExecute then begin
    FAutoExecute := aValue;

    // if now auto execute, do it!
    if AutoExecute then
      Execute;
  end;
end;

//==============================================================================
function TCustomExpressIt.Evaluate(const aText: string): variant;
begin
  // try to evaluate it
  Resultant := inherited Evaluate(aText);

  // back at ya
  Result := Resultant;
end;
function TCustomExpressIt.Execute: boolean;
begin
  // try to evaluate it
  Evaluate(Expression);

  // what to return!
  Result := not Errors;
end;

//==============================================================================
// TCustomExpressioneer = class(TComponent)
//==============================================================================
// alpha and omega
constructor TCustomExpressioneer.Create(aOwner: TComponent);
begin
  // pass the buck
  inherited Create(aOwner);

  // our private bits
  FAutoExecute := False;
  FResultant := Unassigned;
end;
destructor TCustomExpressioneer.Destroy;
begin
  // our private bits
  Dump(FParsedExpression);
  FResultant := Unassigned;

  // pass the buck
  inherited Destroy;
end;
procedure TCustomExpressioneer.Assign(Value: TPersistent);
begin
  // one of us?
  if Value is TCustomExpressioneer then begin
    AutoExecute := False;
    ExpressIt := TCustomExpressioneer(Value).ExpressIt;
    Expression := TCustomExpressioneer(Value).Expression;
    AutoExecute := TCustomExpressioneer(Value).AutoExecute;
    Resultant := TCustomExpressioneer(Value).Resultant;

  // one of them?
  end else if Value is TCustomExpressIt then begin
    AutoExecute := False;
    ExpressIt := TCustomExpressIt(Value);
    Expression := TCustomExpressIt(Value).Expression;
    AutoExecute := TCustomExpressIt(Value).AutoExecute;
    Resultant := TCustomExpressIt(Value).Resultant;

  // pass the buck
  end else
    inherited Assign(Value);
end;
procedure TCustomExpressioneer.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if (Operation = opRemove) and
     (AComponent = ExpressIt) then
    ExpressIt := nil;
  inherited Notification(AComponent, Operation);
end;

// access to private bits
procedure TCustomExpressioneer.SetExpressIt(aExpressIt: TCustomExpressIt);
begin
  // any change?
  if aExpressIt <> ExpressIt then begin
    FExpressIt := aExpressIt;
    Dump(FParsedExpression);

    // do it?
    if AutoExecute then
      Execute;
  end;
end;
procedure TCustomExpressioneer.SetExpression(aExpression: string);
begin
  // any change?
  if aExpression <> Expression then begin
    FExpression := aExpression;
    Dump(FParsedExpression);

    // do it?
    if AutoExecute then
      Execute;
  end;
end;
procedure TCustomExpressioneer.SetResultant(aResultant: variant);
begin
  // store it
  FResultant := aResultant;
  ResultText := VarAsDisplayText(aResultant);
end;
procedure TCustomExpressioneer.SetAutoExecute(aValue: boolean);
begin
  // any change?
  if aValue <> AutoExecute then begin
    FAutoExecute := aValue;

    // if now auto execute, do it!
    if AutoExecute then
      Execute;
  end;
end;

// work
function TCustomExpressioneer.Evaluate: variant;
begin
  // has it been parsed yet?
  if not Assigned(FParsedExpression) and
     Assigned(FExpressIt) then
    FParsedExpression := ExpressIt.ParseToExpression(FExpression);

  // has it really been parsed?
  if Assigned(FParsedExpression) then
    Resultant := ParsedExpression.Evaluate
  else
    Resultant := Null;

  // back at ya
  Result := Resultant;
end;
function TCustomExpressioneer.Execute: boolean;
begin
  // try to evaluate it
  Evaluate;

  // what to return!
  Result := not Errors;
end;

// last error info
procedure TCustomExpressioneer.ClearErrors;
begin
  if Assigned(FParsedExpression) then
    ParsedExpression.ClearErrors;
end;
function TCustomExpressioneer.ErrorAt: integer;
begin
  Result := 0;
  if Assigned(FParsedExpression) then
    Result := ParsedExpression.ErrorAt;
end;
function TCustomExpressioneer.ErrorCode: integer;
begin
  Result := 0;
  if Assigned(FParsedExpression) then
    Result := ParsedExpression.ErrorCode
  else if VarType(Resultant) = varNull then
    Result := -1;
end;
function TCustomExpressioneer.ErrorText: string;
begin
  Result := '';
  if Assigned(FParsedExpression) then
    Result := ParsedExpression.ErrorText
  else if VarType(Resultant) = varNull then
    Result := 'Could not parse expression';
end;
function TCustomExpressioneer.Errors: boolean;
begin
  Result := (VarType(Resultant) = varNull) or
            (Assigned(FParsedExpression) and
             ParsedExpression.Errors);
end;

end.
