unit ExpressionTokens;

//==============================================================================
interface

//==============================================================================
uses
  Windows, Messages, SysUtils, Classes, MiscSupport, TokenSupport,
  ExpressionStack, ExpressionObjects,Variants;

//==============================================================================
type
  //============================================================================
  TExpressionToken = class(TTokenObject)
  public
    procedure Execute(Sender: TObject; aStack: TExpressionStack); virtual; abstract;
  end;

  //============================================================================
  TExpressionCode = class(TExpressionToken)
  public
    function Caption: string; override;
  end;

  //============================================================================
  TExpressionConstant = class(TExpressionToken)
  private
    FValue: variant;
  public
    constructor Create(const aValue: variant);
    procedure Execute(Sender: TObject; aStack: TExpressionStack); override;
    class function Symbol: string; override;
    function Caption: string; override;

    property Value: variant read FValue;
  end;

  //============================================================================
  TExpressionPropertyToken = class(TExpressionToken)
  private
    FIndexed, FValidateIndex: boolean;
  public
    class function Symbol: string; override;
    property Indexed: boolean read FIndexed write FIndexed;
    property ValidateIndex: boolean read FValidateIndex write FValidateIndex;
  end;

  //============================================================================
  TExpressionProperty = class(TExpressionPropertyToken)
  private
    FObjectRef: TExpressionObject;
    FPropertyIdent: string;
  public
    constructor Create(aObjectRef: TExpressionObject; const aPropertyIdent: string);
    procedure Execute(Sender: TObject; aStack: TExpressionStack); override;
    function Caption: string; override;

    property ObjectRef: TExpressionObject read FObjectRef;
    property PropertyIdent: string read FPropertyIdent;
  end;

  //============================================================================
  TExpressionLateProperty = class(TExpressionPropertyToken)
  private
    FObjectIdent, FPropertyIdent: string;
  protected
    function ObjectNProperty: string; virtual;
  public
    constructor Create(const aObjectIdent, aPropertyIdent: string);
    procedure Execute(Sender: TObject; aStack: TExpressionStack); override;
    class function Symbol: string; override;
    function Caption: string; override;

    property ObjectIdent: string read FObjectIdent;
    property PropertyIdent: string read FPropertyIdent;
  end;

//==============================================================================
implementation

//==============================================================================
uses
  ResultSupport, Expression, ExpressionParser;

//==============================================================================
// TExpressionCode = class(TExpressionToken)
//==============================================================================
function TExpressionCode.Caption: string;
begin
  Result := 'Function: ' + Symbol;
end;

//==============================================================================
// TExpressionConstant = class(TExpressionToken)
//==============================================================================
constructor TExpressionConstant.Create(const aValue: variant);
begin
  inherited Create;
  FValue := aValue;
end;
procedure TExpressionConstant.Execute(Sender: TObject; aStack: TExpressionStack);
begin
  aStack.Push(FValue);
end;
class function TExpressionConstant.Symbol: string;
begin
  Result := 'Constant';
end;
function TExpressionConstant.Caption: string;
begin
  Result := VarAsDisplayText(FValue);
end;

//==============================================================================
// TExpressionPropertyToken = class(TExpressionToken)
//==============================================================================
class function TExpressionPropertyToken.Symbol: string;
begin
  Result := 'Property';
end;

//==============================================================================
// TExpressionProperty = class(TExpressionPropertyToken)
//==============================================================================
constructor TExpressionProperty.Create(aObjectRef: TExpressionObject;
                                       const aPropertyIdent: string);
begin
  inherited Create;
  FObjectRef := aObjectRef;
  FPropertyIdent := aPropertyIdent;
end;
procedure TExpressionProperty.Execute(Sender: TObject; aStack: TExpressionStack);
begin
  with aStack do begin
    if Indexed then begin
      ObjectRef.ItemIndex := Pop;

      // valid index?
      if ValidateIndex and
         (not ObjectRef.ValidIndex) then
        SetResult(TYPE_E_AMBIGUOUSNAME);
    end else
      ObjectRef.ItemIndex := Unassigned;
    Push(ObjectRef.ValueOf(PropertyIdent));
  end;
end;
function TExpressionProperty.Caption: string;
begin
  Result := 'Object: ' + ObjectRef.Symbol;
  if Indexed then
    Result := Result + '[..]';
  if PropertyIdent <> '' then
     Result := Result + '.' + PropertyIdent;
end;

//==============================================================================
// TExpressionLateProperty = class(TExpressionPropertyToken)
//==============================================================================
constructor TExpressionLateProperty.Create(const aObjectIdent, aPropertyIdent: string);
begin
  inherited Create;
  FObjectIdent := aObjectIdent;
  FPropertyIdent := aPropertyIdent;
end;
procedure TExpressionLateProperty.Execute(Sender: TObject; aStack: TExpressionStack);
var
  vObject: TExpressionObject;
  vParser: TCustomExpressionParser;
  vValue: variant;
begin
  // find parser
  vParser := TCustomExpressionParser(TExpression(Sender).Parser);

  // find the object
  vObject := vParser.LookupObject(ObjectIdent);
  if Assigned(vObject) then begin

    // if indexed then read index, else bitch
    if Indexed then begin
      if vObject.IndexableValue(PropertyIdent) then begin
        vObject.ItemIndex := aStack.Pop;

        // valid index?
        if ValidateIndex and
           (not vObject.ValidIndex) then
          SetResult(TYPE_E_AMBIGUOUSNAME);
      end else
        SetResult(DISP_E_NOTACOLLECTION);
    end;

    // what we came for
    if not AnyErrors then begin

      // valid property?
      if vObject.Valid(PropertyIdent) then
        aStack.Push(vObject.ValueOf(PropertyIdent))
      else
        SetResult(TYPE_E_ELEMENTNOTFOUND);
    end;

  // possible soft values?
  end else if not Indexed then begin

    // see if we can find it!
    vValue := vParser.LookupValue(ObjectNProperty);

    // what we came for
    if not AnyErrors then

      // push it then
      aStack.Push(vValue);

  // otherwise just house keep
  end else
    SetResult(TYPE_E_AMBIGUOUSNAME);
end;
class function TExpressionLateProperty.Symbol: string;
begin
  Result := 'Late ' + inherited Symbol;
end;
function TExpressionLateProperty.Caption: string;
begin
  Result := ObjectNProperty;
end;
function TExpressionLateProperty.ObjectNProperty: string;
begin
  Result := ObjectIdent;
  if PropertyIdent <> '' then
    Result := Result + '.' + PropertyIdent;
end;

end.
