unit FileInfoSupport;

//==============================================================================
// still to do
// GetFullPathName should be used to make sure we have the fully qualified path

//==============================================================================
interface

//==============================================================================
uses
  Windows, Messages, SysUtils, Classes, ListSupport, VersionInfoSupport;

//==============================================================================
type
  //============================================================================
  TFileInfoAttribute = (fiaReadOnly, fiaHidden, fiaSystem, fiaVolumeID,
                        fiaDirectory, fiaArchive, fiaCompressed, fiaNormal,
                        fiaTemporary, fiaAtomicWrite, fiaXActionWrite, fia11,
                        fia12, fia13, fia14, fia15,
                        fia16, fia17, fia18, fia19,
                        fia20, fia21, fia22, fia23,
                        fia24, fia25, fia26, fia27,
                        fia28, fia29, fia30, fia31);
  TFileInfoAttributes = set of TFileInfoAttribute;

  //============================================================================
  TFileInfo = class(TObject)
  private
    // file specific entries
    FValid: boolean;
    FRoot, FFolder, FName, FShortname, FShortFull: string;
    FSize, FCompressedSize: comp;
    FCreation, FLastAccess, FLastChange: TFileTime;
    FAttributes: TFileInfoAttributes;
    FParentOfPath, FFileDir, FUnc, FSizeAsText, FSizeAsEnglish,
    FCompressedSizeAsText, FCompressedSizeAsEnglish, FAttributesAsText,
    FAttributesAsEnglish, FFileType, FApplication: string;
    FVersionInfo: TVersionInfo;
  protected
    procedure Populate; virtual;
    procedure PopulateWith(const aInfo: TWin32FindData); virtual;
    function GetVersionValue(aIdent: string): string;
  public
    // construction
    constructor Create(aPath: string); virtual;
    constructor CreateWithPathName(const aPath, aName: string); virtual;
    constructor CreateWithRootFolderName(const aRoot, aFolder, aName: string); virtual;
    constructor CreateWithRootFolderInfo(const aRoot, aFolder: string;
      const aInfo: TWin32FindData); virtual;
    destructor Destroy; override;

    // system specific
    property Valid: boolean read FValid;

    // name and path entries
    property Root: string read FRoot;
    property Folder: string read FFolder;
    function Path: string;
    function ParentOfPath: string;
    function FileDir: string;
    property Name: string read FName;
    property Shortname: string read FShortname;
    function ShortFull: string;
    function Full: string;
    function UNC: string;
    function Drive: string;
    function Local: boolean;
    function Extension: string;
    function SansExtension: string;

    // date entries
    property Creation: TFileTime read FCreation;
    function CreationAsDateTime: TDateTime;
    property LastAccess: TFileTime read FLastAccess;
    function LastAccessAsDateTime: TDateTime;
    property LastChange: TFileTime read FLastChange;
    function LastChangeAsDateTime: TDateTime;

    // size entries
    property Size: comp read FSize;
    function SizeAsInteger: integer;
    function SizeAsDouble: double;
    function SizeAsText: string;
    function SizeAsEnglish: string;

    // compressed size
    function CompressedSize: comp;
    function CompressedSizeAsInteger: integer;
    function CompressedSizeAsDouble: double;
    function CompressedSizeAsText: string;
    function CompressedSizeAsEnglish: string;

    // attributes entries
    property Attributes: TFileInfoAttributes read FAttributes;
    function AttributesAsText: string;
    function AttributesAsEnglish: string;
    function ReadOnly: boolean;
    function Hidden: boolean;
    function System: boolean;
    function Directory: boolean;
    function Archive: boolean;
    function Compressed: boolean;

    // file type entries
    function FileType: string;
    function Application: string;

    // version info
    function VersionInfo: TVersionInfo;
    property VersionValue[aIdent: string]: string read GetVersionValue;
  end;

  //============================================================================
  TFileBrowser = class(TObject)
  private
    FFolder: string;
    FHandle: THandle;
    FData: TWin32FindData;
  protected
    procedure SetFolder(const aValue: string); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    //==========================================================================
    function Open: boolean; virtual;
    function Opened: boolean; virtual;
    function Next: boolean; virtual;
    procedure Close; virtual;

    //==========================================================================
    property Folder: string read FFolder write SetFolder;
    property Handle: THandle read FHandle;
    property Data: TWin32FindData read FData;
    function IsFolder: boolean;
    function IsHS: boolean;
    function IsHSD: boolean;
  end;

  //============================================================================
  TFileInfoListUpdate = procedure(Sender: TObject; const aPath: string) of object;
  TFileInfoList = class(TObject)
  private
    FList: TFreeList;
    FOnUpdate: TFileInfoListUpdate;
  protected
    property List: TFreeList read FList;
    function GetFile(aIndex: integer): TFileInfo; virtual;
    procedure DoUpdate(const aPath: string);
  public
    constructor Create; virtual;
    destructor Destroy; override;

    //==========================================================================
    function Count: integer; virtual;
    property Files[aIndex: integer]: TFileInfo read GetFile; default;
    function Add(aFileInfo: TFileInfo): integer; virtual;
    procedure Delete(aIndex: integer); virtual;
    procedure Clear; virtual;

    //==========================================================================
    property OnUpdate: TFileInfoListUpdate read FOnUpdate write FOnUpdate;
    function Populate(const aFrom: string; aRecursive: boolean): boolean; virtual;
    function PopulateWith(aFroms: TStrings; aRecursive: boolean): boolean; virtual;
  end;

//==============================================================================
var
  gMaxFileInfoCacheCount: integer = 20;

//==============================================================================
function FileInfoFor(const aPathname: string): TFileInfo;

//==============================================================================
implementation

//==============================================================================
uses
  MiscConstants, MiscSupport, ResultSupport, DateTimeSupport, StringSupport,
  ShellApi, XProcs, FileSupport;

//==============================================================================
// TFileInfo = class(TObject)
//==============================================================================
constructor TFileInfo.Create(aPath: string);
begin
  inherited Create;

  // find the file and then parse it
  aPath := FindPathForFile(aPath);
  FRoot := ExtractFilePath(aPath);
  FFolder := '';
  FName := ExtractFileName(aPath);

  // do it
  Populate;
end;
constructor TFileInfo.CreateWithPathName(const aPath, aName: string);
begin
  inherited Create;

  // get ready
  FRoot := aPath;
  FFolder := '';
  FName := aName;

  // do it
  Populate;
end;
constructor TFileInfo.CreateWithRootFolderName(const aRoot, aFolder, aName: string);
begin
  inherited Create;

  // get ready
  FRoot := aRoot;
  FFolder := aFolder;
  FName := aName;

  // do it
  Populate;
end;
constructor TFileInfo.CreateWithRootFolderInfo(const aRoot, aFolder: string;
  const aInfo: TWin32FindData);
begin
  inherited Create;

  // get ready
  FRoot := aRoot;
  FFolder := aFolder;

  // do it
  PopulateWith(aInfo);
end;
destructor TFileInfo.Destroy;
begin
  Dump(FVersionInfo);
  inherited Destroy;
end;
procedure TFileInfo.Populate;
var
  vHandle: THandle;
  vData: TWin32FindData;
begin
  // only if not valid
  if not Valid then begin

    // populate things
    vHandle := FindFirstFile(PChar(Full), vData);

    // while item is real
    if vHandle <> INVALID_HANDLE_VALUE then begin
      PopulateWith(vData);

      // finish things up
      Windows.FindClose(vHandle);
    end else
      FValid := False;
  end;
end;
procedure TFileInfo.PopulateWith(const aInfo: TWin32FindData);
begin
  // only if not valid
  if not Valid then with aInfo do begin
    FValid := True;

    // copy over the simple stuff
    FName := cFilename;
    FShortname := cAlternateFilename;

    FileTimeToLocalFileTime(ftCreationTime, FCreation);
    FileTimeToLocalFileTime(ftLastAccessTime, FLastAccess);
    FileTimeToLocalFileTime(ftLastWriteTime, FLastChange);

    FAttributes := TFileInfoAttributes(dwFileAttributes);

    FSize := IntsToComp(nFileSizeHigh, nFileSizeLow);

    // corrections
    if FShortname = '' then
      FShortname := AnsiUppercaseFilename(Name);
    FName := CorrectFilenameCase(Name);
  end;
end;

//==============================================================================
function TFileInfo.Path: string;
begin
  Result := Root + Folder;
end;
function TFileInfo.ParentOfPath: string;
var
  vLast: string;
begin
  if FParentOfPath = '' then begin
    FParentOfPath := Path;
    vLast := Copy(FParentOfPath, Length(FParentOfPath), 1);
    if vLast = ':' then
      FParentOfPath := ''
    else begin
      if vLast = '\' then
        FParentOfPath := Copy(FParentOfPath, 1, Length(FParentOfPath) - 1);
      FParentOfPath := ExtractFilePath(FParentOfPath);
    end;
  end;
  Result := FParentOfPath;
end;
function TFileInfo.FileDir: string;
begin
  if FFileDir = '' then
    FFileDir := ExtractFileDir(Full);
  Result := FFileDir;
end;
function TFileInfo.ShortFull: string;
begin
  if FShortFull = '' then
    FShortFull := ShortpathForPath(Full);
  Result := FShortFull;
end;
function TFileInfo.Full: string;
begin
  Result := Root + Folder + Name;
end;
function TFileInfo.UNC: string;
begin
  if FUnc = '' then
    FUnc := ExpandUNCFileName(Full);
  Result := FUnc;
end;
function TFileInfo.Drive: string;
begin
  Result := ExtractFileDrive(Path);
end;
function TFileInfo.Local: boolean;
begin
  Result := AnsiPos(':', Unc) = 0;
end;
function TFileInfo.Extension: string;
begin
  Result := ExtractFileExt(Name);
  if Copy(Result, 1, 1) = '.' then
    Result := Copy(Result, 2, Length(Result));
end;
function TFileInfo.SansExtension: string;
begin
  Result := ExtractFileExt(Name);
  if Result <> '' then
    Result := Copy(Name, 1, Length(Name) - Length(Result))
  else
    Result := Name;
end;

//==============================================================================
function TFileInfo.CreationAsDateTime: TDateTime;
begin
  if Valid then
    Result := FileTimeToDateTime(Creation)
  else
    Result := 0;
end;
function TFileInfo.LastAccessAsDateTime: TDateTime;
begin
  if Valid then
    Result := FileTimeToDateTime(LastAccess)
  else
    Result := 0;
end;
function TFileInfo.LastChangeAsDateTime: TDateTime;
begin
  if Valid then
    Result := FileTimeToDateTime(LastChange)
  else
    Result := 0;
end;

//==============================================================================
function TFileInfo.SizeAsInteger: integer;
var
  vHigh: integer;
begin
  if Valid then
    CompToInts(Size, vHigh, Result)
  else
    Result := 0;
end;
function TFileInfo.SizeAsDouble: double;
begin
  Result := Size;
end;
function TFileInfo.SizeAsText: string;
begin
  if Valid and
     (FSizeAsText = '') then begin

    if (fiaDirectory in Attributes) then
      FSizeAsText := ''
    else
      FSizeAsText := BytesToText(SizeAsInteger);
  end;
  Result := FSizeAsText;
end;
function TFileInfo.SizeAsEnglish: string;
begin
  if Valid and
     (FSizeAsEnglish = '') then begin

    if (fiaDirectory in Attributes) then
      FSizeAsEnglish := ''
    else
      FSizeAsEnglish := BytesToEnglish(SizeAsInteger);
  end;
  Result := FSizeAsEnglish;
end;

// compressed size
function TFileInfo.CompressedSize: comp;
var
  vHigh, vLow: integer;
begin
  if Valid and
     (FCompressedSize = 0) then begin

    // ask the OS
    vLow := GetCompressedFileSize(PChar(Full), @vHigh);
    FCompressedSize := IntsToComp(vHigh, vLow);
  end;
  Result := FCompressedSize;
end;
function TFileInfo.CompressedSizeAsInteger: integer;
var
  vHigh: integer;
begin
  if Valid then
    CompToInts(CompressedSize, vHigh, Result)
  else
    Result := 0;
end;
function TFileInfo.CompressedSizeAsDouble: double;
begin
  Result := CompressedSize;
end;
function TFileInfo.CompressedSizeAsText: string;
begin
  if Valid and
     (FCompressedSizeAsText = '') then begin

    if (fiaDirectory in Attributes) then
      FCompressedSizeAsText := ''
    else
      FCompressedSizeAsText := BytesToText(CompressedSizeAsInteger);
  end;
  Result := FCompressedSizeAsText;
end;
function TFileInfo.CompressedSizeAsEnglish: string;
begin
  if Valid and
     (FCompressedSizeAsEnglish = '') then begin

    if (fiaDirectory in Attributes) then
      FCompressedSizeAsEnglish := ''
    else
      FCompressedSizeAsEnglish := BytesToEnglish(CompressedSizeAsInteger);
  end;
  Result := FCompressedSizeAsEnglish;
end;

//==============================================================================
function TFileInfo.AttributesAsText: string;
  procedure AppendFlag(aTrue: boolean; aLetter: string);
  begin
    if aTrue then
      FAttributesAsText := FAttributesAsText + aLetter;
  end;
begin
  if Valid and
     (FAttributesAsText = '') then begin
    FAttributesAsText := '';
    AppendFlag(fiaTemporary in Attributes,  'T');
    AppendFlag(fiaNormal in Attributes,     'N');
    AppendFlag(fiaCompressed in Attributes, 'C');
    AppendFlag(fiaArchive in Attributes,    'A');
    AppendFlag(fiaDirectory in Attributes,  'D');
    AppendFlag(fiaVolumeID in Attributes,   'V');
    AppendFlag(fiaSystem in Attributes,     'S');
    AppendFlag(fiaHidden in Attributes,     'H');
    AppendFlag(fiaReadOnly in Attributes,   'R');
  end;
  Result := FAttributesAsText;
end;
function TFileInfo.AttributesAsEnglish: string;
  procedure AppendWord(aTrue: boolean; aWord: string);
  begin
    if aTrue then begin
      if FAttributesAsEnglish <> '' then
        FAttributesAsEnglish := FAttributesAsEnglish + ', ';
      FAttributesAsEnglish := FAttributesAsEnglish + aWord;
    end;
  end;
begin
  if Valid and
     (FAttributesAsEnglish = '') then begin
    FAttributesAsEnglish := '';
    AppendWord(fiaTemporary in Attributes,  'Temporary');
    AppendWord(fiaNormal in Attributes,     'Normal');
    AppendWord(fiaCompressed in Attributes, 'Compressed');
    AppendWord(fiaArchive in Attributes,    'Archived');
    AppendWord(fiaDirectory in Attributes,  'Directory');
    AppendWord(fiaVolumeID in Attributes,   'VolumeID');
    AppendWord(fiaSystem in Attributes,     'System');
    AppendWord(fiaHidden in Attributes,     'Hidden');
    AppendWord(fiaReadOnly in Attributes,   'Readonly');
    if FAttributesAsEnglish = '' then
      FAttributesAsEnglish := 'no flags';
  end;
  Result := FAttributesAsEnglish;
end;
function TFileInfo.ReadOnly: boolean;
begin
  Result := Valid and
            (fiaReadOnly in Attributes);
end;
function TFileInfo.Hidden: boolean;
begin
  Result := Valid and
            (fiaHidden in Attributes);
end;
function TFileInfo.System: boolean;
begin
  Result := Valid and
            (fiaSystem in Attributes);
end;
function TFileInfo.Directory: boolean;
begin
  Result := Valid and
            (fiaDirectory in Attributes);
end;
function TFileInfo.Archive: boolean;
begin
  Result := Valid and
            (fiaArchive in Attributes);
end;
function TFileInfo.Compressed: boolean;
begin
  Result := Valid and
            (fiaCompressed in Attributes);
end;

//==============================================================================
function TFileInfo.FileType: string;
begin
  if Valid and
     (FFileType = '') then
    FFileType := fileTypeName(Full);
  Result := FFileType;
end;
function TFileInfo.Application: string;
var
  vApp: array [0..512] of char;
  vResult: HINST;
begin
  if Valid and
     (FApplication = '') then begin
    vResult := FindExecutable(PChar(Name), PChar(Path), @vApp);
    if vResult >= 32 then
      FApplication := vApp
    else case vResult of
      ERROR_FILE_NOT_FOUND:
        FApplication := '<file was not found>';
      ERROR_PATH_NOT_FOUND:
        FApplication := '<path was not found>';
      ERROR_BAD_FORMAT:
        FApplication := '<file is invalid>';
    else
      FApplication := '<no association>';
    end;
  end;
  Result := FApplication;
end;

//==============================================================================
function TFileInfo.VersionInfo: TVersionInfo;
begin
  if Valid and
     (FVersionInfo = nil) then
    FVersionInfo := TVersionInfo.Create(Full);
  Result := FVersionInfo;
end;
function TFileInfo.GetVersionValue(aIdent: string): string;
begin
  Result := '';
  if VersionInfo <> nil then
    Result := VersionInfo[aIdent];
end;

//==============================================================================
// TFileBrowser = class(TObject)
//==============================================================================
constructor TFileBrowser.Create;
begin
  inherited Create;
  FHandle := INVALID_HANDLE_VALUE;
end;
destructor TFileBrowser.Destroy;
begin
  Close;
  inherited Destroy;
end;

//==========================================================================
procedure TFileBrowser.SetFolder(const aValue: string);
begin
  if (aValue <> Folder) and
     (Copy(aValue, Length(aValue), 1) = '\') then begin
    Close;
    FFolder := aValue;
  end;
end;
function TFileBrowser.Open: boolean;
begin
  if not Opened then begin
    FHandle := Windows.FindFirstFile(PChar(Folder + '*.*'), FData);
    if FHandle = INVALID_HANDLE_VALUE then
      SetToWindowsError;
  end;
  Result := Opened;
end;
function TFileBrowser.Opened: boolean;
begin
  Result := Handle <> INVALID_HANDLE_VALUE;
end;
function TFileBrowser.Next: boolean;
begin
  if not Opened then
    Result := Open
  else begin
    Result := Windows.FindNextFile(Handle, FData);
    if not Result and
       (SetToWindowsError = ERROR_NO_MORE_FILES) then begin
      ClearResult;
      Close;
    end;
  end;
end;
procedure TFileBrowser.Close;
begin
  if Opened then begin
    if Windows.FindClose(FHandle) then
      SetToWindowsError;
    FHandle := INVALID_HANDLE_VALUE;
  end;
end;
function TFileBrowser.IsFolder: boolean;
begin
  Result := Opened and
            (fiaDirectory in TFileInfoAttributes(Data.dwFileAttributes));
end;
function TFileBrowser.IsHS: boolean;
begin
  Result := Opened and
            (fiaHidden in TFileInfoAttributes(Data.dwFileAttributes)) and
            (fiaSystem in TFileInfoAttributes(Data.dwFileAttributes));
end;
function TFileBrowser.IsHSD: boolean;
begin
  Result := Opened and
            (fiaHidden in TFileInfoAttributes(Data.dwFileAttributes)) and
            (fiaSystem in TFileInfoAttributes(Data.dwFileAttributes)) and
            (fiaDirectory in TFileInfoAttributes(Data.dwFileAttributes));
end;

//==============================================================================
// TFileInfoList = class(TObject)
//==============================================================================
constructor TFileInfoList.Create;
begin
  inherited Create;
  FList := TFreeList.Create;
end;
destructor TFileInfoList.Destroy;
begin
  Dump(FList);
  inherited Destroy;
end;

//==============================================================================
procedure TFileInfoList.DoUpdate(const aPath: string);
begin
  if Assigned(FOnUpdate) then
    FOnUpdate(Self, aPath);
end;
function TFileInfoList.GetFile(aIndex: integer): TFileInfo;
begin
  Result := TFileInfo(List[aIndex]);
end;
function TFileInfoList.Count: integer;
begin
  Result := List.Count;
end;
function TFileInfoList.Add(aFileInfo: TFileInfo): integer;
begin
  Result := List.Add(aFileInfo);
end;
procedure TFileInfoList.Delete(aIndex: integer);
begin
  List.Delete(aIndex);
end;
procedure TFileInfoList.Clear;
begin
  List.Clear;
end;

//==============================================================================
function TFileInfoList.Populate(const aFrom: string; aRecursive: boolean): boolean;
var
  vBrowser: TFileBrowser;
  vFolder: string;
  function FolderPopulate(const aFolder: string): boolean;
  var
    vFolders: TStringList;
    vIndex: integer;
    vName: string;
  begin
    // set things up
    Result := False;
    vFolders := TStringList.Create;

    // prep the browser
    with vBrowser do try
      Folder := aFrom + aFolder;
      vFolder := aFolder;
      DoUpdate(Folder);

      // loopy
      while Next do begin
        vName := Data.cFilename;

        // skip the special directories
        if not ((vName = '.') or
                (vName = '..')) then begin

          // remember this one
          Add(TFileInfo.CreateWithRootFolderInfo(aFrom, vFolder, Data));

          // do this directory as well?
          if aRecursive and
             IsFolder then begin
            vFolders.Add(CorrectFilenameCase(vName));
          end;
        end;
      end;
    except
      Result := True;
    end;

    // do the directories?
    if not Result then
      for vIndex := 0 to vFolders.Count-1 do
        if FolderPopulate(aFolder + vFolders[vIndex] + '\') then begin
          Result := True;
          break;
        end;
  end;
begin
  // assume no error
  Result := False;

  // create our slave
  vBrowser := TFileBrowser.Create;
  if Assigned(vBrowser) then try

    // do it dammit
    Result := FolderPopulate('');

  // clean up
  finally
    Dump(vBrowser);
  end;
end;
function TFileInfoList.PopulateWith(aFroms: TStrings; aRecursive: boolean): boolean;
var
  vIndex: integer;
begin
  // assume no error
  Result := False;

  // loop it dude
  for vIndex := 0 to aFroms.Count-1 do begin
    Result := Populate(aFroms[vIndex], aRecursive);
    if Result then
      break;
  end;
end;

//==============================================================================
var
  vFileInfoCache: TFreeStringList = nil;
  vUnitOk: boolean = True;

//==============================================================================
function FileInfoFor(const aPathname: string): TFileInfo;
var
  vIndex: integer;
  vObject: TObject;
begin
  // make room
  if vUnitOk and
     not Assigned(vFileInfoCache) then
    vFileInfoCache := TFreeStringList.Create;

  // see if its defined already
  vIndex := vFileInfoCache.IndexOf(aPathname);

  // if not found create it
  if vIndex = cIndexNotFound then begin

    // delete the oldest fileinfo
    if vFileInfoCache.Count = gMaxFileInfoCacheCount then
      vFileInfoCache.Delete(vFileInfoCache.Count-1);

    // create the file info then
    vIndex := vFileInfoCache.AddObject(aPathname, TFileInfo.Create(aPathname));
  end;

  // keep going?
  if vIndex <> cIndexNotFound then begin

    // move things around a bit
    if vIndex <> 0 then begin

      // copy the reference info and delete it
      vObject := vFileInfoCache.Objects[vIndex];
      vFileInfoCache.Delete(vIndex);

      // put it back and change the index a bit
      vFileInfoCache.InsertObject(0, aPathname, vObject);
      vIndex := 0;
    end;

    // well return it
    Result := TFileInfo(vFileInfoCache.Objects[vIndex]);
  end else
    Result := nil;
end;

//==============================================================================
initialization
  // nothing here
finalization
  vUnitOk := False;
  Dump(vFileInfoCache);
end.
