unit ValueSupport;

//==============================================================================
interface

//==============================================================================
uses
  Windows, Messages, SysUtils, Classes, ListSupport, ExpressionParser,Variants;

//==============================================================================
type
  //============================================================================
  TValueObject = class(TPersistent)
  private
    FSymbol: string;
    FContent: variant;
    FDeferred: boolean;
  protected
    function GetContent: variant; virtual;
  public
    constructor Create(const aSymbol: string; const aContent: variant; aDeferred: boolean);
    property Symbol: string read FSymbol;
    property Content: variant read GetContent write FContent;
    property ActualContent: variant read FContent;
    property Deferred: boolean read FDeferred;
  end;

  //============================================================================
  TValueList = class(TObjectList)
  private
    FDefaultContent: variant;
  protected
    function GetValue(aIndex: integer): TValueObject;
    function GetContent(aSymbol: string): variant;
  public
    constructor Create;
    property DefaultContent: variant read FDefaultContent write FDefaultContent;

    function Add(const aSymbol: string; const aContent: variant;
      aDeferred: boolean): integer;
    property Values[aIndex: integer]: TValueObject read GetValue;
    property Content[aSymbol: string]: variant read GetContent; default;
    function LookupIndex(const aSymbol: string): integer;
    function Exists(const aSymbol: string): boolean;
    function Remove(const aSymbol: string): integer;
  end;

//==============================================================================
var
  DeferredEvaluator: TCustomExpressionParser = nil;
  RecursionErrorMsg: string = '<recursion>';

//==============================================================================
implementation

//==============================================================================
uses
  MiscSupport, MiscConstants, ResultSupport;

//============================================================================
// TValueObject = class(TPersistent)
//============================================================================
constructor TValueObject.Create(const aSymbol: string; const aContent: variant;
  aDeferred: boolean);
begin
  inherited Create;

  // good symbol?
  Assert(IsValidIdent(aSymbol), ResultTextFor(TYPE_E_AMBIGUOUSNAME));
  FSymbol := aSymbol;

  // can we actually do deferred is needed?
  FDeferred := aDeferred;
  Assert(not Deferred or
         Assigned(DeferredEvaluator), ResultTextFor(STG_E_INVALIDHANDLE));

  // store the contents correctly
  if Deferred then
    Content := VarAsType(aContent, varString)
  else
    FContent := aContent;

end;

//==============================================================================
const
  cMaximumRecursion = 50;

//==============================================================================
var
  vRecursionLevel: integer = 0;

//==============================================================================
function TValueObject.GetContent: variant;
begin
  // watch out for recursion
  Inc(vRecursionLevel);

  // but wait do we evaluate?
  if Deferred then begin

    // uncomment this if you dare!
    //Assert(vRecursionLevel < cMaximumRecursion, ResultTextFor(STG_E_INSUFFICIENTMEMORY));
    if (vRecursionLevel < cMaximumRecursion) then
      Result := DeferredEvaluator.Evaluate(FContent)
    else
      Result := RecursionErrorMsg;

  // plain normal content?
  end else
    Result := FContent;

  // make room for more recursion
  Dec(vRecursionLevel);
end;

//============================================================================
// TValueList = class(TObjectList)
//==============================================================================
constructor TValueList.Create;
begin
  inherited Create;
  DefaultContent := Unassigned;
end;
function TValueList.GetValue(aIndex: integer): TValueObject;
begin
  Result := TValueObject(List[aIndex]);
end;
function TValueList.GetContent(aSymbol: string): variant;
var
  vIndex: integer;
begin
  Result := Unassigned;
  vIndex := LookupIndex(aSymbol);
  if vIndex <> cIndexNotFound then
    Result := Values[vIndex].Content
  else if not VarIsEmpty(DefaultContent) then
    Result := DefaultContent;
end;
function TValueList.Add(const aSymbol: string; const aContent: variant;
  aDeferred: boolean): integer;
begin
  Remove(aSymbol);
  Result := inherited Add(TValueObject.Create(aSymbol, aContent, aDeferred));
end;
function TValueList.LookupIndex(const aSymbol: string): integer;
var
  vIndex: integer;
begin
  Result := cIndexNotFound;
  for vIndex := 0 to Count-1 do
    if AnsiCompareText(aSymbol, Values[vIndex].Symbol) = 0 then begin
      Result := vIndex;
      break;
    end;
end;
function TValueList.Exists(const aSymbol: string): boolean;
begin
  Result := (LookupIndex(aSymbol) <> cIndexNotFound) or
            not VarIsEmpty(DefaultContent);
end;
function TValueList.Remove(const aSymbol: string): integer;
begin
  Result := LookupIndex(aSymbol);
  if Result <> cIndexNotFound then
    Delete(Result);
end;

end.
