////////////////////////////////////////////////////////////////////////////////
// File:       PegtopAssemblers.pas
// Classes:    TPegtopAsmCode, TPegtopMacroAsmCode
// Version:    1.01
// Date:       04 Sep 2003 created 1.00
//             28 Feb 2005 modified 1.01 dyn. exception mask (FPU control word)
// Author:     Jens Gruschel (GRU)
// Copyright:  (c) 2003 Jens Gruschel
// Website:    http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////
// Description:
// TPegtopAsmCode extends TPegtopCompilerCode with assembler instructions
// (only the ones that are relevant for TPegtopFormulaCompiler).
// TPegtopMacroAsmCode extends TPegtopAsmCode with macros and procedure calls.
////////////////////////////////////////////////////////////////////////////////
// License:
// Please read the license.txt file that comes with this software
// and visit http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////
// Documentation / Installation:
// Please read the readme.txt file that comes with this software
// and visit http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////

unit PegtopAssemblers;

interface

uses
  PegtopCompilers;

type
  TPegtopProcReg = (regAX, regBX, regCX, regDX, regSP, regBP, regSI, regDI,
                    regEAX, regEBX, regECX, regEDX, regESP, regEBP, regESI, regEDI);
  TPegtopProcRegType = (prt16bit, prt32bit);
  TPegtopMathReg = (st0, st1, st2, st3, st4, st5, st6, st7);
  TPegtopMathRounding = (pmrDefault, pmrNearest, pmrDown, pmrUp, pmrToZero);
  TPegtopMathError = (pmeInvalidOp, pmeDenormal, pmeDivByZero, pmeOverflow, pmeUnderflow, pmePrecision);
  TPegtopMathErrors = set of TPegtopMathError;
  TPegtopMathPrecision = (pmp24bit, pmp53bit, pmp64bit);
  TPegtopMathInfinity = (pmiProjective, pmiAffine);

  EPegtopAsmInvalidOperand = class(EPegtopCompiler)
  public
    constructor CreateProp(const OpCode, Found, Expected: String);
  end;

  TPegtopAsmCode = class(TPegtopCompilerCode)
  private
    FFPUControlWordAddr: TPegtopVirtualAddr;
  public
    function GetFPUControlWord(Errors: TPegtopMathErrors;
      Precision: TPegtopMathPrecision; Infinity: TPegtopMathInfinity): Word;
    procedure AddFPUControlWords(Errors: TPegtopMathErrors;
      Precision: TPegtopMathPrecision; Infinity: TPegtopMathInfinity;
      DefaultRounding: TPegtopMathRounding = pmrNearest);
    procedure SetFPUControlWords(Errors: TPegtopMathErrors;
      Precision: TPegtopMathPrecision; Infinity: TPegtopMathInfinity;
      DefaultRounding: TPegtopMathRounding = pmrNearest);
    // 8086 instructions:
    procedure CodeMOV(Reg1, Reg2: TPegtopProcReg); overload;                    // mov reg, value
    procedure CodeMOV(Reg: TPegtopProcReg; V: TPegtopVirtualAddr); overload;    // mov reg, value
    procedure CodeAND(Reg1, Reg2: TPegtopProcReg); overload;                    // and reg, value
    procedure CodeAND(Reg: TPegtopProcReg; V: TPegtopVirtualAddr); overload;    // and reg, value
    procedure CodeOR(Reg1, Reg2: TPegtopProcReg); overload;                     // or reg, value
    procedure CodeOR(Reg: TPegtopProcReg; V: TPegtopVirtualAddr); overload;     // or reg, value
    procedure CodeCALL(Reg: TPegtopProcReg); overload;                          // call reg
    procedure CodeRET;                                                          // return
    procedure CodeWAIT;                                                         // wait
    procedure CodePUSH(Reg: TPegtopProcReg); overload;                          // push reg
    procedure CodePUSH(Adr: TPegtopVirtualAddr); overload;                      // push Data[adr]
    procedure CodePOP(Reg: TPegtopProcReg);                                     // pop reg
    procedure CodeJMP(Lab: String);                                             // jmp Lab
    procedure CodeJE(Lab: String);                                              // if x = y then jump Lab
    procedure CodeJNE(Lab: String);                                             // if x <> y then jump Lab
    procedure CodeJB(Lab: String);                                              // if x < y then jump Lab (for float or unsigned int)
    procedure CodeJBE(Lab: String);                                             // if x <= y then jump Lab (for float or unsigned int)
    procedure CodeJA(Lab: String);                                              // if x > y then jump Lab (for float or unsigned int)
    procedure CodeJAE(Lab: String);                                             // if x >= y then jump Lab (for float or unsigned int)
    procedure CodeJL(Lab: String);                                              // if x < y then jump Lab (for signed int)
    procedure CodeJLE(Lab: String);                                             // if x <= y then jump Lab (for signed int)
    procedure CodeJG(Lab: String);                                              // if x > y then jump Lab (for signed int)
    procedure CodeJGE(Lab: String);                                             // if x >= y then jump Lab (for signed int)
    procedure CodeSAHF;                                                         // Flags = ah
    // 8087 instructions:
    procedure CodeFLD1;                                                         // st(0) = 1.0
    procedure CodeFLDZ;                                                         // st(0) = 0.0
    procedure CodeFLDL2E;                                                       // st(0) = log2(e)
    procedure CodeFLD(Adr: TPegtopVirtualAddr); overload;                       // st(0) = Data[adr]
    procedure CodeFLD(Reg: TPegtopMathReg); overload;                           // st(0) = st(reg)
    procedure CodeFST(Adr: TPegtopVirtualAddr); overload;                       // Data[adr] = st(0)
    procedure CodeFST(Reg: TPegtopMathReg); overload;                           // st(reg) = st(0)
    procedure CodeFSTP(Adr: TPegtopVirtualAddr); overload;                      // Data[adr] = st(0), pop st()
    procedure CodeFSTP(Reg: TPegtopMathReg); overload;                          // st(reg) = st(0), pop st()
    procedure CodeFADD(Adr: TPegtopVirtualAddr); overload;                      // st(0) = st(0) + Data[adr]
    procedure CodeFADD(Reg1, Reg2: TPegtopMathReg); overload;                   // st(reg1) = st(reg1) + st(reg2)
    procedure CodeFADDP(Reg1, Reg2: TPegtopMathReg);                            // st(reg1) = st(reg1) + st(reg2), pop st()
    procedure CodeFSUB(Adr: TPegtopVirtualAddr); overload;                      // st(0) = st(0) - Data[adr]
    procedure CodeFSUB(Reg1, Reg2: TPegtopMathReg); overload;                   // st(reg1) = st(reg1) - st(reg2)
    procedure CodeFSUBR(Adr: TPegtopVirtualAddr); overload;                     // st(0) = Data[adr] - st(0)
    procedure CodeFSUBR(Reg1, Reg2: TPegtopMathReg); overload;                  // st(reg1) = st(reg2) - st(reg1)
    procedure CodeFMUL(Adr: TPegtopVirtualAddr); overload;                      // st(0) = st(0) * Data[adr]
    procedure CodeFMUL(Reg: TPegtopMathReg); overload;                          // st(0) = st(0) * st(reg)
    procedure CodeFDIV(Adr: TPegtopVirtualAddr);                                // st(0) = st(0) / Data[adr]
    procedure CodeFDIVR(Adr: TPegtopVirtualAddr);                               // st(0) = Data[adr] / st(0)
    procedure CodeFXCHG(Reg: TPegtopMathReg);                                   // swap st(0) with st(Reg)
    procedure CodeFFREE(Reg: TPegtopMathReg);                                   // free st(reg)
    procedure CodeFCHS;                                                         // st(0) = -st(0)
    procedure CodeFABS;                                                         // st(0) = abs(st(0))
    procedure CodeFSQRT;                                                        // st(0) = sqrt(st(0))
    procedure CodeFSTCW(Adr: TPegtopVirtualAddr);                               // Data[adr] = cw
    procedure CodeFLDCW(Adr: TPegtopVirtualAddr);                               // cw = Data[adr]
    procedure CodeFSIN;                                                         // st(0) = sin(st(0))
    procedure CodeFCOS;                                                         // st(0) = cos(st(0))
    procedure CodeFPTAN;                                                        // st(1)/st(0) = tan(st(0))
    procedure CodeFPATAN;                                                       // st(0) = arctan(st(1)/st(0))
    procedure CodeFYL2X;                                                        // st(0) = st(1) * log2(st(0))
    procedure CodeF2XM1;                                                        // st(0) = 2 ^ x - 1 (for -1.0 < x < 1.0)
    procedure CodeFNCLEX;                                                       // clear exceptions
    procedure CodeFSCALE;                                                       // st(0) = st(0) * 2 ^ int(st(1))
    procedure CodeFRNDINT;                                                      // st(0) = rndint(st(0))
    procedure CodeFPREM;                                                        // st(0) = prem(st(0))
    procedure CodeFPREM1;                                                       // st(0) = prem(st(0)) (IEEE)
    procedure CodeFXTRACT;                                                      // st(1) = exponent(st(0)), st(0) = significant(st(0))
    procedure CodeFTST;                                                         // st(0) <=> 0?
    procedure CodeFSTSW(Reg: TPegtopProcReg);                                   // reg = sw
    procedure CodeFLDLN2;                                                       // st(0) = ln(2)
    procedure CodeFLDLG2;                                                       // st(0) = lg(2)
    procedure CodeFCOM(Reg: TPegtopMathReg);                                    // st(0) <=> st(reg)?
    procedure CodeFCOMP(Reg: TPegtopMathReg);                                   // st(0) <=> st(reg)?, pop st()
    procedure CodeFCOMPP;                                                       // st(0) <=> st(1)?, 2x pop st()
    property FPUControlWordAddr: TPegtopVirtualAddr read FFPUControlWordAddr;
  end;

  TPegtopMacroAsmCode = class(TPegtopAsmCode)
  public
    procedure CodeCALL(Adr: Pointer); overload;                                 // call dword ptr adr
    procedure CodeFMOD(Adr: TPegtopVirtualAddr);                                // st(0) = st(0) mod Data[a]
    procedure CodeFMODR(Adr: TPegtopVirtualAddr);                               // st(0) = Data[adr] mod st(0)
    procedure CodeFSQR;                                                         // st(0) = st(0) * st(0)
    procedure CodeFEXP;                                                         // st(0) = exp(st(0))
    procedure CodeFPOWPOS;                                                      // st(0) = st(0) ^ st(1) with st(0) > 0
    procedure CodeFPOW;                                                         // st(0) = st(0) ^ st(1)
    procedure CodeFPOWR(Adr: TPegtopVirtualAddr);                               // st(0) = Data[adr] ^ st(0)
    procedure CodeFROUND(Rounding: TPegtopMathRounding);                        // st(0) = round(st(0))
    procedure CodeFFRAC(Rounding: TPegtopMathRounding);                         // st(0) = frac(st(0))
    procedure CodeFLN;                                                          // st(0) = ln(st(0))
    procedure CodeFLG;                                                          // st(0) = log10(st(0))
    procedure CodeFL2;                                                          // st(0) = log2(st(0))
    procedure CodeFTAN;                                                         // st(0) = tan(st(0))
    procedure CodeFATAN;                                                        // st(0) = arctan(st(0))
    procedure CodeFTSTSYNC(Reg: TPegtopMathReg);                                // st(Reg) <=> 0? (use jx afterwards to branch)
    procedure CodeFCOMSYNC(Reg1, Reg2: TPegtopMathReg);                         // st(Reg1) <=> st(Reg2)? (use jx afterwards to branch)
    procedure CodeFCOMPSYNC(Reg1, Reg2: TPegtopMathReg);                        // st(Reg1) <=> st(Reg2)? (use jx afterwards to branch), pop st()
    procedure CodeFSIGN;                                                        // st(0) = sign(st(0))
    procedure CodeFDIV2;                                                        // st(0) = st(0) / 2
    procedure CodeFMOD2;                                                        // st(0) = st(0) mod 2
  end;

implementation

uses
  SysUtils;

const
  // 80(2)86 registers:
  ProcRegName: array[TPegtopProcReg] of String = (
    'ax', 'bx', 'cx', 'dx', 'sp', 'bp', 'si', 'di',
    'eax', 'ebx', 'ecx', 'edx', 'esp', 'ebp', 'esi', 'edi');
  ProcRegMask: array[TPegtopProcReg] of Word = (
    0, 3, 1, 2, 4, 5, 6, 7,
    0, 3, 1, 2, 4, 5, 6, 7);
  ProcRegType: array[TPegtopProcReg] of TPegtopProcRegType = (
    prt16bit, prt16bit, prt16bit, prt16bit, prt16bit, prt16bit, prt16bit, prt16bit,
    prt32bit, prt32bit, prt32bit, prt32bit, prt32bit, prt32bit, prt32bit, prt32bit);

  // 8087 registers:
  MathRegName: array[TPegtopMathReg] of String = ('st(0)', 'st(1)', 'st(2)', 'st(3)', 'st(4)', 'st(5)', 'si', 'st(6)');
  MathRegMask: array[TPegtopMathReg] of Word = (0, 1, 2, 3, 4, 5, 6, 7);

  // 8087 control word rounding masks:
  RoundingMasks: array[TPegtopMathRounding] of Word = ($0000, $0000, $0400, $0800, $0C00);

var  
  // procedures (are created in initialization section):
  ProcFPOWPOS: TPegtopMacroAsmCode;
  ProcFPOW:    TPegtopMacroAsmCode;

type
  PInt64 = ^Int64;

////////////////////////////////////////////////////////////////////////////////
// EPegtopAsmInvalidArg
////////////////////////////////////////////////////////////////////////////////

constructor EPegtopAsmInvalidOperand.CreateProp(const OpCode, Found, Expected: String);
begin
  inherited CreateFmt('Invalid combination of opcode %s and operands: %s expected but %s found.', [OpCode, Expected, Found]);
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopAsmCode (8086)
////////////////////////////////////////////////////////////////////////////////

function TPegtopAsmCode.GetFPUControlWord(Errors: TPegtopMathErrors;
  Precision: TPegtopMathPrecision; Infinity: TPegtopMathInfinity): Word;
const
  ErrorMasks: array[TPegtopMathError] of Word = ($0001, $0002, $0004, $0008, $0010, $0020);
  PrecisionMasks: array[TPegtopMathPrecision] of Word = ($0000, $0200, $0300);
  InfinityMasks: array[TPegtopMathInfinity] of Word = ($0000, $1000);
var
  E: TPegtopMathError;
begin
  Result := $0040 or PrecisionMasks[Precision] or InfinityMasks[Infinity];
  for E := Low(TPegtopMathError) to High(TPegtopMathError) do
    if not (E in Errors) then Result := Result or ErrorMasks[E];
end;

procedure TPegtopAsmCode.AddFPUControlWords(Errors: TPegtopMathErrors;
  Precision: TPegtopMathPrecision; Infinity: TPegtopMathInfinity;
  DefaultRounding: TPegtopMathRounding = pmrNearest);
var
  W: Word;
  R: TPegtopMathRounding;
begin
  W := GetFPUControlWord(Errors, Precision, Infinity);
  FFPUControlWordAddr := Data.AddWord(W or RoundingMasks[DefaultRounding]);
  for R := Succ(Low(TPegtopMathRounding)) to High(TPegtopMathRounding) do
    Data.AddWord(W or RoundingMasks[R]);
end;

procedure TPegtopAsmCode.SetFPUControlWords(Errors: TPegtopMathErrors;
  Precision: TPegtopMathPrecision; Infinity: TPegtopMathInfinity;
  DefaultRounding: TPegtopMathRounding = pmrNearest);
var
  W: Word;
  R: TPegtopMathRounding;
  P: ^Word;
begin
  W := GetFPUControlWord(Errors, Precision, Infinity);
  P := Data.Ptr[FFPUControlWordAddr];
  P^ := W or RoundingMasks[DefaultRounding];
  Inc(P);
  for R := Succ(Low(TPegtopMathRounding)) to High(TPegtopMathRounding) do begin
    P^ := W or RoundingMasks[R];
    Inc(P);
  end;
end;

procedure TPegtopAsmCode.CodeMOV(Reg1, Reg2: TPegtopProcReg);
begin
  if ProcRegType[Reg1] = ProcRegType[Reg2] then begin
    AddLine('mov ' + ProcRegName[Reg1] + ', ' + ProcRegName[Reg2]);
    if ProcRegType[Reg1] = prt16bit then AddCode($66);
    AddCode($89);
    AddCode($C0 or (ProcRegMask[Reg2] shl 3) or ProcRegMask[Reg1]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('mov', ProcRegName[Reg2], 'another register type');
  end;
end;

//procedure TPegtopAsmCode.CodeMOV(Reg: TPegtopProcReg; V: TPegtopVirtualAddr);
//begin
//  if ProcRegType[Reg] = prt16bit then begin
//    AddLine('mov ' + ProcRegName[Reg] + ', ' + IntToHex(V, 4));
//    AddCode($66);
//    AddCode($8B or ProcRegMask[Reg]);
//    AddWord(V);
//  end
//  else begin
//    AddLine('mov ' + ProcRegName[Reg] + ', ' + IntToHex(V, 8));
//    AddCode($8B or ProcRegMask[Reg]);
//    AddLongInt(V);
//  end;
//end;

//Oups. It was a simple typo. I wrote $8B instead of $B8. Replace 
//TPegtopAsmCode.CodeMOV with following code:

procedure TPegtopAsmCode.CodeMOV(Reg: TPegtopProcReg; V: TPegtopVirtualAddr);
begin
   if ProcRegType[Reg] = prt16bit then begin
     AddLine('mov ' + ProcRegName[Reg] + ', ' + IntToHex(V, 4));
     AddCode($66);
     AddCode($B8 or ProcRegMask[Reg]);
     AddWord(V);
   end
   else begin
     AddLine('mov ' + ProcRegName[Reg] + ', ' + IntToHex(V, 8));
     AddCode($B8 or ProcRegMask[Reg]);
     AddLongInt(V);
   end;
end;

procedure TPegtopAsmCode.CodeAND(Reg1, Reg2: TPegtopProcReg);
begin
  if ProcRegType[Reg1] = ProcRegType[Reg2] then begin
    AddLine('and ' + ProcRegName[Reg1] + ', ' + ProcRegName[Reg2]);
    if ProcRegType[Reg1] = prt16bit then AddCode($66);
    AddCode($21);
    AddCode($C0 or (ProcRegMask[Reg2] shl 3) or ProcRegMask[Reg1]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('and', ProcRegName[Reg2], 'another register type');
  end;
end;

procedure TPegtopAsmCode.CodeAND(Reg: TPegtopProcReg; V: TPegtopVirtualAddr);
begin
  if ProcRegType[Reg] = prt16bit then begin
    AddLine('and ' + ProcRegName[Reg] + ', ' + IntToHex(V, 4));
    AddCode($66);
    AddCode($23 or ProcRegMask[Reg]);
    AddWord(V);
  end
  else begin
    AddLine('and ' + ProcRegName[Reg] + ', ' + IntToHex(V, 8));
    AddCode($23 or ProcRegMask[Reg]);
    AddLongInt(V);
  end;
end;

procedure TPegtopAsmCode.CodeOR(Reg1, Reg2: TPegtopProcReg);
begin
  if ProcRegType[Reg1] = ProcRegType[Reg2] then begin
    AddLine('or ' + ProcRegName[Reg1] + ', ' + ProcRegName[Reg2]);
    if ProcRegType[Reg1] = prt16bit then AddCode($66);
    AddCode($89);
    AddCode($09 or (ProcRegMask[Reg2] shl 3) or ProcRegMask[Reg1]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('or', ProcRegName[Reg2], 'another register type');
  end;
end;

procedure TPegtopAsmCode.CodeOR(Reg: TPegtopProcReg; V: TPegtopVirtualAddr);
begin
  if ProcRegType[Reg] = prt16bit then begin
    AddLine('or ' + ProcRegName[Reg] + ', ' + IntToHex(V, 4));
    AddCode($66);
    AddCode($0B or ProcRegMask[Reg]);
    AddWord(V);
  end
  else begin
    AddLine('or ' + ProcRegName[Reg] + ', ' + IntToHex(V, 8));
    AddCode($0B or ProcRegMask[Reg]);
    AddLongInt(V);
  end;
end;

procedure TPegtopAsmCode.CodeCALL(Reg: TPegtopProcReg);
begin
  if ProcRegType[Reg] = prt32bit then begin
    AddLine('call ' + ProcRegName[Reg]);
    AddCode($FF);
    AddCode($D0 or ProcRegMask[Reg]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('call', ProcRegName[Reg], '32 bit regster');
  end;
end;

procedure TPegtopAsmCode.CodeRET;
begin
  AddLine('ret');
  AddCode($C3);
end;

procedure TPegtopAsmCode.CodeWAIT;
begin
  AddLine('wait');
  AddCode($9B);
end;

procedure TPegtopAsmCode.CodePUSH(Reg: TPegtopProcReg);
begin
  AddLine('push ' + ProcRegName[Reg]);
  if ProcRegType[Reg] = prt16bit then AddCode($66);
  AddCode($50 or ProcRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodePUSH(Adr: TPegtopVirtualAddr);
begin
  AddLine('push dword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($FF);               // push dword ptr [esi+adr]
    AddCode($76);
    AddCode(Adr);
  end
  else begin
    AddCode($FF);               // push dword ptr [esi+adr]
    AddCode($B6);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodePOP(Reg: TPegtopProcReg);
begin
  AddLine('pop ' + ProcRegName[Reg]);
  if ProcRegType[Reg] = prt16bit then AddCode($66);
  AddCode($58 or ProcRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeJMP(Lab: String);
begin
  AddLine('jmp ' + Lab);
  AddCode($EB);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJE(Lab: String);
begin
  AddLine('je ' + Lab);
  AddCode($74);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJNE(Lab: String);
begin
  AddLine('jne ' + Lab);
  AddCode($75);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJB(Lab: String);
begin
  AddLine('jb ' + Lab);
  AddCode($72);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJBE(Lab: String);
begin
  AddLine('jbe ' + Lab);
  AddCode($76);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJA(Lab: String);
begin
  AddLine('ja ' + Lab);
  AddCode($77);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJAE(Lab: String);
begin
  AddLine('jae ' + Lab);
  AddCode($73);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJL(Lab: String);
begin
  AddLine('jl ' + Lab);
  AddCode($7C);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJLE(Lab: String);
begin
  AddLine('jle ' + Lab);
  AddCode($7E);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJG(Lab: String);
begin
  AddLine('jg ' + Lab);
  AddCode($7F);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeJGE(Lab: String);
begin
  AddLine('jge ' + Lab);
  AddCode($7D);
  AddJump(Lab);
end;

procedure TPegtopAsmCode.CodeSAHF;
begin
  AddLine('sahf');
  AddCode($9E);                 // sahf
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopAsmCode (8087)
////////////////////////////////////////////////////////////////////////////////

procedure TPegtopAsmCode.CodeFLD1;
begin
  AddLine('fld1');
  AddCode($D9);
  AddCode($E8);
end;

procedure TPegtopAsmCode.CodeFLDZ;
begin
  AddLine('fldz');
  AddCode($D9);
  AddCode($EE);
end;

procedure TPegtopAsmCode.CodeFLDL2E;
begin
  AddLine('fldl2e');
  AddCode($D9);
  AddCode($EA);
end;

procedure TPegtopAsmCode.CodeFLD(Adr: TPegtopVirtualAddr);
begin
  AddLine('fld qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DD);               // fld qword ptr [esi+adr]
    AddCode($46);
    AddCode(Adr);
  end
  else begin
    AddCode($DD);               // fld qword ptr [esi+adr]
    AddCode($86);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFLD(Reg: TPegtopMathReg);
begin
  AddLine('fld ' + MathRegName[Reg]);
  AddCode($D9);                 // fld st(Reg)
  AddCode($C0 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFST(Adr: TPegtopVirtualAddr);
begin
  AddLine('fst qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DD);               // fst qword ptr [esi+adr]
    AddCode($56);
    AddCode(Adr);
  end
  else begin
    AddCode($DD);               // fst qword ptr [esi+adr]
    AddCode($96);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFST(Reg: TPegtopMathReg);
begin
  AddLine('fst ' + MathRegName[Reg]);
  AddCode($DD);                 // fst st(Reg)
  AddCode($D0 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFSTP(Adr: TPegtopVirtualAddr);
begin
  AddLine('fstp qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DD);               // fstp qword ptr [esi+adr]
    AddCode($5E);
    AddCode(Adr);
  end
  else begin
    AddCode($DD);               // fstp qword ptr [esi+adr]
    AddCode($9E);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFSTP(Reg: TPegtopMathReg);
begin
  AddLine('fstp ' + MathRegName[Reg]);
  AddCode($DD);                 // fstp st(Reg)
  AddCode($D8 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFADD(Adr: TPegtopVirtualAddr);
begin
  AddLine('fadd qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DC);               // fadd qword ptr [esi+adr]
    AddCode($46);
    AddCode(Adr);
  end
  else begin
    AddCode($DC);               // fadd qword ptr [esi+adr]
    AddCode($86);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFADD(Reg1, Reg2: TPegtopMathReg);
begin
  AddLine('fadd ' + MathRegName[Reg2] + ', ' + MathRegName[Reg2]);
  if Reg1 = st0 then begin
    AddCode($D8);                 // fadd st(0), st(Reg2)
    AddCode($C0 or MathRegMask[Reg2]);
  end
  else if Reg2 = st0 then begin
    AddCode($DC);                 // fadd st(Reg1), st(0)
    AddCode($C0 or MathRegMask[Reg1]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('fadd', MathRegName[Reg1], MathRegName[st0]);
  end;
end;

procedure TPegtopAsmCode.CodeFADDP(Reg1, Reg2: TPegtopMathReg);
begin
  AddLine('faddp ' + MathRegName[Reg1] + ', ' + MathRegName[Reg2]);
  if Reg2 = st0 then begin
    AddCode($DE);                 // faddp st(Reg1), st(0)
    AddCode($C0 or MathRegMask[Reg1]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('faddp', MathRegName[Reg2], MathRegName[st0]);
  end;
end;

procedure TPegtopAsmCode.CodeFSUB(Adr: TPegtopVirtualAddr);
begin
  AddLine('fsub qword ptr [esi+' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DC);               // fsub qword ptr [esi+adr]
    AddCode($66);
    AddCode(Adr);
  end
  else begin
    AddCode($DC);               // fsub qword ptr [esi+adr]
    AddCode($A6);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFSUB(Reg1, Reg2: TPegtopMathReg);
begin
  AddLine('fsub ' + MathRegName[Reg1] + ', ' + MathRegName[Reg2]);
  if Reg1 = st0 then begin
    AddCode($D8);                 // fsub st(0), st(Reg2)
    AddCode($E0 or MathRegMask[Reg2]);
  end
  else if Reg2 = st0 then begin
    AddCode($DC);                 // fsub st(Reg1), st(o)
    AddCode($E8 or MathRegMask[Reg1]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('fsub', MathRegName[Reg1], MathRegName[st0]);
  end;
end;

procedure TPegtopAsmCode.CodeFSUBR(Adr: TPegtopVirtualAddr);
begin
  AddLine('fsubr qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DC);               // fsubr qword ptr [esi+adr]
    AddCode($6E);
    AddCode(Adr);
  end
  else begin
    AddCode($DC);               // fsubr qword ptr [esi+adr]
    AddCode($AE);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFSUBR(Reg1, Reg2: TPegtopMathReg);
begin
  AddLine('fsubr ' + MathRegName[Reg1] + ', ' + MathRegName[Reg2]);
  if Reg1 = st0 then begin
    AddCode($D8);                 // fsubr st(0), st(Reg2)
    AddCode($E8 or MathRegMask[Reg2]);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('fsubr', MathRegName[Reg1], MathRegName[st0]);
  end;
end;

procedure TPegtopAsmCode.CodeFMUL(Adr: TPegtopVirtualAddr);
begin
  AddLine('fmul qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DC);               // fmul qword ptr [esi+adr]
    AddCode($4E);
    AddCode(Adr);
  end
  else begin
    AddCode($DC);               // fmul qword ptr [esi+adr]
    AddCode($8E);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFMUL(Reg: TPegtopMathReg);
begin
  AddLine('fmul st(0), ' + MathRegName[Reg]);
  AddCode($D8);                 // fmul st(0), st(Reg)
  AddCode($C8 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFDIV(Adr: TPegtopVirtualAddr);
begin
  AddLine('fdiv qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DC);               // fdiv qword ptr [esi+adr]
    AddCode($76);
    AddCode(Adr);
  end
  else begin
    AddCode($DC);               // fdiv qword ptr [esi+adr]
    AddCode($B6);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFDIVR(Adr: TPegtopVirtualAddr);
begin
  AddLine('fdivr qword ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($DC);               // fdivr qword ptr [esi+adr]
    AddCode($7E);
    AddCode(Adr);
  end
  else begin
    AddCode($DC);               // fdivr qword ptr [esi+adr]
    AddCode($BE);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFXCHG(Reg: TPegtopMathReg);
begin
  AddLine('fxchg ' + MathRegName[Reg]);
  AddCode($D9);                 // fxchg st(Reg)
  AddCode($C8 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFFREE(Reg: TPegtopMathReg);
begin
  AddLine('ffree ' + MathRegName[Reg]);
  AddCode($DD);                 // ffree st(Reg)
  AddCode($C0 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFCHS;
begin
  AddLine('fchs');
  AddCode($D9);                 // fchs
  AddCode($E0);
end;

procedure TPegtopAsmCode.CodeFABS;
begin
  AddLine('fabs');
  AddCode($D9);                 // fabs
  AddCode($E1);
end;

procedure TPegtopAsmCode.CodeFSQRT;
begin
  AddLine('fsqrt');
  AddCode($D9);                 // fsqrt
  AddCode($FA);
end;

procedure TPegtopAsmCode.CodeFSTCW(Adr: TPegtopVirtualAddr);
begin
  CodeWAIT;
  AddLine('fstcw word ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($D9);               // fstcw word ptr [esi+adr]
    AddCode($7E);
    AddCode(Adr);
  end
  else begin
    AddCode($D9);               // fstcw word ptr [esi+adr]
    AddCode($BE);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFLDCW(Adr: TPegtopVirtualAddr);
begin
  AddLine('fldcw word ptr [esi' + AddrToHex(Adr) + ']');
  if (Adr < 128) and (Adr >= -128) then begin
    AddCode($D9);               // fldcw word ptr [esi+adr]
    AddCode($6E);
    AddCode(Adr);
  end
  else begin
    AddCode($D9);               // fldcw word ptr [esi+adr]
    AddCode($AE);
    AddAddr(Adr);
  end;
end;

procedure TPegtopAsmCode.CodeFSIN;
begin
  AddLine('fsin');
  AddCode($D9);                 // fsin
  AddCode($FE);
end;

procedure TPegtopAsmCode.CodeFCOS;
begin
  AddLine('fcos');
  AddCode($D9);                 // fcos
  AddCode($FF);
end;

procedure TPegtopAsmCode.CodeFPTAN;
begin
  AddLine('fptan');
  AddCode($D9);                 // fptan
  AddCode($F2);
end;

procedure TPegtopAsmCode.CodeFPATAN;
begin
  AddLine('fpatan');
  AddCode($D9);                 // fpatan
  AddCode($F3);
end;

procedure TPegtopAsmCode.CodeFYL2X;
begin
  AddLine('fyl2x');
  AddCode($D9);                 // fyl2x
  AddCode($F1);
end;

procedure TPegtopAsmCode.CodeF2XM1;
begin
  AddLine('f2xm1');
  AddCode($D9);                 // f2xm1
  AddCode($F0);
end;

procedure TPegtopAsmCode.CodeFNCLEX;
begin
  AddLine('fnclex');
  AddCode($DB);                 // fnclex
  AddCode($E2);
end;

procedure TPegtopAsmCode.CodeFSCALE;
begin
  AddLine('fscale');
  AddCode($D9);                 // fscale
  AddCode($FD);
end;

procedure TPegtopAsmCode.CodeFRNDINT;
begin
  AddLine('frndint');
  AddCode($D9);                 // frndint
  AddCode($FC);
end;

procedure TPegtopAsmCode.CodeFPREM;
begin
  AddLine('fprem');
  AddCode($D9);                 // fprem
  AddCode($F8);
end;

procedure TPegtopAsmCode.CodeFPREM1;
begin
  AddLine('fprem1');
  AddCode($D9);                 // fprem1
  AddCode($F5);
end;

procedure TPegtopAsmCode.CodeFXTRACT;
begin
  AddLine('fxtract');
  AddCode($D9);                 // fxtract
  AddCode($F4);
end;

procedure TPegtopAsmCode.CodeFTST;
begin
  AddLine('ftst');
  AddCode($D9);                 // ftst
  AddCode($E4);
end;

procedure TPegtopAsmCode.CodeFSTSW(Reg: TPegtopProcReg);
begin
  CodeWAIT;
  AddLine('fstsw ' + ProcRegName[Reg]);
  if Reg = regAX then begin
    AddCode($DF);                 // fstsw ax
    AddCode($E0);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('fstsw', ProcRegName[Reg], ProcRegName[regAX]);
  end;
end;

procedure TPegtopAsmCode.CodeFLDLN2;
begin
  AddLine('fldln2');
  AddCode($D9);                 // fldln2
  AddCode($ED);
end;

procedure TPegtopAsmCode.CodeFLDLG2;
begin
  AddLine('fldlg2');
  AddCode($D9);                 // fldlg2
  AddCode($EC);
end;

procedure TPegtopAsmCode.CodeFCOM(Reg: TPegtopMathReg);
begin
  AddLine('fcom ' + MathRegName[Reg]);
  AddCode($D8);                 // fcom st(reg)
  AddCode($D0 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFCOMP(Reg: TPegtopMathReg);
begin
  AddLine('fcomp ' + MathRegName[Reg]);
  AddCode($D8);                 // fcomp st(reg)
  AddCode($D8 or MathRegMask[Reg]);
end;

procedure TPegtopAsmCode.CodeFCOMPP;
begin
  AddLine('fcompp');
  AddCode($DE);                 // fcompp
  AddCode($D9);
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopMacroAsmCode
////////////////////////////////////////////////////////////////////////////////

procedure TPegtopMacroAsmCode.CodeCALL(Adr: Pointer);
begin
  CodePUSH(regEAX);
  CodeMOV(regEAX, LongWord(Adr));
  CodeCALL(regEAX);
  CodePOP(regEAX);
end;

procedure TPegtopMacroAsmCode.CodeFMOD(Adr: TPegtopVirtualAddr);
begin
  CodeFLD(Adr);
  CodeFXCHG(st1);
  CodeFPREM;
  CodeFSTP(st1); // throw away st(1)
end;

procedure TPegtopMacroAsmCode.CodeFMODR(Adr: TPegtopVirtualAddr);
begin
  CodeFLD(Adr);
  CodeFPREM;
  CodeFSTP(st1); // throw away st(1)
end;

procedure TPegtopMacroAsmCode.CodeFSQR;
begin
  CodeFMUL(st0);
end;

procedure TPegtopMacroAsmCode.CodeFEXP;
begin
  // idea: exp(x) = 2 ^ (x * log2(e)) = 2 ^ x'
  //       2 ^ x' = 2 ^ frac(x') * 2 ^ int(x')
  CodeFLDL2E;               // load log2(e)
  CodeFMUL(st1);            // x' = x * log2(e)
  CodeFST(st1);             // duplicate st(0)
  CodeFROUND(pmrDefault);   // i = int(x')
  CodeFSUB(st1, st0);       // f = frac(x')
  CodeFXCHG(st1);
  CodeF2XM1;                // 2 ^ f - 1
  CodeFLD1;
  CodeFADDP(st1, st0);      // + 1
  CodeFSCALE;               // * 2 ^ i
  CodeFSTP(st1);            // throw away st(1)
end;

procedure TPegtopMacroAsmCode.CodeFPOWPOS;
begin
  // idea: b ^ x = 2 ^ (x * log2(b)) = 2 ^ x'
  //       2 ^ x' = 2 ^ frac(x') * 2 ^ int(x')
  CodeFYL2X;                // x' = x * log2(b)
  CodeFST(st1);             // duplicate st(0)
  CodeFROUND(pmrDefault);   // i = int(x')
  CodeFSUB(st1, st0);       // f = frac(x')
  CodeFXCHG(st1);
  CodeF2XM1;                // 2 ^ f - 1
  CodeFLD1;
  CodeFADDP(st1, st0);      // + 1
  CodeFSCALE;               // * 2 ^ i
  CodeFSTP(st1);            // throw away st(1)
end;

procedure TPegtopMacroAsmCode.CodeFPOW;
begin
  OpenNameSpace('FPOW');
  try
    CodeFTSTSYNC(st0);      // compare base with 0
    CodeJA('@def');         // greater? => default case
    CodeJB('@neg');         // smaller?

    // base = 0 (=> result = 1):
    CodeFSTP(st0);          // throw away base...
    CodeFSTP(st0);          // ...and exponent
    CodeFLD1;               // return 1
    CodeJMP('@end');

    // base < 0:
    AddLabel('@neg');
    CodeFLD(st1);           // load exponent
    CodeFROUND(pmrDefault);
    CodeFCOMSYNC(st0, st2);
    CodeJNE('@invalid');    // raise exception
    CodeFMOD2;
    CodeFTSTSYNC(st0);
    CodeFSTP(st0);          // throw away st(0)
    CodeFCHS;               // neg. base
    CodeJE('@def');

    // odd exponent:
    // idea: (-b) ^ x = -(b ^ x) for odd x
    AddLabel('@odd');
    CodeCALL(ProcFPOWPOS.Ptr[0]);
    CodeFCHS;               // neg. result
    CodeJMP('@end');

    // invalid (neg. base and non-int. exponent):
    AddLabel('@invalid');
    CodeFSTP(st0);            // throw away st(0)
    CodeFYL2X;                // raises exception
    CodeFSTP(st1);            // throw away st(1)
    CodeJMP('@end');

    // default case (base > 0):
    AddLabel('@def');
    CodeCALL(ProcFPOWPOS.Ptr[0]);

    AddLabel('@end');
  finally
    CloseNameSpace;
  end;
end;

procedure TPegtopMacroAsmCode.CodeFPOWR(Adr: TPegtopVirtualAddr);
begin
  CodeFLD(Adr);
  CodeCALL(ProcFPOW.Ptr[0]);
end;

procedure TPegtopMacroAsmCode.CodeFROUND(Rounding: TPegtopMathRounding);
begin
  if Rounding = pmrDefault then begin
    CodeFRNDINT;
  end
  else begin
    CodeFSTCW(FFPUControlWordAddr);
    CodeFLDCW(FFPUControlWordAddr + Ord(Rounding) * SizeOf(Word));
    CodeFRNDINT;
    CodeFLDCW(FFPUControlWordAddr);
  end;
end;

procedure TPegtopMacroAsmCode.CodeFFRAC(Rounding: TPegtopMathRounding);
begin
  CodeFLD(st0);         // duplicate st(0) to st(1)
  CodeFROUND(pmrToZero); // transform x.y to x.0
  CodeFSUBR(st0, st1);  // subtract x.0 from x.y
  CodeFSTP(st1);        // throw away st(1)
end;

procedure TPegtopMacroAsmCode.CodeFLN;
begin
  CodeFLDLN2;
  CodeFXCHG(st1);
  CodeFYL2X;
end;

procedure TPegtopMacroAsmCode.CodeFLG;
begin
  CodeFLDLG2;
  CodeFXCHG(st1);
  CodeFYL2X;
end;

procedure TPegtopMacroAsmCode.CodeFL2;
begin
  CodeFLD1;
  CodeFXCHG(st1);
  CodeFYL2X;
end;

procedure TPegtopMacroAsmCode.CodeFATAN;
begin
  CodeFLD1;
  CodeFPATAN;
end;

procedure TPegtopMacroAsmCode.CodeFTAN;
begin
  CodeFPTAN;
  CodeFSTP(st0);        // throw away st(1)
end;

procedure TPegtopMacroAsmCode.CodeFTSTSYNC(Reg: TPegtopMathReg);
begin
  CodePUSH(regAX);
  if Reg <> st0 then CodeFXCHG(Reg);
  CodeFTST;
  CodeFSTSW(regAX);
  CodeSAHF;
  if Reg <> st0 then CodeFXCHG(Reg);
  CodePOP(regAX);
end;

procedure TPegtopMacroAsmCode.CodeFCOMSYNC(Reg1, Reg2: TPegtopMathReg);
begin
  if (Reg1 = st0) then begin
    CodePUSH(regAX);
    CodeFCOM(Reg2);
    CodeFSTSW(regAX);
    CodeSAHF;
    CodePOP(regAX);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('fcompare', MathRegName[Reg1], MathRegName[st0]);
  end;
end;

procedure TPegtopMacroAsmCode.CodeFCOMPSYNC(Reg1, Reg2: TPegtopMathReg);
begin
  if (Reg1 = st0) then begin
    CodePUSH(regAX);
    CodeFCOMP(Reg2);
    CodeFSTSW(regAX);
    CodeSAHF;
    CodePOP(regAX);
  end
  else begin
    raise EPegtopAsmInvalidOperand.CreateProp('fcompare', MathRegName[Reg1], MathRegName[st0]);
  end;
end;

procedure TPegtopMacroAsmCode.CodeFSIGN;
begin
  OpenNameSpace('FSIGN');
  try
    CodeFTSTSYNC(st0);      // compare st(0) with 0
    CodeJE('@end');
    CodeFSTP(st0);          // throw away st(0)
    CodeFLD1;               // load 1.0
    CodeJA('@end');
    CodeFCHS;               // change sign
    AddLabel('@end');
  finally
    CloseNameSpace;
  end;
end;

procedure TPegtopMacroAsmCode.CodeFDIV2;
begin
  CodeFLD1;             // 1
  CodeFLD1;             // + 1
  CodeFADDP(st1, st0);  // = 2
  CodeFXCHG(st1);
  CodeFSCALE;
  CodeFSTP(st1);        // throw away st(1)
end;

procedure TPegtopMacroAsmCode.CodeFMOD2;
begin
  CodeFLD1;             // 1
  CodeFLD1;             // + 1
  CodeFADDP(st1, st0);  // = 2
  CodeFXCHG(st1);
  CodeFPREM;
  CodeFSTP(st1);        // throw away st(1)
end;

initialization
  ProcFPOWPOS := TPegtopMacroAsmCode.Create(NIL);
  ProcFPOWPOS.CodeFPOWPOS;
  ProcFPOWPOS.CodeRET;
  ProcFPOW := TPegtopMacroAsmCode.Create(NIL);
  ProcFPOW.CodeFPOW; // ProcFPOWPOS must be created first
  ProcFPOW.CodeRET;
finalization
  ProcFPOWPOS.Free;
  ProcFPOW.Free;
end.
