////////////////////////////////////////////////////////////////////////////////
// Components: TPegtopFormula
// Version:    1.01
// Date:       01 Aug 2003 created 1.00
//             14 Feb 2004 modified 1.01 (user defined functions added)
// Author:     Jens Gruschel (GRU)
// Copyright:  (c) 2003 Jens Gruschel
// Website:    http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////
// Description:
// TPegtopFormula is a fast expression evaluator. The formula (which is given
// as a string at runtime) is parsed and compiled once, so evaluation is done
// with speed of assembler code (comparable to the speed of Delphi code).
// Arguments and result have double precision (the default 64 bit double type),
// all calculation is done with the math coprocessor (80x87 instructions).
////////////////////////////////////////////////////////////////////////////////
// Usage:
// Before you assign a string (like 'x+2') to the Formula property, you have
// to define all parameters (like Parameters.Add('x')). The formula is compiled
// automatically before the first evaluation is done (or by calling Compile).
// You can either set arguments by index (like SetArg(0, 2.3), where the first
// parameter has index = 0) or by name (like SetArg('x', 2.3)), which is a bit
// slower, or directly by calling the overloaded Eval method with parameters
// (like Eval([2.3, 7] to set the first parameter to 2.3 and the second to 7).
////////////////////////////////////////////////////////////////////////////////
// License:
// Please read the license.txt file that comes with this software
// and visit http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////
// Documentation / Installation:
// Please read the readme.txt file that comes with this software
// and visit http://www.pegtop.net/delphi
////////////////////////////////////////////////////////////////////////////////

unit PegtopFormulas;

interface

uses
  Classes, SysUtils,
  PegtopCompilers, PegtopAssemblers,
{$IFDEF LINUX}
 QDialogs;
{$ENDIF}
{$IFDEF MSWINDOWS}
 Dialogs;
{$ENDIF}


type
  EPegtopFormula = class(Exception);

  EPegtopFormulaArgument = class(EPegtopFormula);

  EPegtopFormulaInvalid = class(EPegtopFormula);

  EPegtopFormulaParameter = class(EPegtopFormula);

  EPegtopFormulaParse = class(EPegtopFormula)
  private
    FExpression: String;
    FPosition: Integer;
    FInvalid: String;
    FExpected: String;
  public
    constructor CreateProp(const AExpression: String; const APosition: Integer; const AExpected: String);
    property Expression: String read FExpression;
    property Position: Integer read FPosition;
    property Invalid: String read FInvalid;
    property Expected: String read FExpected;
  end;

  EPegtopFormulaBadSpecifier = class(EPegtopFormulaParse)
  private
    FSpecifier: String;
  public
    constructor CreateProp(const ASpecifier: String);
    property Specifier: String read FSpecifier;
  end;

  TPegtopFormulaCompiler = class; // forward declaration

  TPegtopFormulaElement = class
  private
    FLevel: Integer;
    FPrev: TPegtopFormulaElement;
    FNegative: Boolean;
  protected
    function GetPrio: Integer; virtual; abstract;
  public
    procedure Compile(Code: TPegtopMacroAsmCode); virtual; abstract;
    property Prio: Integer read GetPrio;
    property Level: Integer read FLevel write FLevel;
    property Prev: TPegtopFormulaElement read FPrev write FPrev;
    property Negative: Boolean read FNegative;
  end;

  TPegtopFormulaOperator = (opNone, opAdd, opSub, opMul, opDiv, opMod, opPow);

  TPegtopFormulaVariable = class(TPegtopFormulaElement)
  private
    FDataAdr: Integer;
    FOperator: TPegtopFormulaOperator;
  protected
    function GetPrio: Integer; override;
  public
    constructor Create(ADataAdr: Integer; ANegative: Boolean);
    procedure Compile(Code: TPegtopMacroAsmCode); override;
    property DataAdr: Integer read FDataAdr;
    property Operator: TPegtopFormulaOperator read FOperator write FOperator;
  end;

  TPegtopFormulaFunction = class(TPegtopFormulaElement)
  protected
    function GetPrio: Integer; override;
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
    constructor Create(ANegative: Boolean);
  end;

  TPegtopFormulaNeg = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaAbs = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaSqr = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaSqrt = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaRound = class(TPegtopFormulaFunction)
  private
    FRounding: TPegtopMathRounding;
  public
    constructor Create(ANegative: Boolean; ARounding: TPegtopMathRounding);
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaFrac = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaSin = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaCos = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaTan = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaArcTan = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaLn = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaLog10 = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaLog2 = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaExp = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

  TPegtopFormulaSign = class(TPegtopFormulaFunction)
  public
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;

{  TPegtopFormulaUserDefined1 = class(TPegtopFormulaFunction)
  private
    FFunc: TPegtopFormulaFunction1;
  public
    constructor Create(AFunc: TPegtopFormulaFunction1);
    procedure Compile(Code: TPegtopMacroAsmCode); override;
  end;}

  TPegtopFormulaCompiler = class
  private
    FCode: TPegtopMacroAsmCode;
    FParameters: TStrings;
    FStack: TPegtopFormulaElement;
    FVariable: TPegtopFormulaVariable;
    // FLabels: TStringList;
    // FJumps: TObjectList;
    function AddSpecifier(Specifier: String; Negative: Boolean; Level: Integer): Boolean;
    procedure AddConstant(Value: Double; Negative: Boolean);
    procedure AddOperator(C: Char; Level: Integer);
    procedure AddStack(V: TPegtopFormulaVariable; Lev: Integer); overload;
    procedure AddStack(F: TPegtopFormulaFunction; Lev: Integer); overload;
  public
    constructor Create(ACode: TPegtopMacroAsmCode; Parameters: TStrings);
    destructor Destroy; override;
    procedure Parse(Expression: String;
      AutoCloseBrackets: Boolean = True;
      NationalDecimalSeparator: Boolean = False);
    procedure CleanUp;
  end;

  TPegtopFormulaOption = (foAutoCloseBrackets, foNationalDecimalSeparator);
  TPegtopFormulaOptions = set of TPegtopFormulaOption;

  TPegtopFormula = class(TComponent)
  private
    FData: TPegtopCompilerData;
    FCode: TPegtopMacroAsmCode;

    FExpression: String;
    FParameters: TStrings;
    FParamCount: Integer;
    FCompiled: Boolean;
    FValid: Boolean;
    FOptions: TPegtopFormulaOptions;
    FMathErrors: TPegtopMathErrors;
    FMathPrecision: TPegtopMathPrecision;
    FMathInfinity: TPegtopMathInfinity;

    procedure ParametersChange(Sender: TObject);
    function GetParamIndex(Name: String): Integer;
    procedure SetParameters(Value: TStrings);
    procedure SetExpression(Value: String);
    procedure SetOptions(V: TPegtopFormulaOptions);
    procedure SetMathErrors(V: TPegtopMathErrors);
    procedure SetMathPrecision(V: TPegtopMathPrecision);
    procedure SetMathInfinity(V: TPegtopMathInfinity);
    function GetCodeSize: Integer;
    function GetCodeLines: TStrings;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Compile;
    procedure Recompile;
    function Eval(const Args: array of Double): Double; overload;
    function Eval: Double; overload;
    function GetArg(Index: Integer): Double; overload;
    function GetArg(Name: String): Double; overload;
    procedure SetArg(Index: Integer; Value: Double); overload;
    procedure SetArg(Name: String; Value: Double); overload;
    property CodeSize: Integer read GetCodeSize;
    property CodeLines: TStrings read GetCodeLines;
    property Valid: Boolean read FValid;
  published
    property Parameters: TStrings read FParameters write SetParameters;
    property Options: TPegtopFormulaOptions read FOptions write SetOptions;
    property MathErrors: TPegtopMathErrors read FMathErrors write SetMathErrors;
    property MathPrecision: TPegtopMathPrecision read FMathPrecision write SetMathPrecision;
    property MathInfinity: TPegtopMathInfinity read FMathInfinity write SetMathInfinity;
    property Expression: String read FExpression write SetExpression;
  end;

function IsInfiniteOrNaN(V: Double): Boolean;
function IsNaN(V: Double): Boolean;
function IsInfinite(V: Double): Boolean;
function IsPosInfinite(V: Double): Boolean;
function IsNegInfinite(V: Double): Boolean;
function IsZero(V: Double): Boolean;
function IsPosZero(V: Double): Boolean;
function IsNegZero(V: Double): Boolean;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Pegtop Math', [TPegtopFormula]);
end;

const
  ErrorMsgArgIndex = 'The argument index %d refers no valid parameter of this formula (must be between %d and %d).';
  ErrorMsgArgName = 'The argument name %s refers no valid parameter of this formula.';
  ErrorMsgNoParams = 'No parameters for this formula specified.';
  ErrorMsgTooManyArgs = 'Not more than %d arguments can be passed to this formula.';
  ErrorMsgInvalid = '"%s" is not a valid formula.';
  ErrorMsgParamNotUnique = '"%s" must be a unique parameter name.';

const
  BracketPrecedence = 10;

type
  TCharClass = (ccUnknown, ccSpace, ccAlpha, ccDigit, ccDecSep, ccOperator, ccBraOpen, ccBraClose);
  TPrecedence = 0..BracketPrecedence;
  TOperatorPrecedence = 0..BracketPrecedence-1;

const
  OperatorPrecedence: array[TPegtopFormulaOperator] of TOperatorPrecedence = (0, 1, 1, 2, 2, 2, 3);

var
  CharClass: array[Char] of TCharClass;

function IsInfiniteOrNaN(V: Double): Boolean;
begin
  Result := (PInt64(@V)^ and $7FF0000000000000) = $7FF0000000000000;
end;

function IsNaN(V: Double): Boolean;
begin
  Result := ((PInt64(@V)^ and $7FF0000000000000) = $7FF0000000000000)
  and ((PInt64(@V)^ and $000FFFFFFFFFFFFF) <> $0000000000000000);
end;

function IsInfinite(V: Double): Boolean;
begin
  Result := (PInt64(@V)^ and $7FFFFFFFFFFFFFFF) = $7FF0000000000000;
end;

function IsPosInfinite(V: Double): Boolean;
begin
  Result := PInt64(@V)^ = $7FF0000000000000;
end;

function IsNegInfinite(V: Double): Boolean;
begin
  Result := PInt64(@V)^ = $FFF0000000000000;
end;

function IsZero(V: Double): Boolean;
begin
  Result := ((PInt64(@V)^ and $7FFFFFFFFFFFFFFF) = $0000000000000000);
end;

function IsPosZero(V: Double): Boolean;
begin
  Result := PInt64(@V)^ = $0000000000000000;
end;

function IsNegZero(V: Double): Boolean;
begin
  Result := PInt64(@V)^ = $8000000000000000;
end;

////////////////////////////////////////////////////////////////////////////////
// CallCode (run code that's stored in Code array)
////////////////////////////////////////////////////////////////////////////////

function CallCode(Code, Data: Pointer): Double;
// Code = eax, Data = edx, Result = st(0)
asm
  push esi
  mov esi, Data
  // add esi, $80 // might be useful to generate more 8 bit addresses (instead of 32 bit)
  call Code
  pop esi
end;

////////////////////////////////////////////////////////////////////////////////
// EPegtopFormulaParse
////////////////////////////////////////////////////////////////////////////////

constructor EPegtopFormulaParse.CreateProp(const AExpression: String; const APosition: Integer;
  const AExpected: String);
const
  ErrorMsgParse = 'Error parsing formula at position %d after "%s": %s expected but %s found.';
begin
  FExpression := AExpression;
  FPosition := APosition;
  FExpected := AExpected;
  if FPosition > Length(FExpression) then FInvalid := 'end of formula'
  else if CharClass[FExpression[FPosition]] = ccSpace then FInvalid := 'white space'
  else if FExpression[FPosition] < #32 then FInvalid := 'invalid character'
  else FInvalid := '"' + FExpression[FPosition] + '"';
  inherited CreateFmt(ErrorMsgParse, [FPosition, Copy(FExpression, 1, FPosition-1), FExpected, FInvalid]);
end;

////////////////////////////////////////////////////////////////////////////////
// EPegtopFormulaBadSpecifier
////////////////////////////////////////////////////////////////////////////////

constructor EPegtopFormulaBadSpecifier.CreateProp(const ASpecifier: String);
const
  ErrorMsgBadSpecifier = '"%s" is not a valid specifier (function / parameter name) for this formula.';
begin
  FSpecifier := ASpecifier;
  inherited CreateFmt(ErrorMsgBadSpecifier, [FSpecifier]);
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopFormula
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopFormula.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FData := TPegtopCompilerData.Create;
  FCode := TPegtopMacroAsmCode.Create(FData);
  FParameters := TStringList.Create;
  TStringList(FParameters).OnChange := ParametersChange;
  FOptions := [foAutoCloseBrackets];
  FMathErrors := [pmeInvalidOp, pmeDivByZero, pmeOverflow];
  FMathPrecision := pmp64bit;
  FMathInfinity := pmiAffine;
  FParamCount := 0;
  FValid := False;
  FCompiled := False;
end;

destructor TPegtopFormula.Destroy;
begin
  FParameters.Free;
  FCode.Free;
  FData.Free;
  inherited;
end;

function TPegtopFormula.Eval(const Args: array of Double): Double;
var
  I, L: Integer;
  P: ^Double;
  ControlWordBackup: Word;
begin
  Compile;
  if not FValid then raise EPegtopFormulaInvalid.CreateFmt(ErrorMsgInvalid, [FExpression]);
  L := Length(Args);
  if L > FParamCount then raise EPegtopFormulaArgument.CreateFmt(ErrorMsgTooManyArgs, [FParamCount]);
  P := FData.Ptr[0];
  for I := 0 to L-1 do begin
    P^ := Args[I];
    Inc(P);
  end;
  asm
    fstcw ControlWordBackup
  end;
  try
    Result := CallCode(FCode.Ptr[0], FData.Ptr[0]);
  finally
    asm
      fclex
      fldcw ControlWordBackup
    end;
  end;
end;

function TPegtopFormula.Eval: Double;
begin
  Compile;
  if not FValid then raise EPegtopFormulaInvalid.CreateFmt(ErrorMsgInvalid, [FExpression]);
  Result := CallCode(FCode.Ptr[0], FData.Ptr[0]);
end;

procedure TPegtopFormula.Compile;
begin
  if not FCompiled then Recompile;
end;

procedure TPegtopFormula.Recompile;
var
  Compiler: TPegtopFormulaCompiler;
begin
  FValid := False;
  FCompiled := True;
  FData.Clear;
  FCode.Clear;
  FData.IncSize(FParameters.Count * SizeOf(Double), SizeOf(Double));
  FCode.AddFPUControlWords(FMathErrors, FMathPrecision, FMathInfinity);
  Compiler := TPegtopFormulaCompiler.Create(FCode, FParameters);
  try
    Compiler.Parse(FExpression, foAutoCloseBrackets in FOptions);
    Compiler.CleanUp;
  finally
    Compiler.Free;
  end;
  FValid := True;
end;

function TPegtopFormula.GetParamIndex(Name: String): Integer;
begin
  Result := FParameters.IndexOf(Name);
end;

function TPegtopFormula.GetArg(Index: Integer): Double;
var
  P: ^Double;
begin
  if (Index >= 0) and (Index < FParamCount) then begin
    P := FData.Ptr[0];
    Inc(P, Index);
    Result := P^;
  end
  else if FParamCount > 0 then begin
    raise EPegtopFormulaArgument.CreateFmt(ErrorMsgArgIndex, [Index, 0, FParamCount-1]);
  end
  else begin
    raise EPegtopFormulaArgument.Create(ErrorMsgNoParams);
  end;
end;

function TPegtopFormula.GetArg(Name: String): Double;
var
  Index: Integer;
begin
  Index := GetParamIndex(Name);
  if Index > 0 then begin
    Result := GetArg(Index);
  end
  else if FParamCount > 0 then begin
    raise EPegtopFormulaArgument.CreateFmt(ErrorMsgArgName, [Name]);
  end
  else begin
    raise EPegtopFormulaArgument.Create(ErrorMsgNoParams);
  end;
end;

procedure TPegtopFormula.SetArg(Index: Integer; Value: Double);
var
  P: ^Double;
begin
  if (Index >= 0) and (Index < FParamCount) then begin
    P := FData.Ptr[0];
    Inc(P, Index);
    P^ := Value;
  end
  else if FParamCount > 0 then begin
    raise EPegtopFormulaArgument.CreateFmt(ErrorMsgArgIndex, [Index, 0, FParamCount-1]);
  end
  else begin
    raise EPegtopFormulaArgument.Create(ErrorMsgNoParams);
  end;
end;

procedure TPegtopFormula.SetArg(Name: String; Value: Double);
var
  Index: Integer;
begin
  Index := GetParamIndex(Name);
  if Index > 0 then begin
    SetArg(Index, Value);
  end
  else if FParamCount > 0 then begin
    raise EPegtopFormulaArgument.CreateFmt(ErrorMsgArgName, [Name]);
  end
  else begin
    raise EPegtopFormulaArgument.Create(ErrorMsgNoParams);
  end;
end;

procedure TPegtopFormula.SetParameters(Value: TStrings);
begin
  if Value = NIL then FParameters.Clear
  else FParameters.Assign(Value);
  ParametersChange(Self);
end;

procedure TPegtopFormula.SetExpression(Value: String);
begin
  FExpression := Value;
  FValid := False;
  FCompiled := False;
end;

procedure TPegtopFormula.SetOptions(V: TPegtopFormulaOptions);
begin
  if (foAutoCloseBrackets in V) <> (foAutoCloseBrackets in FOptions) then begin
    FCompiled := False; // formula has to be recompiled
    FValid := False;
  end;
  FOptions := V;
end;

procedure TPegtopFormula.SetMathErrors(V: TPegtopMathErrors);
begin
  if FMathErrors <> V then begin
    FMathErrors := V;
    if FValid then FCode.SetFPUControlWords(FMathErrors, FMathPrecision, FMathInfinity);
  end;
end;

procedure TPegtopFormula.SetMathPrecision(V: TPegtopMathPrecision);
begin
  if FMathPrecision <> V then begin
    FMathPrecision := V;
    if FValid then FCode.SetFPUControlWords(FMathErrors, FMathPrecision, FMathInfinity);
  end;
end;

procedure TPegtopFormula.SetMathInfinity(V: TPegtopMathInfinity);
begin
  if FMathInfinity <> V then begin
    FMathInfinity := V;
    if FValid then FCode.SetFPUControlWords(FMathErrors, FMathPrecision, FMathInfinity);
  end;
end;

procedure TPegtopFormula.ParametersChange(Sender: TObject);
var
  I, J: Integer;
begin
  FParamCount := FParameters.Count;
  FValid := False;
  FCompiled := False;
  FData.Clear;
  FCode.Clear;
  FData.IncSize(FParameters.Count * SizeOf(Double), SizeOf(Double));
  for I := 1 to FParameters.Count-1 do begin
    for J := 0 to I-1 do begin
      if LowerCase(FParameters[I]) = LowerCase(FParameters[J]) then
        raise EPegtopFormulaParameter.CreateFmt(ErrorMsgParamNotUnique, [FParameters[J]]);
    end;
  end;
end;

function TPegtopFormula.GetCodeSize: Integer;
begin
  Compile;
  if FValid then Result := FCode.Size
  else Result := 0;
end;

function TPegtopFormula.GetCodeLines: TStrings;
begin
  Compile;
  if not FValid then FCode.Clear;
  Result := FCode.Lines;
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopFormulaVariable
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopFormulaVariable.Create(ADataAdr: Integer; ANegative: Boolean);
begin
  FDataAdr := ADataAdr;
  FNegative := ANegative;
  FOperator := opNone;
end;

function TPegtopFormulaVariable.GetPrio: Integer;
begin
  Result := BracketPrecedence * FLevel + OperatorPrecedence[FOperator];
end;

procedure TPegtopFormulaVariable.Compile(Code: TPegtopMacroAsmCode);
begin
  if not FNegative then begin
    case FOperator of
      opNone: begin
        Code.CodeFLD(FDataAdr);
      end;
      opAdd: begin
        Code.CodeFADD(FDataAdr);
      end;
      opSub: begin
        Code.CodeFSUBR(FDataAdr);
      end;
      opMul: begin
        Code.CodeFMUL(FDataAdr);
      end;
      opDiv: begin
        Code.CodeFDIVR(FDataAdr);
      end;
      opMod: begin
        Code.CodeFMODR(FDataAdr);
      end;
      opPow: begin
        Code.CodeFPOWR(FDataAdr);
      end;
    end;
  end
  else begin
    case FOperator of
      opNone: begin
        Code.CodeFLD(FDataAdr);
        Code.CodeFCHS;
      end;
      opAdd: begin
        Code.CodeFSUBR(FDataAdr);
        Code.CodeFCHS;
      end;
      opSub: begin
        Code.CodeFADD(FDataAdr);
        Code.CodeFCHS;
      end;
      opMul: begin
        Code.CodeFMUL(FDataAdr);
        Code.CodeFCHS;
      end;
      opDiv: begin
        Code.CodeFDIVR(FDataAdr);
        Code.CodeFCHS;
      end;
      opMod: begin
        Code.CodeFMODR(FDataAdr);
        Code.CodeFCHS;
      end;
      opPow: begin
        Code.CodeFPOWR(FDataAdr);
        Code.CodeFCHS;
      end;
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TPegtopFormulaFunction
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopFormulaFunction.Create(ANegative: Boolean);
begin
  FNegative := ANegative;
end;

procedure TPegtopFormulaFunction.Compile(Code: TPegtopMacroAsmCode);
begin
  if Negative then Code.CodeFCHS;
end;

function TPegtopFormulaFunction.GetPrio: Integer;
begin
  Result := BracketPrecedence * FLevel + High(TOperatorPrecedence);
end;

procedure TPegtopFormulaNeg.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFCHS;
  inherited;
end;

procedure TPegtopFormulaAbs.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFABS;
  inherited;
end;

procedure TPegtopFormulaSqr.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFSQR;
  inherited;
end;

procedure TPegtopFormulaSqrt.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFSQRT;
  inherited;
end;

constructor TPegtopFormulaRound.Create(ANegative: Boolean; ARounding: TPegtopMathRounding);
begin
  inherited Create(ANegative);
  FRounding := ARounding;
end;

procedure TPegtopFormulaRound.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFROUND(FRounding);
  inherited;
end;

procedure TPegtopFormulaFrac.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFFRAC(pmrToZero);
  inherited;
end;

procedure TPegtopFormulaSin.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFSIN;
  inherited;
end;

procedure TPegtopFormulaCos.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFCOS;
  inherited;
end;

procedure TPegtopFormulaTan.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFTAN;
  inherited;
end;

procedure TPegtopFormulaArcTan.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFATAN;
  inherited;
end;

procedure TPegtopFormulaLn.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFLN;
  inherited;
end;

procedure TPegtopFormulaLog10.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFLG;
  inherited;
end;

procedure TPegtopFormulaLog2.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFL2;
  inherited;
end;

procedure TPegtopFormulaExp.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFEXP;
  inherited;
end;

procedure TPegtopFormulaSign.Compile(Code: TPegtopMacroAsmCode);
begin
  Code.CodeFSIGN;
  inherited;
end;

{constructor TPegtopFormulaUserDefined1.Create(AFunc: TPegtopFormulaFunction1);
begin
  FFunc := AFunc;
end;

procedure TPegtopFormulaUserDefined1.Compile(Code: TPegtopMacroAsmCode);
var
  Adr: Integer;
begin
  Adr := Code.Data.AddDouble(0.0); // ...reserve space...
  // move current value from math register to memory:
  Code.CodeFSTP(Adr);
  // push value form memory to stack:
  Code.CodePUSH(Adr + 4);
  Code.CodePUSH(Adr);
  // call function:
  Code.CodeCALL(Addr(FFunc));
  inherited;
end;}

////////////////////////////////////////////////////////////////////////////////
// TPegtopFormulaCompiler
////////////////////////////////////////////////////////////////////////////////

constructor TPegtopFormulaCompiler.Create(ACode: TPegtopMacroAsmCode; Parameters: TStrings);
begin
  FCode := ACode;
  FParameters := Parameters;
end;

destructor TPegtopFormulaCompiler.Destroy;
begin
  inherited;
end;

function TPegtopFormulaCompiler.AddSpecifier(Specifier: String; Negative: Boolean; Level: Integer): Boolean;
var
  I: Integer;
  S: String;
begin
  Result := False;
  S := LowerCase(Specifier);
  if S = 'neg' then AddStack(TPegtopFormulaNeg.Create(Negative), Level)
  else if S = 'abs' then AddStack(TPegtopFormulaAbs.Create(Negative), Level)
  else if S = 'sqr' then AddStack(TPegtopFormulaSqr.Create(Negative), Level)
  else if S = 'sqrt' then AddStack(TPegtopFormulaSqrt.Create(Negative), Level)
  else if S = 'round' then AddStack(TPegtopFormulaRound.Create(Negative, pmrNearest), Level)
  else if S = 'int' then AddStack(TPegtopFormulaRound.Create(Negative, pmrToZero), Level)
  else if S = 'floor' then AddStack(TPegtopFormulaRound.Create(Negative, pmrDown), Level)
  else if S = 'ceil' then AddStack(TPegtopFormulaRound.Create(Negative, pmrUp), Level)
  else if S = 'frac' then AddStack(TPegtopFormulaFrac.Create(Negative), Level)
  else if S = 'sin' then AddStack(TPegtopFormulaSin.Create(Negative), Level)
  else if S = 'cos' then AddStack(TPegtopFormulaCos.Create(Negative), Level)
  else if S = 'tan' then AddStack(TPegtopFormulaTan.Create(Negative), Level)
  else if S = 'arctan' then AddStack(TPegtopFormulaArcTan.Create(Negative), Level)
  else if S = 'ln' then AddStack(TPegtopFormulaLn.Create(Negative), Level)
  else if (S = 'lg') or (S = 'log') or (S = 'log10') then AddStack(TPegtopFormulaLog10.Create(Negative), Level)
  else if S = 'log2' then AddStack(TPegtopFormulaLog2.Create(Negative), Level)
  else if S = 'exp' then AddStack(TPegtopFormulaExp.Create(Negative), Level)
  else if (S = 'sgn') or (S = 'sign') then AddStack(TPegtopFormulaSign.Create(Negative), Level)
  else if S = 'pi' then begin
    AddConstant(Pi, Negative);
    Result := True;
  end
  else begin
    I := FParameters.IndexOf(Specifier);
    if I < 0 then raise EPegtopFormulaBadSpecifier.CreateProp(Specifier);
    FVariable := TPegtopFormulaVariable.Create(I * 8, Negative);
    Result := True;
  end;
end;

procedure TPegtopFormulaCompiler.AddConstant(Value: Double; Negative: Boolean);
var
  Adr: Integer;
begin
  if Negative then Adr := FCode.Data.AddDouble(-Value)
  else Adr := FCode.Data.AddDouble(Value);
  FVariable := TPegtopFormulaVariable.Create(Adr, False);
end;

procedure TPegtopFormulaCompiler.AddOperator(C: Char; Level: Integer);
var
  Op: TPegtopFormulaOperator;
begin
  case C of
    '+':           Op := opAdd;
    '-':           Op := opSub;
    '*', '':      Op := opMul;
    '/', ':', '': Op := opDiv;
    '%':           Op := opMod;
    '^':           Op := opPow;
    else raise EPegtopFormulaParse.CreateFmt('Error parsing formula: %s is not a valid operator', [C]);
  end;
  FVariable.Operator := Op;
  AddStack(FVariable, Level);
end;

procedure TPegtopFormulaCompiler.AddStack(V: TPegtopFormulaVariable; Lev: Integer);
var
  Element: TPegtopFormulaElement;
  Prio: Integer;
  ResultVariable: TPegtopFormulaVariable;
  Adr: Integer;
begin
  V.Level := Lev;
  Prio := V.Prio;
  if (V.Operator = opNone) // new variable is last element of formula
  or ((FStack <> NIL) and (FStack.Prio >= Prio)) // or last element of subformula (like a*b+...)
  then begin
    // begin compilation with new variable...
    FCode.CodeFLD(V.DataAdr);
    if V.Negative then FCode.CodeFCHS;
    // ... and remove elements from stack and compile them:
    while (FStack <> NIL) and (FStack.Prio >= Prio) do begin
      Element := FStack;
      Element.Compile(FCode);
      FStack := Element.Prev;
      Element.Free;
    end;
    if (V.Operator <> opNone) then begin
      // we need the result of this subformula later, so...
      Adr := FCode.Data.AddDouble(0.0); // ...reserve space...
      FCode.CodeFSTP(Adr); // ...and store result
      // add result to stack:
      ResultVariable := TPegtopFormulaVariable.Create(Adr, False);
      ResultVariable.Operator := V.Operator;
      ResultVariable.Level := Lev;
      ResultVariable.Prev := FStack;
      FStack := ResultVariable;
    end;
    V.Free;
  end
  else begin
    // add new variable to stack:
    V.Prev := FStack;
    FStack := V;
  end;
end;

procedure TPegtopFormulaCompiler.AddStack(F: TPegtopFormulaFunction; Lev: Integer);
begin
  F.Level := Lev;
  F.Prev := FStack;
  FStack := F;
end;

procedure TPegtopFormulaCompiler.Parse(Expression: String;
  AutoCloseBrackets: Boolean = True;
  NationalDecimalSeparator: Boolean = False);
type
  TParseState = (psOperator, psOperand, psSpecifier, psConstant, psDecimals, psNegative);
var
  Decimal: Double;
  I, L: Integer;
  State: TParseState;
  Level: Integer;
  Specifier: String;                         // current specifier
  Constant: Double;                          // current constant
  Negative: Boolean;                         // value < 0 (minus found)
  C: Char;
  CClass: TCharClass;
  CharDecSep: Char;
begin
  FCode.CodeFLDCW(FCode.FPUControlWordAddr); // load default control word
  if NationalDecimalSeparator then CharDecSep := DecimalSeparator
  else CharDecSep := '.';
  FStack := NIL;
  Level := 0;
  // we begin with state psOperator, because we expect an operand
  // (specifier, constant) first (just like after an operator):
  State := psOperator;
  L := Length(Expression);
  I := 1;
  Constant := 0; // not necessary, but to avoid compiler warnings
  Decimal := 0; // not necessary, but to avoid compiler warnings
  Negative := False; // not necessary, but to avoid compiler warnings
  while I <= L do begin

    C := Expression[I];
    if C = CharDecSep then CClass := ccDecSep
    else CClass := CharClass[C];

    case State of
      psOperator:
        if CClass = ccSpace then begin
          // ignore
        end
        else if CClass = ccDigit then begin
          Negative := False;
          State := psConstant;
          Constant := Ord(C)-48;
        end
        else if CClass = ccAlpha then begin
          Negative := False;
          State := psSpecifier;
          Specifier := C;
        end
        else if CClass = ccDecSep then begin
          Negative := False;
          State := psDecimals;
          Decimal := 1.0;
          Constant := 0.0;
        end
        else if CClass = ccOperator then begin
          if C = '-' then State := psNegative
          else raise EPegtopFormulaParse.CreateProp(Expression, I, 'expression');
        end
        else if CClass = ccBraOpen then begin
          Inc(Level);
        end
        else begin // ccBraClose / unknown character
          raise EPegtopFormulaParse.CreateProp(Expression, I, 'expression');
        end;
      psOperand:
        if CClass = ccSpace then begin
          // ignore
        end
        else if CClass = ccOperator then begin
          AddOperator(C, Level);
          State := psOperator;
        end
        else if CClass = ccBraClose then begin
          if Level > 0 then Dec(Level)
          else raise EPegtopFormulaParse.CreateProp(Expression, I, 'operator');
        end
        else begin // ccDigit / ccAlpha / ccDecSep / ccBraOpen / unknown character
          raise EPegtopFormulaParse.CreateProp(Expression, I, 'operator')
        end;
      psSpecifier:
        if CClass = ccSpace then begin
          if AddSpecifier(Specifier, Negative, Level) then State := psOperand // wait for operator
          else State := psOperator; // wait for operand
        end
        else if CClass in [ccAlpha, ccDigit] then begin
          Specifier := Specifier + C;
        end
        else if CClass = ccOperator then begin
          if AddSpecifier(Specifier, Negative, Level) then begin
            AddOperator(C, Level);
            State := psOperator;
          end
          else begin
            raise EPegtopFormulaParse.CreateProp(Expression, I, '"("');
          end;
        end
        else if CClass = ccBraOpen then begin
          if AddSpecifier(Specifier, Negative, Level) then begin
            raise EPegtopFormulaParse.CreateProp(Expression, I, 'alphanumeric character or operator');
          end
          else begin
            Inc(Level);
            State := psOperator; // wait for operand
          end;
        end
        else if CClass = ccBraClose then begin
          if AddSpecifier(Specifier, Negative, Level) then begin
            if Level > 0 then Dec(Level)
            else raise EPegtopFormulaParse.CreateProp(Expression, I, 'alphanumeric character or operator');
            State := psOperand; // wait for operator
          end
          else begin
            raise EPegtopFormulaParse.CreateProp(Expression, I, '"("');
          end;
        end
        else begin // ccDecSep / unknown character
          raise EPegtopFormulaParse.CreateProp(Expression, I, 'alphanumeric character or operator');
        end;
      psConstant:
        if CClass = ccSpace then begin
          AddConstant(Constant, Negative);
          State := psOperand; // wait for operator
        end
        else if CClass = ccDigit then begin
          Constant := Constant * 10 + (Ord(C)-48);
        end
        {else if CClass in ['E', 'e']  then begin
          // not supported yet
        end}
        else if CClass = ccDecSep then begin
          State := psDecimals;
          Decimal := 1.0;
        end
        else if CClass = ccOperator then begin
          AddConstant(Constant, Negative);
          AddOperator(C, Level);
          State := psOperator;
        end
        else if CClass = ccBraClose then begin
          AddConstant(Constant, Negative);
          if Level > 0 then Dec(Level)
          else raise EPegtopFormulaParse.CreateProp(Expression, I, 'digit, operator or decimal separator');
          State := psOperand; // wait for operator
        end
        else begin // ccAlpha / ccBraOpen / unknown character
          raise EPegtopFormulaParse.CreateProp(Expression, I, 'digit, operator or decimal separator');
        end;
      psDecimals:
        if CClass = ccSpace then begin
          AddConstant(Constant, Negative);
          State := psOperand; // wait for operator
        end
        else if CClass = ccDigit then begin
          Decimal := Decimal * 0.1;
          Constant := Constant + (Ord(C)-48) * Decimal;
        end
        else if CClass = ccOperator then begin
          AddConstant(Constant, Negative);
          AddOperator(C, Level);
          State := psOperator;
        end
        else if CClass = ccBraClose then begin
          AddConstant(Constant, Negative);
          if Level > 0 then Dec(Level)
          else raise EPegtopFormulaParse.CreateProp(Expression, I, 'digit or operator');
          State := psOperand; // wait for operator
        end
        else begin // ccAlpha / ccDecSep / ccBraOpen / unknown character
          raise EPegtopFormulaParse.CreateProp(Expression, I, 'digit or operator');
        end;
      psNegative:
        if CClass = ccSpace then begin
          // ignore
        end
        else if CClass = ccDigit then begin
          Negative := True;
          State := psConstant;
          Constant := Ord(C)-48;
        end
        else if CClass = ccAlpha then begin
          Negative := True;
          State := psSpecifier;
          Specifier := C;
        end
        else if CClass = ccDecSep then begin
          Negative := True;
          State := psDecimals;
          Decimal := 1.0;
          Constant := 0.0;
        end
        else if CClass = ccBraOpen then begin
          AddStack(TPegtopFormulaNeg.Create(False), Level);
          Inc(Level);
          State := psOperator; // wait for operand
        end
        else begin // ccOperator / ccBraClose / unknown character
          raise EPegtopFormulaParse.CreateProp(Expression, I, 'expression');
        end;
    end;

    Inc(I);
  end;

  case State of
    psOperator: raise EPegtopFormulaParse.CreateProp(Expression, I, 'expression');
    psOperand: begin
      // everything fine (operand already added)
    end;
    psSpecifier: begin
      if not AddSpecifier(Specifier, Negative, Level) then begin
        raise EPegtopFormulaParse.CreateProp(Expression, I, '"("');
      end;
    end;
    psConstant, psDecimals: begin
      AddConstant(Constant, Negative);
    end;
    psNegative: raise EPegtopFormulaParse.CreateProp(Expression, I, 'expression');
  end;

  AddStack(FVariable, -1); // add last operand (without operator)

  // all brackets closed?
  if (Level <> 0) and (not AutoCloseBrackets) then
    raise EPegtopFormulaParse.CreateProp(Expression, I, '")"');

  FCode.CodeRET;
end;

procedure TPegtopFormulaCompiler.CleanUp;
begin
  FCode.CleanUp;
end;

////////////////////////////////////////////////////////////////////////////////
// initialization
////////////////////////////////////////////////////////////////////////////////

var
  C: Char;
initialization
  for C := #0 to #255 do CharClass[C] := ccUnknown;

  // spaces:
  CharClass[' '] := ccSpace;
  CharClass[#9] := ccSpace; // tab
  CharClass[#10] := ccSpace; // lf
  CharClass[#13] := ccSpace; // cr

  // alphanumeric:
  for C := '0' to '9' do CharClass[C] := ccDigit;
  for C := 'A' to 'Z' do CharClass[C] := ccAlpha;
  for C := 'a' to 'z' do CharClass[C] := ccAlpha;
  CharClass['_'] := ccAlpha;

  // operators:
  CharClass['+'] := ccOperator;
  CharClass['-'] := ccOperator;
  CharClass['*'] := ccOperator;
  CharClass[''] := ccOperator;
  CharClass['/'] := ccOperator;
  CharClass[':'] := ccOperator;
  CharClass[''] := ccOperator;
  CharClass['%'] := ccOperator;
  CharClass['^'] := ccOperator;

  // brackets:
  CharClass['('] := ccBraOpen;
  CharClass[')'] := ccBraClose;
  CharClass['['] := ccBraOpen;
  CharClass[']'] := ccBraClose;
  CharClass['{'] := ccBraOpen;
  CharClass['}'] := ccBraClose;
end.

