// @abstract(The @name unit declares @Link(TRbwParser) along with
// associated classes and types.)  

unit RbwParser;

interface

uses
{$IFDEF MSWINDOWS}
  // Indcluding Windows, allows AnsiCompareStr to be inlined with Delphi 2005.
  Windows,
{$ENDIF}
  Types, SysUtils, Classes, Contnrs;

type
  {
    @Name is the prototype for functions in @link(TRbwParser) that
    return a real number.  To use a @link(TRbwRealFunction), assign it to a
    @link(TFunctionRecord.RFunctionAddr), assign the remainder of the fields
    in the @link(TFunctionRecord) and then call @link(TFunctionStringList.Add)
    of @Link(TRbwParser.Functions).
    Be sure that the @link(TFunctionRecord.ResultType)
    is rdtDouble when assigning a @link(TRbwRealFunction).
    Failure to do so will
    lead to access violations or invalid results.
  }
  TRbwRealFunction = function(Values: array of pointer): double;
  {
    @Name is the prototype for functions in TRbwParser that
    return an integer.  To use a TRbwIntegerFunction, assign it to the
    IFunctionAddr of a TFunctionRecord, assign the remainder of the fields
    in the TFunctionRecord and then call the Add function of
    TRbwParser.Functions.  Be sure that the ResultType of the TFunctionRecord
    is rdtInteger when assigning a TRbwIntegerFunction.  Failure to do so will
    lead to access violations or invalid results.
  }
  TRbwIntegerFunction = function(Values: array of pointer): Integer;
  {
    @Name is the prototype for functions in TRbwParser that
    return a boolean.  To use a TRbwBooleanFunction, assign it to the
    BFunctionAddr of a TFunctionRecord, assign the remainder of the fields
    in the TFunctionRecord and then call the Add function of
    TRbwParser.Functions.  Be sure that the ResultType of the TFunctionRecord
    is rdtBoolean when assigning a TRbwBooleanFunction.  Failure to do so will
    lead to access violations or invalid results.
  }
  TRbwBooleanFunction = function(Values: array of pointer): Boolean;
  {
    @Name is the prototype for functions in TRbwParser that
    return a string.  To use a TRbwStringFunction, assign it to the
    SFunctionAddr of a TFunctionRecord, assign the remainder of the fields
    in the TFunctionRecord and then call the Add function of
    TRbwParser.Functions.  Be sure that the ResultType of the TFunctionRecord
    is rdtString when assigning a TRbwStringFunction.  Failure to do so will
    lead to access violations or invalid results.
  }
  TRbwStringFunction = function(Values: array of pointer): string;

  {
    @Name is used to define the data type passed into or returned by
    a function.  In the case of rdtString used to define the data passed into
    a function, the data are passed into the function as a PString.
    Data are
    passed into a function as pointers to variables of the correct
    type. Results of the functions are values of the correct type rather
    than pointers.
  }
  TRbwDataType = (rdtDouble, rdtInteger, rdtBoolean, rdtString);

  {@name is defined for convenience here but is not used by @link(TRbwParser).}
  TRbwDataTypes = set of TRbwDataType;

  // @name is a pointer to a @link(TFunctionRecord).
  PFunctionRecord = ^TFunctionRecord;
  {

    @abstract(A @Name is used to define a function that can be used in a
    TRbwParser.  To use the function, you must first assign the fields
    of the TFunctionRecord and then call TRbwParser.Functions.Add.)
    

    @longcode(#
  TFunctionRecord = record
    InputDataTypes: array of TRbwDataType;
    OptionalArguments: integer;
    CanConvertToConstant: boolean;
    Name: string;
    Prototype: string;
    Hidden: boolean;
    case ResultType: TRbwDataType of
      rdtDouble: (RFunctionAddr: TRbwRealFunction);
      rdtInteger: (IFunctionAddr: TRbwIntegerFunction);
      rdtBoolean: (BFunctionAddr: TRbwBooleanFunction);
      rdtString: (SFunctionAddr: TRbwStringFunction);
  end;
    #)
  }
  TFunctionRecord = record
    {
      @Name defines the type of data passed into the function
      assigned to RFunctionAddr, IFunctionAddr, BFunctionAddr, or
      SFunctionAddr.  The data are passed in as pointers to variables
      of the correct type.

      The length of @Name normally defines the maximum number of
      arguments that can be passed into the function.  However, if
      OptionalArguments is less than 0, an unlimited number of arguments can
      be passed to the function and the types of all those arguments must
      match the type of the last member of InputDataTypes;  In that case,
      the minimum number of required arguments is the length of InputDataTypes
      minus 1.

      See also : @link(TRbwDataType);
    }
    InputDataTypes: array of TRbwDataType;
    {
      If @Name is greater than 0, some of the arguments passed to
      the function may be nil pointers.  The nil pointers will always follow
      the non-nil pointers; no non-nil pointer may follow a nil pointer.
      The maximum number of nil pointers will be @Name.

      If @Name  is less than 0, the number of arguments passed to
      the function may be greater than the length of @link(InputDataTypes).
      The minimum number of arguments is the length of @link(InputDataTypes)
      minus 1.
      If more arguments are used, the types of those arguments must correspond
      to the value defined in the last member of InputDataTypes.
    }
    OptionalArguments: integer;
    {
      @Name defines whether the result of a function may be
      considered a constant value if all of the values passed to the function
      in the values array are constants.

      Normally @Name should be set to True but if the function
      makes reference to global variables that may change between one
      evaluation of the expression and the next, @Name should
      be set to False.

      Pi is an example of a function for which CanConvertToConstant should
      be true.

      @Name is used when optimizing compiled expressions.

      Example:

      @LongCode(#
        var
          // ... others omitted in example.
          PiFunction : TFunctionRecord;
          SinFunction : TFunctionRecord;
      #)

         Pi is a constant so CanConvertToConstant is true.  Pi has no
         input variables so it will always be converted to a constant.
         When PiFunction is used to create a TFunctionClass,
         CanConvertToConstant will set the value of the
         TFunctionClass.AllowConversionToConstant property.
         That in turn will set the value of
         TExpression.AllowConversionToConstant.  The same happens with
         RFunctionAddr and InputDataTypes.

         _Pi is a function defined in the implementation

      @LongCode(#
        function _Pi(Values : array of pointer) : double;
        begin
          result := Pi;
        end;
      #)

      Sin will always return the same value if it's input is a constant
      so CanConvertToConstant is set true.  However, it will only be converted
      to a constant if all of it's input variables are constants.

      _Sin is a function defined in the implementation

      @LongCode(#
      function _Sin(Values : array of pointer) : double;
      begin
        result := Sin(PDouble(Values[0])^);
      end;
      #)

      @LongCode(#
      constructor TFunctionStringList.Create;
      begin
        inherited;
        CaseSensitive := False;
        Duplicates := dupError;
        Sorted := True;

        // ... others lines omitted in example.

        PiFunction.ResultType := rdtDouble;
        PiFunction.Name := 'Pi';
        SetLength(PiFunction.InputDataTypes, 0);
        PiFunction.OptionalArguments := 0;
        PiFunction.CanConvertToConstant := True;
        PiFunction.RFunctionAddr := _Pi;
        Add(PiFunction);

        // ... others lines omitted in example.

        PiFunction.ResultType := rdtDouble;
        PiFunction.Name := 'Pi';
        SetLength(PiFunction.InputDataTypes, 0);
        PiFunction.OptionalArguments := 0;
        PiFunction.CanConvertToConstant := True;
        PiFunction.RFunctionAddr := _Pi;
        Add(PiFunction);

        // ... others lines omitted in example.

      end;
      #)
    }
    CanConvertToConstant: boolean;
    {
      @Name is the unique identifier of the function.
    }
    Name: string;
    {
      @Name gives the names and arguments for a function as they
      should be shown on a GUI interface.
    }
    Prototype: string;
    {
      @Name determines whether or not the function should be
      visible in a GUI interface.
    }
    Hidden: boolean;
    { @name represents alternative valid spellings for the same function. 
    }
    Synonyms: array of string;
    {
     @Name defines the type of data returned by the function.
     It must correspond to the value assigned to the RFunctionAddr,
     IFunctionAddr, BFunctionAddr, or SFunctionAddr.
    }
    case ResultType: TRbwDataType of
      {
        @Name is the address of the @Link(TRbwRealFunction) that is to be
        evaluated.
      }
      rdtDouble: (RFunctionAddr: TRbwRealFunction);
      {
        @Name is the address of the @Link(TRbwIntegerFunction) that is to be
        evaluated.
      }
      rdtInteger: (IFunctionAddr: TRbwIntegerFunction);
      {
        @Name is the address of the @Link(TRbwBooleanFunction) that is to be
        evaluated.
      }
      rdtBoolean: (BFunctionAddr: TRbwBooleanFunction);
      {
        @Name is the address of the @Link(TRbwStringFunction) that is to be
        evaluated.
      }
      rdtString: (SFunctionAddr: TRbwStringFunction);
  end;

  // @name is a pointer to a boolean.
  PBoolean = ^Boolean;

  {
    @abstract(@Name are used in the protected section of @Link(TExpression).)
    @longcode(#
  TVariables = record
    Datum: Pointer;
    DataType: TRbwDataType;
  end;
  #)
  }
  TVariables = record
    {
      @Name is a pointer to a value to be passed to a function.
    }
    Datum: Pointer;
    {
      @Name defines the type of data to which @Link(Datum) points.
    }
    DataType: TRbwDataType;
  end;

  {
    @abstract(@Name is raised if by @Link(TRbwParser) to report user errors
    and some programmer errors.)

    (Assertion failures represent programmer errors.)
  }
  ERbwParserError = class(Exception)
    // @name: integer;
    ErrorType: integer;
    // sets @link(ErrorType) to ErrorMode.
    constructor CreateMode(const Msg: string; const ErrorMode: integer);
  end;

  {
    @abstract(@Name defines a function that can be called by @Link(TRbwParser).)
  }
  TFunctionClass = class(TObject)
  private
    // @Name: @link(TFunctionRecord);
    // @name stores the data represented by many of the public properties.
    FunctionRecord: TFunctionRecord;
    // @Name: string;
    // See @link(Prototype).
    FPrototype: string;
    // See @link(AllowConversionToConstant).
    function GetAllowConversionToConstant: boolean;
    // Gets value of BFunctionAddr property
    function GetBFunctionAddr: TrbwBooleanFunction;
    // See @link(Hidden).
    function GetHidden: boolean;
    // Gets value of IFunctionAddr property
    function GetIFunctionAddr: TrbwIntegerFunction;
    // See @link(InputDataCount).
    function GetInputDataCount: integer;
    // See @link(InputDataTypes).
    function GetInputDataTypes(const Index: integer): TRbwDataType;
    // See @link(Name).
    function GetName: string;
    // See @link(OptionalArguments).
    function GetOptionalArguments: integer;
    // See @link(ResultType).
    function GetResultType: TRbwDataType;
    // Gets value of RFunctionAddr property
    function GetRFunctionAddr: TrbwRealFunction;
    // Gets value of SFunctionAddr property
    function GetSFunctionAddr: TrbwStringFunction;
    // See @link(AllowConversionToConstant).
    procedure SetAllowConversionToConstant(const Value: boolean);
    // Sets value of BFunctionAddr property
    procedure SetBFunctionAddr(const Value: TrbwBooleanFunction);
    // See @link(Hidden).
    procedure SetHidden(const Value: boolean);
    // Sets value of IFunctionAddr property
    procedure SetIFunctionAddr(const Value: TrbwIntegerFunction);
    // See @link(InputDataCount).
    procedure SetInputDataCount(const Value: integer);
    // See @link(InputDataTypes).
    procedure SetInputDataTypes(const Index: integer;
      const Value: TRbwDataType);
    // See @link(Name).
    procedure SetName(const Value: string);
    // See @link(OptionalArguments).
    procedure SetOptionalArguments(const Value: integer);
    // Sets value of RFunctionAddr property
    procedure SetRFunctionAddr(const Value: TrbwRealFunction);
    // Sets value of SFunctionAddr property
    procedure SetSFunctionAddr(const Value: TrbwStringFunction);
  public
    {
      @Name defines whether the result of a function may be
      considered a constant value if all of the values passed to the function
      in the values array are constants.

      Normally @Name should be set to True but if the
      function makes reference to global variables that may change between one
      evaluation of the expression and the next, @Name
      should be set to False.

      Pi is an example of a function for which @Name should
      be true.

      @Name is used when optimizing compiled expressions.
    }
    property AllowConversionToConstant: boolean
      read GetAllowConversionToConstant write SetAllowConversionToConstant;
    {
      @Name calls the inherited @Name and sets
      @Link(AllowConversionToConstant) to true.

      Users should generally not call @Name directly but instead create
      a @Link(TFunctionClass) by calling
      TFunctionStringList.@Link(TFunctionStringList.Add).
    }
    constructor Create;
    {
      @Name is the address of a @Link(TRbwBooleanFunction) assigned to
      the @Link(TFunctionClass).  If @Link(ResultType) is of the wrong type,
      reading @Name will raise an exception. Writing @Name sets
      @Link(ResultType).
    }
    property BFunctionAddr: TRbwBooleanFunction read GetBFunctionAddr
      write SetBFunctionAddr;
    {
      @Name is the address of a @Link(TRbwIntegerFunction) assigned to
      the @Link(TFunctionClass).  If @Link(ResultType) is of the wrong type,
      reading @Name will raise an exception. Writing @Name sets
      @Link(ResultType).
    }
    property IFunctionAddr: TRbwIntegerFunction read GetIFunctionAddr
      write SetIFunctionAddr;
    {
      @Name normally defines the maximum number of
      arguments that can be passed into the function.  However, if
      @Link(OptionalArguments) is less than 0, an unlimited number of arguments
      can be passed to the function and the types of all those arguments must
      match the type of the last member of @Link(InputDataTypes).  In that case,
      the minimum number of required arguments is @Name
      minus 1.
    }
    property InputDataCount: integer read GetInputDataCount
      write SetInputDataCount;
    {
      @Name is used to define the data types passed into or returned
      by a function.  Data are
      passed into a function as pointers to variables of the correct
      type. Results of the functions are values of the correct type rather
      than pointers.
    }
    property InputDataTypes[const Index: integer]: TRbwDataType
    read GetInputDataTypes write SetInputDataTypes;
    {
      @Name defines the name of the function.  The @Name of each function and
      variable in a @Link(TRbwParser) must be unique.
    }
    property Name: string read GetName write SetName;
    {
      @Name gives suggestions as to how the names and arguments
      for a function could be shown on a GUI interface.  The "|"
      character is used in the prototype to classify functions.  Portions
      of the prototype that appear before a "|" character is the
      classification to which the rest of the prototype belongs.
      Thus, the portions before the "|" could be used to populate a
      tree control.
    }
    property Prototype: string read FPrototype write FPrototype;
    {
      If @Name is greater than 0, up to that number of nil
      pointers may be passed to the function when it is evaluated.  All
      the nil pointers must follow all the non-nil pointers.

      If @Name is less than 0, an unlimited number of arguments
      may be passed to the function.  The data type of these pointers will
      correspond to the data type defined in the last member of
      @Link(InputDataTypes).
    }
    property OptionalArguments: integer read GetOptionalArguments
      write SetOptionalArguments;
    {
      @Name is used to define the data type passed returned by
      a function.  The results of the functions are values of the correct
      type rather than pointers.

      Example:

      Because OptionalArguments is less than 0, the CaseB, CaseI, CaseR, and
      CaseS functions can take an unlimited number of arguments.

      @longcode(#
      function _CaseBoolean(Values : array of pointer) : boolean;
      begin
        result := PBoolean(Values[PInteger(Values[0])^])^;
      end;

      function _CaseInteger(Values : array of pointer) : integer;
      begin
        result := PInteger(Values[PInteger(Values[0])^])^;
      end;

      function _CaseDouble(Values : array of pointer) : double;
      begin
        result := PDouble(Values[PInteger(Values[0])^])^;
      end;

      function _CaseString(Values : array of pointer) : String;
      begin
        result := PString(Values[PInteger(Values[0])^])^;
      end;

      var
        CaseBooleanFunction : TFunctionRecord;
        CaseIntegerFunction : TFunctionRecord;
        CaseDoubleFunction : TFunctionRecord;
        CaseStringFunction : TFunctionRecord;

      constructor TFunctionStringList.Create;
      begin
        inherited;
        CaseSensitive := False;
        Duplicates := dupError;
        Sorted := True;

        // ... others lines omitted in example.

        CaseBooleanFunction.ResultType := rdtBoolean;
        CaseBooleanFunction.Name := 'CaseB';
        SetLength(CaseBooleanFunction.InputDataTypes, 4);

        CaseBooleanFunction.InputDataTypes[0] := rdtInteger;
        CaseBooleanFunction.InputDataTypes[1] := rdtBoolean;
        CaseBooleanFunction.InputDataTypes[2] := rdtBoolean;
        CaseBooleanFunction.InputDataTypes[3] := rdtBoolean;
        CaseBooleanFunction.CanConvertToConstant := True;
        CaseBooleanFunction.OptionalArguments := -1;
        CaseBooleanFunction.BFunctionAddr := _CaseBoolean;
        Add(CaseBooleanFunction);

        CaseIntegerFunction.ResultType := rdtInteger;
        CaseIntegerFunction.Name := 'CaseI';
        SetLength(CaseIntegerFunction.InputDataTypes, 4);

        CaseIntegerFunction.InputDataTypes[0] := rdtInteger;
        CaseIntegerFunction.InputDataTypes[1] := rdtInteger;
        CaseIntegerFunction.InputDataTypes[2] := rdtInteger;
        CaseIntegerFunction.InputDataTypes[3] := rdtInteger;
        CaseIntegerFunction.OptionalArguments := -1;
        CaseIntegerFunction.CanConvertToConstant := True;
        CaseIntegerFunction.IFunctionAddr := _CaseInteger;
        Add(CaseIntegerFunction);

        CaseDoubleFunction.ResultType := rdtDouble;
        CaseDoubleFunction.Name := 'CaseR';
        SetLength(CaseDoubleFunction.InputDataTypes, 4);
        CaseDoubleFunction.InputDataTypes[0] := rdtInteger;
        CaseDoubleFunction.InputDataTypes[1] := rdtDouble;
        CaseDoubleFunction.InputDataTypes[2] := rdtDouble;
        CaseDoubleFunction.InputDataTypes[3] := rdtDouble;
        CaseDoubleFunction.OptionalArguments := -1;

        CaseDoubleFunction.CanConvertToConstant := True;
        CaseDoubleFunction.RFunctionAddr := _CaseDouble;
        Add(CaseDoubleFunction);

        CaseStringFunction.ResultType := rdtString;
        CaseStringFunction.Name := 'CaseS';
        SetLength(CaseStringFunction.InputDataTypes, 4);
        CaseStringFunction.InputDataTypes[0] := rdtInteger;
        CaseStringFunction.InputDataTypes[1] := rdtString;
        CaseStringFunction.InputDataTypes[2] := rdtString;
        CaseStringFunction.InputDataTypes[3] := rdtString;
        CaseStringFunction.OptionalArguments := -1;
        CaseStringFunction.CanConvertToConstant := True;
        CaseStringFunction.SFunctionAddr := _CaseString;
        Add(CaseStringFunction);
        // ... others lines omitted in example.
      end;

      #)
    }
    property ResultType: TRbwDataType read GetResultType;
    {
      @Name is the address of a @Link(TRbwRealFunction) assigned to the
      @Link(TFunctionClass).  If @Link(ResultType) is of the wrong type, reading
      @Name will raise an exception. Writing @Name sets
      @Link(ResultType).
    }
    property RFunctionAddr: TRbwRealFunction read GetRFunctionAddr
      write SetRFunctionAddr;
    {
      @Name is the address of a @Link(TRbwStringFunction) assigned to the
      @Link(TFunctionClass).  If @Link(ResultType) is of the wrong type, reading
      @Name will raise an exception. Writing @Name sets
      @Link(ResultType).
    }
    property SFunctionAddr: TRbwStringFunction read GetSFunctionAddr
      write SetSFunctionAddr;
    {
      @Name has no effect.  It is intended to be used to indicate whether or
      not the function specified by the @Link(TFunctionClass) should be visible
      to the user.
    }
    property Hidden: boolean read GetHidden write SetHidden;
  end;

  {
    @abstract(@Name maintains a sorted list of @Link(TFunctionClass)es.)

    In the functions below, except where otherwise noted,
    angles are expressed in radians.

    The following functions are included:

    @bold(AbsI)(Value) returns the absolute value of Value.
    Value must be an integer.
    The value returned by AbsI will be an integer.

    @bold(AbsR)(Value) returns the absolute value of Value.
    Value can be either an integer or a real number.
    The value returned by AbsR will be a real number.

    @bold(ArcCos)(Value) returns the inverse cosine of Value.
    The return value is in the range from zero to Pi.

    @bold(ArcCosh)(Value) returns the inverse hyperbolic cosine of Value.

    @bold(ArcSin)(Value) returns the inverse sine of Value.
    The return value is in the range from -Pi/2 to +Pi/2.

    @bold(ArcSinh)(Value) returns the inverse hyperbolic sine of Value.

    @bold(ArcTan2)(Y, X) returns the inverse tangent of Y/X in the correct quadrant.
    The return value is in the range from -Pi to +Pi.

    @bold(ArcTanh)(Value) returns the inverse hyperbolic tangent of Value.

    @bold(CaseB)(Index, boolean_Result1, boolean_Result2, ...).
    CaseB uses Index to determine which of the boolean_Result1,
    boolean_Result2, ... arguments will be returned as a result. 
    If Index equals 1, boolean_Result1 is returned; if Index
    equals 2, boolean_Result2 is returned; if Index equals 3,
    boolean_Result3 is returned; and so forth. Only "Index", constant
    expressions and the result that is returned will be evaluated.
    The types of boolean_Result1, boolean_Result2, ...
    must all be booleans. The type that is returned will be a
    boolean.

    @bold(CaseI)(Index, integer_Result1, integer_Result2, ...).
    CaseI uses Index to determine which of the integer_Result1,
    integer_Result2, ... arguments will be returned as a result. 
    If Index equals 1, integer_Result1 is returned; if Index
    equals 2, integer_Result2 is returned; if Index equals 3,
    integer_Result3 is returned; and so forth. Only "Index", constant
    expressions and the result that is returned will be evaluated.
    The types of integer_Result1, integer_Result2, ...
    must all be integers. The type that is returned will be an
    integer.

    @bold(CaseR)(Index, real_Result1, real_Result2, ...).
    CaseR uses Index to determine which of the real_Result1,
    real_Result2, ... arguments will be returned as a result. 
    If Index equals 1, real_Result1 is returned; if Index equals 2,
    real_Result2 is returned; if Index equals 3,
    real_Result3 is returned; and so forth. Only "Index", constant
    expressions and the result that is returned will be evaluated.
    The types of real_Result1, real_Result2, ...
    must all be real numbers. The type that is returned will be a
    real number.

    @bold(CaseT)(Index, text_Result1, text_Result2, ...).
    CaseT uses Index to determine which of the text_Result1,
    text_Result2, ... arguments will be returned as a result.
    If Index equals 1, text_Result1 is returned;
    if Index equals 2, text_Result2 is returned;
    if Index equals 3, text_Result3 is returned; and so forth.
    Only "Index", constant expressions and the result that is
    returned will be evaluated. The types of text_Result1,
    text_Result2, ... must all be character strings.
    The type that is returned will be a character string.
    CaseS is a synonym for CaseT.

    @bold(Copy)(text_Value, StartIndex, Count) returns a substring
    of text_Value starting at the character indicated by
    StartIndex and extending for either Count characters or
    until the end text_Value is reached whichever is smaller.

    @bold(Cos)(Value) returns the cosine of Value.

    @bold(Cosh)(Value) returns the hyperbolic cosine of Value.

    @bold(DegToRad)(Value) converts Value from degrees to radians.
    See also RadToDeg.

    @bold(Distance)(X1, Y1, X2, Y2) calculates the distance between
    points (X1,Y1) and (X2, Y2).

    @bold(FactorialI)(Value_Less_than_13) returns the factorial of
    Value_Less_than_13 as an integer.

    @bold(FactorialR)(Value_Less_than_171) returns the factorial of
    Value_Less_than_171 as a real number.

    @bold(FloatToText)(Value) converts the real number Value to
    its string representation.
    FloatToStr is a synonym for FloatToText.

    @bold(Frac)(Value) returns the fractional part of Value.

    @bold(IfB)(Boolean_Value, If_True_boolean_Result,
    If_False_boolean_Result).
    IfB uses Boolean_Value to determine whether
    If_True_boolean_Result or If_False_boolean_Result is returned
    as a result. If Boolean_Value is true, If_True_boolean_Result is
    returned;
    if Boolean_Value is false, If_False_boolean_Result is returned.
    Only "Boolean_Value", constant expressions and the result
    that is returned will be evaluated. The types of
    If_True_boolean_Result and If_False_boolean_Result
    must be boolean. The type that is returned will be a boolean.

    @bold(IfI)(Boolean_Value, If_True_integer_Result,
    If_False_integer_Result).
    IfI uses Boolean_Value to determine whether
    If_True_integer_Result or If_False_integer_Result is returned
    as a result.  If Boolean_Value is true, If_True_integer_Result
    is returned; if Boolean_Value is false, If_False_integer_Result
    is returned.  Only "Boolean_Value", constant expressions and
    the result that is returned will be evaluated.  The types of
    If_True_integer_Result and If_False_integer_Result must be integers.  The type that is returned will be an integer.

    @bold(IfR)(Boolean_Value, If_True_real_Result, If_False_real_Result).
    IfR uses Boolean_Value to determine whether If_True_real_Result
    or If_False_real_Result is returned as a result.
    If Boolean_Value is true, If_True_real_Result is returned;
    if Boolean_Value is false, If_False_real_Result is returned.
    Only "Boolean_Value", constant expressions and the result that
    is returned will be evaluated.  The types of
    If_True_real_Result and If_False_real_Result must be real
    numbers.  The type that is returned will be a real number.

    @bold(IfT)(Boolean_Value, If_True_text_Result, If_False_text_Result).
    IfT uses Boolean_Value to determine whether If_True_text_Result
    or If_False_text_Result is returned as a result.
    If Boolean_Value is true, If_True_text_Result is returned;
    if Boolean_Value is false, If_False_text_Result is returned.
    Only "Boolean_Value", constant expressions and the result
    that is returned will be evaluated.  The types of
    If_True_text_Result and If_False_text_Result must be
    character strings.  The type that is returned will be a
    character string.
    IfS is a synonym for IfT.

    @bold(Interpolate)(Position, Value1, Distance1, Value2, Distance2).
    Interpolate returns

    (Position-Distance1)/(Distance2-Distance1)*(Value2-Value1)+Value1.

    As its name implies, this is an interpolation between Value1 and
    Value2 based on where Position is between Distance1 and Distance2.
    See also @link(OverloadedFunctionList MultiInterpolate).

    @bold(IntPower)(Base, Exponent) returns Base raised to the Exponent
    power.  Base is a real number or integer.
    Exponent is an integer.  See also Power.

    @bold(IntToText)(Value) converts the integer number Value
    to its string representation.
    IntToStr is a synonym for IntToText.

    @bold(Length)(text_Value) returns the number of
    characters in text_Value.

    @bold(ln)(Value) returns the natural log of Value.

    @bold(log10)(Value) returns the log to the base 10 of Value.

    @bold(logN)(Base, Value) returns the log to the base N of Value.

    @bold(LowerCase)(text_Value) returns text_Value with all its
    characters converted to lower case.  See also: UpperCase.

    @bold(MaxI)(integer_Value1, integer_Value2, ...) returns whichever of its
    arguments is the largest.  Its arguments must be integers.
    The result will be an integer.

    @bold(MaxR)(real_Value1, real_Value2, ...) returns whichever of its
    arguments is the largest.
    Its arguments must be either integers or real numbers.
    The result will be a real number.

    @bold(MinI)(integer_Value1, integer_Value2, ...) returns whichever
    of its arguments is the smallest.
    Its arguments must be integers.  The result will be an integer.

    @bold(MinR)(real_Value1, real_Value2, ...) returns whichever of its
    arguments is the smallest.
    Its arguments must be either integers or real numbers.
    The result will be a real number.

    @bold(Odd)(Value) returns True if Value is an odd number.
    Otherwise it returns False.  Value must be an integer.

    @bold(Pi) returns the ratio of the circumference of a circle to its diameter.

    @bold(Pos)(SubString, StringValue) returns the position of the first
    instance of SubString within StringValue.
    If Substring does not occur within StringValue, Pos returns 0.

    @bold(PosEx)(SubText, TextValue, Offset) returns the position of the first
    instance of SubText within TextValue that starts on or after Offset.
    If SubText does not occur within TextValue, on or after Offset,
    PosEx returns 0.  If Offset equals one, PosEx is equivalent to Pos.

    @bold(Power)(Base, Exponent) returns Base raised to the Exponent power.
    Base and Exponent are real numbers or integers.
    See also IntPower.

    @bold(RadToDeg)(Value) converts Value from radians to degrees.
    See also DegToRad.

    @bold(Round)(Value) converts Value to the nearest integer.
    In the case of a number that is exactly halfway between two
    integers, it converts it to whichever one is even.
    See also Trunc.

    @bold(Sin)(Value) returns the sine of Value.

    @bold(Sinh)(Value) returns the hyperbolic sine of Value.

    @bold(SqrI)(integer_Value) returns integer_Value squared. Integer_Value
    must be an integer.  The result of SqrI will be an integer.

    @bold(SqrR)(real_Value) returns real_Value squared. Real_Value
    can be either an integer or a real number.
    The result of SqrR will be a real number.

    @bold(Sqrt)(Value) returns the square root of Value.

    @bold(TextToFloat)(text_Value) converts text_Value to a real number.
    If text_Value can not be converted, TextToFloat causes an error.
    StrToFloat is a synonym for TextToFloat.

    @bold(TextToFloatDef)(text_Value, DefaultResult) converts
    text_Value to a real number.
    If text_Value can not be converted, DefaultResult
    is returned instead.
    StrToFloatDef is a synonym for TextToFloatDef.

    @bold(TextToInt)(text_Value) converts text_Value to an integer.
    If text_Value can not be converted, TextToInt causes an error.
    StrToInt is a synonym for TextToInt.

    @bold(TextToIntDef)(text_Value, DefaultResult) converts text_Value
    to an integer.  If text_Value can not be converted,
    DefaultResult is returned instead.
    StrToIntDef is a synonym for TextToIntDef.

    @bold(Tan)(Value) returns the tangent of Value.

    @bold(Tanh)(Value) returns the hyperbolic tangent of Value.

    @bold(Trunc)(Value) truncates Value to an integer by rounding it towards zero.
    See also Round.

    @bold(UpperCase)(text_Value) returns text_Value with all its characters
    converted to upper case.  See also LowerCase.

  }
  TFunctionStringList = class(TStringList)
  private
    // See @link(FunctionClass).
    function GetFunctionClass(const Index: integer): TFunctionClass;
  public
    {
      @Name creates a @Link(TFunctionClass) based on FunctionRecord, calls the
      TStringList.AddObject method using the function name and the
      @Link(TFunctionClass) and returns the @Link(TFunctionClass) that it
      created.  Any @Link(TFunctionClass) created in this way is owned by the
      TFunctionStringList.
    }
    function Add(const FunctionRecord: TFunctionRecord): Integer; reintroduce;
    {
      @Name sets the number of items in the TFunctionStringList to 0 and Free's
      all the TFunctionClass it created.
    }
    procedure Clear; override;
    {
      @Name creates a TFunctionStringList and also creates a variety of
      @Link(TFunctionClass)'es.

      Users should not generally call @Name directly but instead use the
      @Link(TFunctionStringList) already created in
      TRbwParser.@Link(TRbwParser.Functions).
    }
    constructor Create;
    {
      @Name destroys the @Link(TFunctionClass) at the position Index.
    }
    procedure Delete(Index: Integer); override;
    {
      @Name destroys the @Link(TFunctionStringList).
      Do not call @Name directly. Call Free instead.
    }
    destructor Destroy; override;
    {
      @Name returns the @Link(TFunctionClass) at position index.
    }
    property FunctionClass[const Index: integer]: TFunctionClass
      read GetFunctionClass; default;
  end;

  {
    @abstract(@name defines storage for a constant value such as "Abc",
    1, 2.5, or @True.)
  }
  TConstant = class(TObject)
  private
    // @Name: string;
    // See @Link(ResultString).
    // Ordinarily access to FResultString should be avoided.
    FResultString: string;
    // See @Link(ResultString).
    procedure SetResultString(const Value: string);
  protected
    // @name: Pointer;
    // @name points to the address of the result.
    FResult: Pointer;
    // @name: @link(TRbwDataType);
    // @name specifies what type of data is stored in FResult.
    FResultType: TRbwDataType;
    // Create a TConstant and specify @link(FResultType).
    constructor Create(const DataType: TRbwDataType); overload;
    // Convert the result to a string.
    function ValueToString: string;
    // if @link(FResultType) is rdtString, ResultString is the string that is stored.
    property ResultString: string read FResultString write SetResultString;
  public
    {
      @Name creates a @classname and sets @Link(ResultType).
    }
    constructor Create(Value: string); overload;
    {
      @Name creates a @classname and sets @Link(ResultType).
    }
    constructor Create(const Value: Boolean); overload;
    {
      @Name creates a @classname and sets @Link(ResultType).
    }
    constructor Create(const Value: integer); overload;
    {
      @Name creates a @classname and sets @Link(ResultType).
    }
    constructor Create(const Value: double); overload;
    {
      @Name destroys the @Link(TConstant).  Do not call @Name directly.
      Call Free instead.
    }
    destructor Destroy; override;
    {
      @Name defines the type of data stored in the @Link(TConstant).
    }
    property ResultType: TRbwDataType read FResultType;
    {
      @Name returns the value stored in @classname
      if that value is a string. Other wise it generates an error.
    }
    function StringResult: string;
    {
      @Name returns the value stored in @classname
      if that value is a boolean. Other wise it generates an error.
    }
    function BooleanResult: boolean;
    {
      @Name returns the value stored in @classname
      if that value is a integer. Other wise it generates an error.
    }
    function IntegerResult: Integer;
    {
      @Name returns the value stored in @classname
      if that value is a integer or double. Other wise it generates an error.
    }
    function DoubleResult: double;
    {
      @Name converts the value stored in the @classname to a
      string that can be compiled into an equivalent @classname.
    }
    function Decompile: string; virtual;
  end;

  {
    @abstract(@Name is the abstract ancestor of @Link(TCustomVariable)
    and @Link(TExpression).
    It adds a @Link(TCustomValue.Name) field to @Link(TConstant).)
  }
  TCustomValue = class(TConstant)
  private
    // @Name: string;
    // Always upper case
    FName: string;
    // @Name: string;
    // FUserName is the mixed-case version of @link(FName).
    // It is used in @link(Decompile).
    FUserName: string;
  public
    {
      @Name defines how the @Link(TCustomValue) should be identified.
      It is set via Create.
    }
    property Name: string read FName;
    {
      @Name calls the inherited constructor and sets @Link(Name).
    }
    constructor Create(const VariableName: string; const DataType:
      TRbwDataType);
    {
      @Name returns the @Link(Name) of the @Link(TCustomValue).
    }
    function Decompile: string; override;
  end;

  {
    @abstract(@Name is the abstract ancestor of @Link(TRealVariable),
    @Link(TIntegerVariable), @Link(TBooleanVariable)
    and @Link(TStringVariable).)
    It validates the @Link(TCustomValue.Name)
    in @Link(TCustomVariable.Create).
  }
  TCustomVariable = class(TCustomValue)
    {
      @Name calls inherited create and then validates VariableName.
      Variable @Link(TCustomValue.Name)s must start with a letter or underscore.
      The rest of the name must consist of letters underscores and digits.
    }
    constructor Create(const VariableName: string; const DataType:
      TRbwDataType);
  end;

  {
    @abstract(@Name stores a double that may change from one execution of
    a @Link(TExpression) to the next.)
  }
  TRealVariable = class(TCustomVariable)
  private
    // See @link(Value).
    function GetValue: double;
    // See @link(Value).
    procedure SetValue(const Value: double);
  public
    {
      @Name calls the inherited create.
    }
    constructor Create(const VariableName: string);
    {
      @Name is the double stored by @Link(TRealVariable).
    }
    property Value: double read GetValue write SetValue;
  end;

  {
    @abstract(@Name stores a integer that may change from one execution of
    a @Link(TExpression) to the next.)
  }
  TIntegerVariable = class(TCustomVariable)
  private
    // See @link(Value).
    function GetValue: Integer;
    // See @link(Value).
    procedure SetValue(const Value: Integer);
  public
    {
      @Name calls the inherited create.
    }
    constructor Create(const VariableName: string);
    {
      @Name is the Integer stored by @Link(TIntegerVariable).
    }
    property Value: Integer read GetValue write SetValue;
  end;

  {
    @abstract(@Name stores a boolean that may change from one execution of
    a @Link(TExpression) to the next.)
  }
  TBooleanVariable = class(TCustomVariable)
  private
    // See @link(Value).
    function GetValue: Boolean;
    // See @link(Value).
    procedure SetValue(const Value: Boolean);
  public
    {
      @Name calls the inherited create.
    }
    constructor Create(const VariableName: string);
    {
      @Name is the Boolean stored by @Link(TBooleanVariable).
    }
    property Value: Boolean read GetValue write SetValue;
  end;

  {
    @abstract(@Name stores a string that may change from one execution of
    a @Link(TExpression) to the next.)
  }
  TStringVariable = class(TCustomVariable)
  private
    // See @link(Value).
    function GetValue: string;
    // See @link(Value).
    procedure SetValue(const Value: string);
  public
    {
      @Name calls the inherited create.
    }
    constructor Create(const VariableName: string);
    {
      @Name is the String stored by @Link(TStringVariable).
    }
    property Value: string read GetValue write SetValue;
  end;

  {
    @abstract(@Name is the compiled version of an Expression
    generated by the
    TRbwParser.@Link(TRbwParser.Compile) method.)  It can be evaluated by using
    the Evaluate method and the result can then be read
    using the @Link(TConstant.BooleanResult) method,
    @Link(TConstant.DoubleResult) method,
    @Link(TConstant.IntegerResult) method, or
    @Link(TConstant.StringResult) method.
    The correct one to read can be determined from the
    @Link(TConstant.ResultType) property.

    Every instance of @Name is owned by the instance of @Link(TRbwParser)
    that compiled it.
  }
  TExpression = class(TCustomValue)
  private
    // @name: integer;
    // @name is the number of optional arguments in the function that will
    // be evaluated.
    FOptionalArguments: integer;
    // @name: Pointer;
    // @name is the address of the function used to evaluate the
    // TExpression.
    FunctionAddr: Pointer;
    // @name: array of Pointer;
    // @name is the argument passed to the function used
    // to evaluate the @classname.
    VariablesForFunction: array of Pointer;
    // @name: boolean;
    // See @link(AllowConversionToConstant).
    FAllowConversionToConstant: boolean;
    // @name: integer;
    // See @link(Tag)
    FTag: integer;
    // @name: array of integer;
    // StringVariableIndicies indicates which members of
    // @link(VariablesForFunction) refer to strings.
    StringVariableIndicies: array of integer;
    // @name: integer;
    // StringVariableCount is the number of strings in
    // @link(VariablesForFunction).
    StringVariableCount: integer;
    // @name: TStringList;
    // FVariablesUsed is used to hold the result of @link(VariablesUsed).
    FVariablesUsed: TStringList;
    // @name: boolean;
    // FTopLevel is used in @link(Decompile) to determine whether or not to include
    // parenthesis around the outermost item.
    FTopLevel: boolean;
    // Create a TExpression.
    constructor Create(const VariableName: string; const DataType: TRbwDataType;
      const CanConvertToConstant: boolean); overload;
    // Create a TExpression.
    constructor Create(const FunctionRecord: TFunctionRecord); overload;
    // Create a TExpression.
    constructor Create(const VariableName: string; const DataType:
      TRbwDataType);
      overload;
    // Initializes certain variables.  Called by Create.
    procedure Initalize(const FunctionAddress: Pointer;
      const DataTypes: array of TRbwDataType; const OptionalArguments: integer);
    // See @link(Variables).
    function GetVariables(const Index: integer): TConstant;
    // See @link(Variables).
    procedure SetVariables(const Index: integer; const Value: TConstant);
    // if the expression can be represented as a constant value,
    // ConvertToConstant returns a @link(TConstant) that represents
    // that value.  Otherwise, it returns nil.
    function ConvertToConstant: TConstant;
    // If optional arguements are used, ResetDataLength ise used to
    // resets the length of arrays to the correct length.
    procedure ResetDataLength(const Count: integer);
    // See @link(AllowConversionToConstant).
    procedure SetAllowConversionToConstant(const Value: boolean);
    {
      @Name defines whether the result of a function may be
      considered a constant value if all of the values passed to the function
      in the values array are constants.

      Normally @Name should be set to True but if the
      function makes reference to global variables that may change between one
      evaluation of the expression and the next, @Name
      should be set to False.

      Pi is an example of a function for which @Name should
      be true.

      @Name is used when optimizing compiled expressions.
    }
    property AllowConversionToConstant: boolean read FAllowConversionToConstant
      write SetAllowConversionToConstant;
    {
      @Name define the data passed to the function when @Link(Evaluate)
      is called.
      Any descendent of @Link(TConstant) may be assigned to @Name.
    }
    property Variables[const Index: integer]: TConstant read GetVariables
      write SetVariables;
    // @name initializes @link(VariablesForFunction) and @link(StringVariableIndicies)
    procedure FillVariables;
    // See @link(VariablesUsed).
    function GetVariablesUsed: TStringList;
  protected
    {
      @name: boolean;

      @Name is set to false if a @Link(TExpression) is equivalent to a
      @Link(TConstant). In such cases, the @Link(TExpression)
      is evaluated when it
      is created and does not need to be reevaluated later.  An example would
      be if the expression used to create the @Link(TExpression) was "1 + 1".
      This would be converted to "2".
    }
    ShouldEvaluate: boolean;
    {
      @name: array of @link(TVariables);
      @Name holds the arguments used to evaluate the @Link(TExpression).
    }
    Data: array of TVariables;
  public
    {
      @Name creates a @Link(TExpression) based on a @Link(TFunctionClass).
      The values of any variables used by the @Link(TExpression)
      must also be set before calling @Link(Evaluate).
    }
    constructor Create(const FunctionClass: TFunctionClass); overload; virtual;
    {
      @Name converts the value stored in the @Link(TExpression)
      to a string that can be compiled into an equivalent @Link(TExpression)
      together with its arguments.
    }
    function Decompile: string; override;
    {
      @Name destroys the @Link(TExpression).  Do not call Destroy directly.
      Call Free instead.
    }
    destructor Destroy; override;
    {
      @Name evaluates the expression and sets the result which may then be
      read using @Link(TConstant.BooleanResult),
      @Link(TConstant.DoubleResult),
      @Link(TConstant.IntegerResult), or
      @Link(TConstant.StringResult)
      depending on the @Link(TConstant.ResultType).

      Example:

      @Longcode(#
      procedure TForm1.Button1Click(Sender: TObject);
      begin
        RbwParser1.Compile(Edit1.Text);
        RbwParser1.CurrentExpression.Evaluate;
        case RbwParser1.CurrentExpression.ResultType of
          rdtDouble:
            begin
              Label1.Caption := 'result: '
                + FloatToStr(RbwParser1.CurrentExpression.DoubleResult);
            end;
          rdtInteger:
            begin
              Label1.Caption := 'result: '
                + IntToStr(RbwParser1.CurrentExpression.IntegerResult);
            end;
          rdtBoolean:
            begin
              if RbwParser1.CurrentExpression.BooleanResult then
              begin
                Label1.Caption := 'result: ' + 'True';
              end
              else
              begin
                Label1.Caption := 'result: ' + 'False';
              end;
            end;
          rdtString:
            begin
              Label1.Caption := 'result: '
                + RbwParser1.CurrentExpression.StringResult;
            end;
        else Assert(False);
        end;
      end;
      #)
    }
    procedure Evaluate; virtual;
    {
      @Name usually calls Create to create a @Link(TExpression) based
      on FunctionClass. However, in some cases, it creates a descendent
      of @Link(TExpression).  Many of the descendents are declared
      in the implementation section of
      RBW_Parser.pas.  Others can be generated via @Link(SpecialImplentorList).
      One such descendent is @Link(TSelectExpression) which overrides
      @Link(Evaluate).
    }
    class function New(const FunctionClass: TFunctionClass): TExpression;
      virtual;
    {
      @Name has no predefined meaning. The Tag property is provided for the
      convenience of developers. It can be used for storing an
      additional integer value or it can be typecast to any 32-bit value
      such as a component reference or a pointer.
    }
    property Tag: integer read FTag write FTag;
    {
      @Name returns True if Variable is used by the @Link(TExpression).
    }
    function UsesVariable(const Variable: TCustomVariable): boolean;
    {
      @Name is a list of all the variables used by the
      @Link(TExpression).
    }
    property VariablesUsed: TStringList read GetVariablesUsed;
  end;

  // @name is used in @link(TSpecialImplementor) to create a descendent
  // of @link(TExpression).
  TExpressionClass = class of TExpression;

  {
    @abstract(@Name is used in conjunction with
    @Link(SpecialImplentorList) to create a descendent of @Link(TExpression)
    in TExpression.@Link(TExpression.New).)
  }
  TSpecialImplementor = class
  public
    // @Name: @link(TFunctionClass);
    // @Name defines the interface of the expression.
    FunctionClass: TFunctionClass;
    // @name: @link(TExpressionClass);
    // @Name is the class of TExpression that will be created.
    Implementor: TExpressionClass;
  end;

  {
    @abstract(@Name is the type of the global variable
    @Link(SpecialImplentorList).  It is used together with
    @Link(TSpecialImplementor) to create a descendent of @Link(TExpression)
    in TExpression.@Link(TExpression.New).)
  }
  TSpecialImplentorList = class(TObject)
  private
    // @name: TList;
    // is used to store @link(TSpecialImplementor)s.
    FList: TList;
    // See @link(Count).
    function GetCount: integer;
    // See @link(Items).
    function GetItems(const Index: integer): TSpecialImplementor;
    // See @link(Items).
    procedure SetItems(const Index: integer;
      const Value: TSpecialImplementor);
    // See @link(Capacity).
    function GetCapacity: integer;
    // See @link(Capacity).
    procedure SetCapacity(const Value: integer);
  public
    {
      @Name creates an instance of TSpecialImplentorList.
    }
    constructor Create;
    {
      @Name destroys a TSpecialImplentorList. Do not call @Name.
      Call Free instead.
    }
    destructor Destroy; override;
    {
      @Name is the number of @Link(TSpecialImplementor)s in the
      TSpecialImplentorList.
    }
    property Count: integer read GetCount;
    {
      @Name is the number of @Link(TSpecialImplementor)s that the
      TSpecialImplentorList can hold without reallocating memory.

      Setting @Name to a value high enough to hold all the
      @Link(TSpecialImplementor)s that it will hold before @Link(Add)ing
      them, can speed up @Link(Add)ing the @Link(TSpecialImplementor)s.
    }
    property Capacity: integer read GetCapacity write SetCapacity;
    {
      @Name provides access to the @Link(TSpecialImplementor)s held by the
      @Link(TSpecialImplentorList).
    }
    property Items[const Index: integer]: TSpecialImplementor read GetItems write
      SetItems; default;
    {
      @Name adds a TSpecialImplementor to @Link(Items).  The items that are
      added are not owned by the @Link(TSpecialImplentorList).
    }
    function Add(const Item: TSpecialImplementor): Integer;
    {
      @Name  removes everything from @Link(Items) and sets the @Link(Count)
      and @Link(Capacity) to 0.
    }
    procedure Clear;
    {
      @Name removes the @Link(TSpecialImplementor) at position Index from
      @Link(Items);
    }
    procedure Delete(const Index: Integer);
    {
      @Name removes the @Link(TSpecialImplementor) indicated by Item
      from @Link(Items) and returns its former position.  If it was not
      in @Link(Items) it returns -1.
    }
    function Remove(const Item: TSpecialImplementor): Integer;
  end;

  {
    @abstract(@Name compiles expression in strings
    via the @Link(TRbwParser.Compile)
    method into @Link(TExpression) objects that it owns.)  It can also create
    @Link(TRealVariable)s, @Link(TIntegerVariable)s, @Link(TBooleanVariable)s,
    and @Link(TStringVariable)s that can be used in the @Link(TExpression)
    via the @Link(TRbwParser.CreateVariable) method. Alternately, such variables
    created by another instance of TRbwParser can be used via the
    @Link(TRbwParser.RegisterVariable) method.

    All variables and expressions created by an instance of
    @Name are owned by it.

    If several TRbwParsers are used and variables owned by one are used in
    another one, @Link(TRbwParser.ClearExpressions) and
    @Link(TRbwParser.ClearVariables) should
    be called in all of them before any of them is destroyed.

    See @link(TFunctionStringList) and @link(OverloadedFunctionList)
    for a list of the functions included by default.
  }
  TRbwParser = class(TComponent)
  private
    // @name: @link(TExpression);
    // See @link(CurrentExpression).
    FCurrentExpression: TExpression;
    // @name: TStringList;
    // See @link(Expressions).
    FExpressions: TStringList;
    // @name: @link(TFunctionStringList);
    // See @link(Functions).
    FFunctions: TFunctionStringList;
    // @name: TStringList;
    // See @link(Variables).
    FVariables: TStringList;
    // @name: TObjectList;
    // @name holds all the @link(TCustomVariable)s owned by the
    // @link(TRbwParser).
    FOwnedVariables: TObjectList;
    // See @link(Expressions).
    function GetExpressions(const Index: integer): TExpression;
    // See @link(Variables).
    function GetVariable(const Index: integer): TCustomValue;
    { Private declarations }
  protected
    { Protected declarations }
  public
    {
      @Name destroys all compiled @Link(TExpression)s in the
      @Link(TRbwParser).

      If several @Link(TRbwParser)s are used and variables owned by one
      are used in another one, ClearExpressions and
      @Link(ClearVariables) should be called
      in all of them before any of them is destroyed.
    }
    procedure ClearExpressions;
    {
      @Name removes all variables from the list of variables used by
      @Link(TRbwParser) but does not free them (because they may be used
      by another @Link(TRbwParser)).

      If several @Link(TRbwParser)s are used and variables owned by one
      are used in another one, @Link(ClearExpressions) and
      ClearVariables should be called
      in all of them before any of them is destroyed.
    }
    procedure ClearVariables;
    {
      @Name creates a @Link(TBooleanVariable)
      owned by @Link(TRbwParser) and allows
      it to be used with the @Link(TRbwParser).

      See also:  @link(RegisterVariable).
    }
    function CreateVariable(const Name: string; const Value: boolean):
      TBooleanVariable; overload;
    {
      @Name creates a @Link(TIntegerVariable)
      owned by @Link(TRbwParser) and allows
      it to be used with the @Link(TRbwParser).

      See also:  @link(RegisterVariable).
    }
    function CreateVariable(const Name: string; const Value: integer):
      TIntegerVariable; overload;
    {
      @Name creates a @Link(TRealVariable)
      owned by @Link(TRbwParser) and allows
      it to be used with the @Link(TRbwParser).

      See also:  @link(RegisterVariable).

      Example:
      @longcode(#
      procedure TForm1.Button2Click(Sender: TObject);
      var
        Index : integer;
        Variable : TRealVariable;
      begin
        RbwParser1.CreateVariable('A', 2.5);
        RbwParser1.CreateVariable('B', 3.5);
        RbwParser1.Compile('A + B');
        RbwParser1.CurrentExpression.Evaluate;
        Label1.Caption := 'result: '
          + FloatToStr(RbwParser1.CurrentExpression.DoubleResult);
        // Label1.Caption is set to "result: 6".
        Index := RbwParser1.IndexOfVariable('a');
        // variable names are not case sensitive.
        Variable := RbwParser1.Variables[Index] as TRealVariable;
        Variable.Value := 6.5;

        RbwParser1.CurrentExpression.Evaluate;
        Label2.Caption := 'result: '
          + FloatToStr(RbwParser1.CurrentExpression.DoubleResult);
        // Label2.Caption is set to "result: 10".
        // As it stands, this event handler would cause
        // an error if it was called twice because
        // it would try to create variables named "A" and "B" twice.
        // Calling ClearVariables would prevent this if that is
        // what you really want to do.
      end;
      #)
    }
    function CreateVariable(const Name: string; const Value: double):
      TRealVariable; overload;
    {
      @Name creates a @Link(TStringVariable)
      owned by @Link(TRbwParser) and allows
      it to be used with the @Link(TRbwParser).

      See also:  @link(RegisterVariable).
    }
    function CreateVariable(const Name: string; const Value: string):
      TStringVariable; overload;
    {
      @Name searches for AString in @Link(Expressions). If it finds it,
      it sets @Link(CurrentExpression) to that @Link(TExpression)
      and returns it's position in @Link(Expressions).  If it doesn't find it,
      Compile converts AString into a @Link(TExpression), sets
      @Link(CurrentExpression) to the @Link(TExpression) it created,
      adds the @Link(TExpression) to @Link(Expressions)
      and returns the position of the @Link(TExpression) in @Link(Expressions).

      Expressions are owned by @Link(TRbwParser).
    }
    function Compile(var AString: string): integer;
    {
      @Name is the most recently @Link(Compile)d @Link(TExpression).
    }
    property CurrentExpression: TExpression read FCurrentExpression;
    {
      @Name creates an instance of @Link(TRbwParser);
    }
    constructor Create(AOwner: TComponent); override;
    {
      @Name destroys the @Link(TExpression) at Index and deletes it
      from the list of @Link(Expressions).
    }
    procedure DeleteExpression(const Index: integer);
    {
      @Name destroys the @Link(TRbwParser).  Do not call Destroy directly.
      Call Free instead.
    }
    destructor Destroy; override;
    {
      @Name returns the number of @Link(TExpression)s in
      @Link(Expressions).
    }
    function ExpressionCount: integer;
    {
      @Name is an array of @Link(TExpression)s.
    }
    property Expressions[const Index: integer]: TExpression
    read GetExpressions;
    {
      @Name locates the @Link(TCustomVariable) whose name is
      VariableName in @Link(Variables) and returns it's position.  If it is
      not in @Link(Variables) it returns -1.
    }
    function IndexOfVariable(VariableName: string): integer;
    {
      @Name stores @Link(TFunctionClass)es in it's Objects property
      and manages the @Link(TFunctionClass)es used by TRbwParser.
    }
    property Functions: TFunctionStringList read FFunctions;
    {
      @Name allows a @Link(TCustomValue) to be used by
      the @Link(TRbwParser) even though the @Link(TCustomValue) isn't
      owned by the @Link(TRbwParser).
    }
    procedure RegisterVariable(const Value: TCustomValue);
    {
      @Name deletes Expression from the list of
      @Link(Expressions) and destroys it.
    }
    procedure RemoveExpression(const Expression: TExpression);
    {
      If Variable is owned by or registered with @Link(TRbwParser),
      @Name removes it and, if it is owned by the current
      instance of @Link(TRbwParser),
      also destroys it.  It also destroys any TExpression's in the
      current @Link(TRbwParser) that use Variable.
    }
    procedure RemoveVariable(const Variable: TCustomVariable);
    {
      @Name changes the name of the variable at position
      Index to a NewName and changes Index to the new position of
      the variable.
    }
    procedure RenameVariable(var Index: integer; NewName: string);
    {
      @Name is an array of variables (@Link(TCustomValue)s) used by
      the current @Link(TRbwParser).
    }
    property Variables[const Index: integer]: TCustomValue read GetVariable;
    {
      @Name is the number of variables in @Link(Variables).
    }
    function VariableCount: integer;
    { Public declarations }
  published
    { Published declarations }
  end;

  {
    @abstract(@Name is used for "if" and "case" statements. It allows
    Evaluate to be faster and safer by only evaluating the arguments
    that will be used.)
  }
  TSelectExpression = class(TExpression)
  public
    {
      @Name calls Evaluate for its first arguement and based on
      the result of that argument calls Evaluate for one of its other
      arguments and sets its own result to the result of that arguement.
    }
    procedure Evaluate; override;
  end;

  // @Name installs the component on the component palette.
procedure Register;

{
  @Name converts a @Link(TRbwDataType) to a string.  This function
  is primarily used in error messages.
}
function DataTypeToString(const DataType: TRbwDataType): string;

{
  @Name uses root to generate the name of a variable
  that could be accepted by @link(TRbwParser) but does check whether a variable
  with that name already exists.
}
function GenerateVariableName(const root: string): string;

var
  {
    @Name contains a series of @link(TFunctionClass)es
    that define overloaded functions.

    The following functions are included by default.

    @bold(Abs)(Value) returns the absolute value of Value.
    Value can be either an integer or a real number.
    The value returned by Abs will have the same type as Value.

    @bold(Case)(Index, Result1, Result2, ...).
    Case uses Index to determine which of the Result1,
    Result2, ... arguments will be returned as a result.
    If Index equals 1, Result1 is returned; if Index equals 2,
    Result2 is returned; if Index equals 3, Result3 is returned;
    and so forth.  Only "Index", constant expressions
    and the result that is returned will be evaluated.
    The types of Result1, Result2, ... must all be the same
    but they can be of any type.  The type that is returned
    will be the same as the type of Result1, Result2, ...

    @bold(If)(Boolean_Value, If_True_Result, If_False_Result).
    If uses Boolean_Value to determine whether If_True_Result or
    If_False_Result is returned as a result.
    If Boolean_Value is true, If_True_Result is returned;
    if Boolean_Value is false, If_False_Result is returned.
    Only "Boolean_Value", constant expressions and the result
    that is returned will be evaluated.
    The types of If_True_Result and If_False_Result
    must be the same but they can be of any type.
    The type that is returned will be the same as the type of
    If_True_Result and If_False_Result.

    @bold(Max)(Value1, Value2, ...) returns whichever of its arguments
    is the largest.  Its arguments must be either integers or
    real numbers.  The result will be a real number if any of
    the arguments is a real number.  If all the arguments are
    integers, the result will be an integer.

    @bold(Min)(Value1, Value2, ...) returns whichever of its arguments
    is the smallest.  Its arguments must be either integers or
    real numbers.  The result will be a real number if any of
    the arguments is a real number.  If all the arguments are
    integers, the result will be an integer.

    @bold(MultiInterpolate)(Position, Value1, Distance1,
    [Value2, Distance2,] ...). If Position is less than or equal
    to Distance1, MultiInterpolate returns Value1.
    If Position is greater than or equal to DistanceN,
    MultiInterpolate returns ValueN.
    If  Position is between any two adjacent distances,
    linear interpolation between the associated values will be
    used to determine the value that will be returned.
    See also @link(TFunctionStringList Interpolate).

    @bold(Sqr)(Value) returns Value Squared. Value can be either an
    integer or a real number.  The result of Sqr will be an
    integer if Value is an integer.
    Otherwise it will be a real number.

  }
  OverloadedFunctionList: TObjectList;

  {
    See: @Link(TSpecialImplentorList).
  }
  SpecialImplentorList: TSpecialImplentorList;

implementation

uses Math, StrUtils;

const
  ValidCharacters = ['A'..'Z', 'a'..'z', '0'..'9', '_'];
  ValidFirstCharacters = ['A'..'Z', 'a'..'z', '_'];

procedure Register;
begin
  RegisterComponents('RBW', [TRbwParser]);
end;

function GenerateVariableName(const root: string): string;
var
  Index: integer;
begin
  result := root;
  if Length(result) < 1 then
  begin
    raise ERbwParserError.Create('Error: A root name must be supplied when '
      + 'generating a variable name.');
  end;
  for Index := 1 to Length(result) do
  begin
    if not (result[Index] in ValidCharacters) then
    begin
      result[Index] := '_'
    end;
  end;
  if not (result[1] in ValidFirstCharacters) then
  begin
    result := '_' + result;
  end;
end;

procedure ValidateVariableName(const VariableName: string);
var
  Index: integer;
begin
  if Length(VariableName) < 1 then
  begin
    raise ERbwParserError.Create('Error: Variables must be named.');
  end;
  if not (VariableName[1] in ValidFirstCharacters) then
  begin
    raise ERbwParserError.Create('Error: "'
      + VariableName + '" is illegal. '
      +
      'The first character of a variable name must be a letter or an underscore.');
  end;
  for Index := 2 to Length(VariableName) do
  begin
    if not (VariableName[Index] in ValidCharacters) then
    begin
      raise ERbwParserError.Create('Error: "'
        + VariableName + '" is illegal.  Variable names may only include '
        + 'letters, digits, or the underscore symbol.');
    end;
  end;
end;

type
  TIntegerStack = class(TObject)
  private
    FList: TList;
    function GetCapacity: integer;
    procedure SetCapacity(const Value: integer);
  public
    property Capacity: integer read GetCapacity write SetCapacity;
    procedure Clear;
    function Count: integer;
    constructor Create;
    destructor Destroy; override;
    function Peek: Integer;
    function Pop: integer;
    procedure Push(const AnInteger: integer);
  end;

  TTokenStringList = class(TStringList)
  private
    function FixToken(const Token: string): string;
    procedure EvaluateNotAndSigns(Start: integer; var Stop: integer);
    procedure EvaluateMultDivEtc(Start: integer; var Stop: integer);
    procedure EvaluatePlusMinusOrXor(Start: integer; var Stop: integer);
    procedure EvaluateBoolOperators(Start: integer; var Stop: integer);
    procedure MakeVariableList(Start, Stop: integer);
    function Add(const S: string): Integer; override;
    procedure Insert(Index: Integer; const S: string); override;
    function Compile(const Functions: TFunctionStringList;
      const Variables: TStringList): TExpression;
    procedure EvaluateOverloadedFunctions(Start: integer;
      var Stop: integer);
    procedure EvaluateSpecialImplementors(Start: integer;
      var Stop: integer);
  end;

  TSignOperator = class(TExpression)
  public
    {
      Decompile converts the value stored in the TSignOperator to a
      string that can be compiled into an equivalent TSignOperator
      together with its argument.
    }
    function Decompile: string; override;
  end;

  TOperator = class(TExpression)
  public
    {
      Decompile converts the value stored in the TOperator to a
      string that can be compiled into an equivalent TOperator
      together with its argument.
    }
    function Decompile: string; override;
  end;

  TIntToDoubleExpression = class(TExpression)
  public
    {
      Decompile returns the Decompile of the single argument
      of the TIntToDoubleExpression.
    }
    function Decompile: string; override;
  end;

  TPartialEvaluationOperator = class(TOperator)
  protected
    //function MakeLinkedList(Prior: TExpression): TExpression; override;
  end;

  TAndExpression = class(TPartialEvaluationOperator)
  public
    procedure Evaluate; override;
  end;

  TOrExpression = class(TPartialEvaluationOperator)
    procedure Evaluate; override;
  end;

  TMultiInterpolateExpression = class(TSelectExpression)
  public
    procedure Evaluate; override;
  end;

  TVariableExpression = class(TExpression)
  public
    constructor Create(const Variable: TCustomValue);
    function Decompile: string; override;
    procedure Evaluate; override;
  end;

const
  Digits: set of Char = ['0'..'9'];
  Operators: array[0..11] of string = ('+', '-', '*', '/', '^',
    '=', '>=', '<=', '<>', '>', '<', ',');
  WordOperators: array[0..4] of string = ('and', 'or', 'div', 'mod', 'not');
  Parenthesis: set of Char = ['(', ')'];
  Signs: set of Char = ['+', '-'];
  WhiteSpace: set of Char = [' ', #9, #10, #13];

var
  OperatorList: TObjectList;

  IntToDoubleFunction: TFunctionClass;

  NotOperator: TFunctionClass;
  TimesIOperator: TFunctionClass;
  TimesROperator: TFunctionClass;
  DivideIOperator: TFunctionClass;
  DivideROperator: TFunctionClass;
  DivOperator: TFunctionClass;
  ModOperator: TFunctionClass;
  AndOperator: TFunctionClass;
  OrOperator: TFunctionClass;
  PlusSignIOperator: TFunctionClass;
  PlusSignROperator: TFunctionClass;
  MinusSignIOperator: TFunctionClass;
  MinusSignROperator: TFunctionClass;
  PlusIOperator: TFunctionClass;
  PlusROperator: TFunctionClass;
  PlusSOperator: TFunctionClass;
  MinusIOperator: TFunctionClass;
  MinusROperator: TFunctionClass;
  EqualIOperator: TFunctionClass;
  EqualROperator: TFunctionClass;
  EqualSOperator: TFunctionClass;
  EqualBOperator: TFunctionClass;
  NotEqualIOperator: TFunctionClass;
  NotEqualROperator: TFunctionClass;
  NotEqualSOperator: TFunctionClass;
  NotEqualBOperator: TFunctionClass;
  LessThanBOperator: TFunctionClass;
  LessThanIOperator: TFunctionClass;
  LessThanROperator: TFunctionClass;
  LessThanSOperator: TFunctionClass;
  GreaterThanBOperator: TFunctionClass;
  GreaterThanIOperator: TFunctionClass;
  GreaterThanROperator: TFunctionClass;
  GreaterThanSOperator: TFunctionClass;
  LessThanOrEqualsBOperator: TFunctionClass;
  LessThanOrEqualsIOperator: TFunctionClass;
  LessThanOrEqualsROperator: TFunctionClass;
  LessThanOrEqualsSOperator: TFunctionClass;
  GreaterThanOrEqualsBOperator: TFunctionClass;
  GreaterThanOrEqualsIOperator: TFunctionClass;
  GreaterThanOrEqualsROperator: TFunctionClass;
  GreaterThanOrEqualsSOperator: TFunctionClass;
  XorOperator: TFunctionClass;

  AbsIOverloadedFunction: TFunctionClass;
  AbsROverloadedFunction: TFunctionClass;
  CaseBOverloadedFunction: TFunctionClass;
  CaseIOverloadedFunction: TFunctionClass;
  CaseROverloadedFunction: TFunctionClass;
  CaseSOverloadedFunction: TFunctionClass;
  IfBOverloadedFunction: TFunctionClass;
  IfIOverloadedFunction: TFunctionClass;
  IfROverloadedFunction: TFunctionClass;
  IfSOverloadedFunction: TFunctionClass;
  MaxIOverloadedFunction: TFunctionClass;
  MaxROverloadedFunction: TFunctionClass;
  MinIOverloadedFunction: TFunctionClass;
  MinROverloadedFunction: TFunctionClass;
  SqrIOverloadedFunction: TFunctionClass;
  SqrROverloadedFunction: TFunctionClass;
  MultiInterpolateFunction: TFunctionClass;

var
  AbsFunctionI: TFunctionRecord;
  AbsFunctionR: TFunctionRecord;
  ArcCosFunction: TFunctionRecord;
  ArcCoshFunction: TFunctionRecord;
  ArcSinFunction: TFunctionRecord;
  ArcSinhFunction: TFunctionRecord;
  ArcTan2Function: TFunctionRecord;
  ArcTanhFunction: TFunctionRecord;

  CaseBooleanFunction: TFunctionRecord;
  CaseIntegerFunction: TFunctionRecord;
  CaseDoubleFunction: TFunctionRecord;
  CaseStringFunction: TFunctionRecord;

  CopyFunction: TFunctionRecord;
  CosFunction: TFunctionRecord;
  CoshFunction: TFunctionRecord;
  DegToRadFunction: TFunctionRecord;
  FactorialFunction: TFunctionRecord;
  FactorialFFunction: TFunctionRecord;
  FloatToStrFunction: TFunctionRecord;
  FracFunction: TFunctionRecord;
  IfBooleanFunction: TFunctionRecord;
  IfIntegerFunction: TFunctionRecord;
  IfRealFunction: TFunctionRecord;
  IfStringFunction: TFunctionRecord;
  IntPowerFunction: TFunctionRecord;
  IntToStrFunction: TFunctionRecord;
  LengthFunction: TFunctionRecord;
  LnFunction: TFunctionRecord;
  Log10Function: TFunctionRecord;
  LogNFunction: TFunctionRecord;
  LowerCaseFunction: TFunctionRecord;
  MaxIFunction: TFunctionRecord;
  MaxRFunction: TFunctionRecord;
  MinIFunction: TFunctionRecord;
  MinRFunction: TFunctionRecord;
  OddFunction: TFunctionRecord;
  PiFunction: TFunctionRecord;
  PowerFunction: TFunctionRecord;
  PosFunction: TFunctionRecord;
  PosExFunction: TFunctionRecord;
  RadToDegFunction: TFunctionRecord;
  RoundFunction: TFunctionRecord;
  SinFunction: TFunctionRecord;
  SinhFunction: TFunctionRecord;
  SqrIFunction: TFunctionRecord;
  SqrRFunction: TFunctionRecord;
  SqrtFunction: TFunctionRecord;
  StrToIntFunction: TFunctionRecord;
  StrToIntDefFunction: TFunctionRecord;
  StrToFloatFunction: TFunctionRecord;
  StrToFloatDefFunction: TFunctionRecord;
  TanFunction: TFunctionRecord;
  TanhFunction: TFunctionRecord;
  TruncFunction: TFunctionRecord;
  UpperCaseFunction: TFunctionRecord;

  InterpolateFunction: TFunctionRecord;
  DistanceFunction: TFunctionRecord;

function DataTypeToString(const DataType: TRbwDataType): string;
begin
  case DataType of
    rdtDouble:
      begin
        result := 'a Real';
      end;
    rdtInteger:
      begin
        result := 'an Integer';
      end;
    rdtBoolean:
      begin
        result := 'a Boolean';
      end;
    rdtString:
      begin
        result := 'a String';
      end;
  else
    Assert(False);
  end;

end;

function IsStringConstant(const Token: string): boolean;
begin
  result := (Length(Token) > 0) and (Token[1] = '"');
end;

function IsBooleanConstant(const Token: string; var Value: boolean): boolean;
var
  Temp: string;
begin
  Temp := UpperCase(Token);
  if Temp = 'TRUE' then
  begin
    result := True;
    Value := True;
  end
  else if Temp = 'FALSE' then
  begin
    result := True;
    Value := False;
  end
  else
  begin
    result := False;
  end;
end;

function IsIntegerConstant(const Token: string; var Value: integer): boolean;
var
  Code: integer;
begin
  Assert(Length(Token) >= 1);
  // Val('x1', Value, Code) will convert to a numeric value.
  // Use this test to prevent that problem.
  result := Token[1] in ['+', '-', '0', '1','2','3','4','5','6','7','8','9'];
  if result then
  begin
    Val(Token, Value, Code);
    result := Code = 0;
  end;
end;

function IsFloatConstant(const Token: string; var Value: double): boolean;
var
  Code: integer;
begin
  Assert(Length(Token) >= 1);
  result := Token[1] in ['+', '-', '0', '1','2','3','4','5','6','7','8','9'];
  if result then
  begin
    Val(Token, Value, Code);
    result := Code = 0;
  end;
end;

{ TRbwParser }

function TRbwParser.CreateVariable(const Name: string;
  const Value: integer): TIntegerVariable;
begin
  if FVariables.IndexOf(Trim(UpperCase(Name))) >= 0 then
  begin
    raise ErbwParserError.CreateMode('Error: A Variable named ' + Name +
      ' already exists', 1);
  end;
  result := TIntegerVariable.Create(Name);
  result.Value := Value;
  FVariables.AddObject(result.Name, result);
  FOwnedVariables.Add(result);
end;

function TRbwParser.CreateVariable(const Name: string;
  const Value: boolean): TBooleanVariable;
begin
  if FVariables.IndexOf(Trim(UpperCase(Name))) >= 0 then
  begin
    raise ErbwParserError.CreateMode('Error: A Variable named ' + Name +
      ' already exists', 1);
  end;
  result := TBooleanVariable.Create(Name);
  result.Value := Value;
  FVariables.AddObject(result.Name, result);
  FOwnedVariables.Add(result);
end;

function TRbwParser.CreateVariable(const Name, Value: string): TStringVariable;
begin
  if FVariables.IndexOf(Trim(UpperCase(Name))) >= 0 then
  begin
    raise ErbwParserError.CreateMode('Error: A Variable named ' + Name +
      ' already exists', 1);
  end;
  result := TStringVariable.Create(Name);
  result.Value := Value;
  FVariables.AddObject(result.Name, result);
  FOwnedVariables.Add(result);
end;

function TRbwParser.CreateVariable(const Name: string;
  const Value: double): TRealVariable;
begin
  if FVariables.IndexOf(Trim(UpperCase(Name))) >= 0 then
  begin
    raise ErbwParserError.CreateMode('Error: A Variable named ' + Name +
      ' already exists', 1);
  end;
  result := TRealVariable.Create(Name);
  result.Value := Value;
  FVariables.AddObject(result.Name, result);
  FOwnedVariables.Add(result);
end;

function TRbwParser.Compile(var AString: string): integer;
var
  Index: integer;
  StringConstants: TStringList;
  IsStringConstantA: boolean;
  LastPosition: integer;
  LineIndex: integer;
  ALine: string;
  IsWhiteSpace: boolean;
  OperatorIndex: integer;
  Operator: string;
  TestString: string;
  Skip: integer;
  Level: integer;
  Tokens: TTokenStringList;
begin
  AString := (Trim(AString));
  if AString = '' then
  begin
    result := -1;
    Exit;
  end;

  if FExpressions.Find(AString, result) then
  begin
    FCurrentExpression := Expressions[Result];
  end
  else
  begin
    FCurrentExpression := nil;
    Tokens := TTokenStringList.Create;
    try
      // Find string constants (starting with the " character.
      // For now, at least, string constants may not contain the " character.
      LastPosition := 0;
      IsStringConstantA := False;
      StringConstants := TStringList.Create;
      try
        for Index := 1 to Length(AString) do
        begin
          if AString[Index] = '"' then
          begin
            IsStringConstantA := not IsStringConstantA;
            if (Index > 1) then
            begin
              if IsStringConstantA then
              begin
                StringConstants.Add(UpperCase(Copy(AString, LastPosition + 1,
                  Index - LastPosition - 1)));
              end
              else
              begin
                StringConstants.Add(Copy(AString, LastPosition,
                  Index - LastPosition + 1));
              end;
            end;
            LastPosition := Index;
          end;
        end;
        if LastPosition < Length(AString) then
        begin
          if IsStringConstantA then
          begin
            // raise error: unterminated string
            raise ErbwParserError.Create('Error: Unterminated string');
          end
          else
          begin
            StringConstants.Add(Copy(AString, LastPosition + 1,
              Length(AString) - LastPosition));
          end;
        end;
        // At this point StringConstants contains string constants
        // and text that requires further processing.
        for LineIndex := 0 to StringConstants.Count - 1 do
        begin
          ALine := StringConstants[LineIndex];
          // If the current line is a string constant, add it to the tokens and
          // then go on to the next line.
          if Length(ALine) > 0 then
          begin
            if ALine[1] = '"' then
            begin
              Tokens.Add(ALine);
              Continue;
            end;
          end;
          // ALine is not a string constant. Extract tokens from it.
          // LastPosition is used to identify the beginning of the current token.
          // LastPosition is one space before the beginning of the current token.
          LastPosition := 0;
          ALine := Trim(ALine);
          IsWhiteSpace := True;
          Skip := 0;
          for Index := 1 to Length(ALine) do
          begin
            // if we found the beginning of an operator in the last character and the
            // length of the operator was > 1, skip this character.
            if Skip > 0 then
            begin
              LastPosition := Index;
              Dec(Skip);
              Continue;
            end;
            // test if this character is white space.
            if ALine[Index] in WhiteSpace then
            begin
              // if this character is white space, but the previous character wasn't
              // white space, the last character must be the end of a token.
              if not IsWhiteSpace then
              begin
                Tokens.Add(copy(ALine, LastPosition + 1, Index - LastPosition -
                  1));
              end;
              IsWhiteSpace := True;
              LastPosition := Index;
              Continue;
            end
            else
            begin
              IsWhiteSpace := False;
            end;
            // The current character is not white space.
            // Add parentheses to the tokens.
            if ALine[Index] in Parenthesis then
            begin
              Tokens.Add(copy(ALine, LastPosition + 1, Index - LastPosition -
                1));
              Tokens.Add(ALine[Index]);
              LastPosition := Index;
              Continue;
            end;
            // See if the current character is the beginning of an operator.
            // If so, add the prior token and the operator to the list of tokens.
            for OperatorIndex := 0 to Length(Operators) - 1 do
            begin
              Operator := Operators[OperatorIndex];
              TestString := Copy(ALine, Index, Length(Operator));
              if Operator = TestString then
              begin
                if (Index > 1)
                  and ((Operator = '-') or (Operator = '+'))
                  and (ALine[Index - 1] in ['e', 'E']) then
                begin
                  try
                    StrToFloat(Copy(ALine, LastPosition + 1, Index -
                      LastPosition));
                    break;
                  except on EConvertError do
                    begin
                      // do nothing = don't break;
                    end;
                  end;

                end;

                Tokens.Add(copy(ALine, LastPosition + 1, Index - LastPosition -
                  1));
                Tokens.Add(TestString);
                LastPosition := Index;
                // if the length of the token is greater than 1, skip characters
                // until after the end of the operator is reached.
                Skip := Length(TestString) - 1;
                break;
              end;
            end;
            for OperatorIndex := 0 to Length(WordOperators) - 1 do
            begin
              Operator := WordOperators[OperatorIndex];
              TestString := Copy(ALine, Index, Length(Operator));
              if Operator = TestString then
              begin
                if ((Index = 1) or (ALine[Index - 1] in [' ', ')', ',']))
                  and ((Index + Length(Operator) > Length(ALine))
                  or (ALine[Index + Length(Operator)] in [' ', '(', ','])) then
                begin
                  Tokens.Add(copy(ALine, LastPosition + 1, Index - LastPosition
                    - 1));
                  Tokens.Add(TestString);
                  LastPosition := Index;
                  // if the length of the token is greater than 1, skip characters
                  // until after the end of the operator is reached.
                  Skip := Length(TestString) - 1;
                  break;
                end;
              end;
            end;
          end;
          // Add last token
          if LastPosition < Length(ALine) then
          begin
            Tokens.Add(Copy(ALine, LastPosition + 1, Length(ALine) -
              LastPosition));
          end;
        end;
      finally
        StringConstants.Free;
      end;
      Level := 0;
      for LineIndex := 0 to Tokens.Count - 1 do
      begin
        ALine := Tokens[LineIndex];
        if not IsStringConstant(ALine) then
        begin
          if ALine = '(' then
          begin
            Inc(Level)
          end;
          if ALine = ')' then
          begin
            Dec(Level)
          end;
          if Level < 0 then
          begin
            raise ErbwParserError.Create('Error: Unmatched parentheses');
          end;
        end;
      end;
      if Level <> 0 then
      begin
        raise ErbwParserError.Create('Error: Unmatched parentheses');
      end;
      FCurrentExpression := Tokens.Compile(FFunctions, FVariables);
      AString := FCurrentExpression.Decompile;
      if FExpressions.Find(AString, result) then
      begin
        Assert(FCurrentExpression.ResultType = Expressions[Result].ResultType);
        FCurrentExpression.Free;
        FCurrentExpression := Expressions[Result];
      end
      else
      begin
        result := FExpressions.AddObject(AString, FCurrentExpression);
        //FCurrentExpression.MakeLinkedList(nil);
      end;

    finally
      Tokens.Free;
    end;
  end;
end;

constructor TRbwParser.Create(AOwner: TComponent);
begin
  inherited;
  FOwnedVariables := TObjectList.Create;
  FExpressions := TStringList.Create;
  FExpressions.Sorted := True;
  FFunctions := TFunctionStringList.Create;
  FVariables := TStringList.Create;
  FVariables.Sorted := True;
end;

destructor TRbwParser.Destroy;
var
  Index: integer;
begin
  for Index := 0 to FExpressions.Count - 1 do
  begin
    FExpressions.Objects[Index].Free;
  end;
  FExpressions.Free;
  FFunctions.Free;
  FVariables.Free;
  FOwnedVariables.Free;
  inherited;
end;

function TRbwParser.GetExpressions(const Index: integer): TExpression;
begin
  result := FExpressions.Objects[Index] as TExpression;
end;

function TRbwParser.GetVariable(const Index: integer): TCustomValue;
begin
  result := FVariables.Objects[Index] as TCustomValue;
end;

function TRbwParser.IndexOfVariable(VariableName: string): integer;
begin
  VariableName := Trim(UpperCase(VariableName));
  if not FVariables.Find(VariableName, result) then
  begin
    result := -1;
  end;
end;

function TRbwParser.VariableCount: integer;
begin
  result := FVariables.Count;
end;

procedure TRbwParser.RegisterVariable(const Value: TCustomValue);
begin
  if (Value is TBooleanVariable) or (Value is TIntegerVariable)
    or (Value is TRealVariable) or (Value is TStringVariable) then
  begin
    if FVariables.IndexOf(Value.Name) >= 0 then
    begin
      raise ErbwParserError.CreateMode('Error: A Variable named ' + Name +
        ' already exists', 1);
    end
    else
    begin
      FVariables.AddObject(Value.Name, Value)
    end;
  end
  else
  begin
    raise ERbwParserError.Create('Error: Invalid variable type.');
  end;
end;

procedure TRbwParser.ClearExpressions;
var
  Index: integer;
begin
  for Index := 0 to FExpressions.Count - 1 do
  begin
    FExpressions.Objects[Index].Free;
  end;
  FExpressions.Clear;
  FCurrentExpression := nil;
end;

procedure TRbwParser.ClearVariables;
begin
  FVariables.Clear;
end;

procedure TRbwParser.RenameVariable(var Index: integer; NewName: string);
var
  AVariable: TCustomValue;
  NewExpressions: TStringList;
  ExpressionIndex: integer;
  AnExpression: TExpression;
  NewUserName: string;
  Position: integer;
begin
  NewUserName := Trim(NewName);
  NewName := UpperCase(NewUserName);
  ValidateVariableName(NewName);
  if Variables[Index].Decompile <> NewUserName then
  begin
    Position := FVariables.IndexOf(NewName);
    if (Position >= 0) and (Position <> Index) then
    begin
      raise ErbwParserError.CreateMode('Error: A Variable named ' + NewName +
        ' already exists', 1);
    end;
    AVariable := FVariables.Objects[Index] as TCustomValue;
    FVariables.Delete(Index);
    AVariable.FName := NewName;
    AVariable.FUserName := NewUserName;
    Index := FVariables.AddObject(NewName, AVariable);
    NewExpressions := TStringList.Create;
    try
      for ExpressionIndex := 0 to FExpressions.Count - 1 do
      begin
        AnExpression := FExpressions.Objects[ExpressionIndex] as TExpression;
        NewExpressions.AddObject(AnExpression.Decompile, AnExpression)
      end;
      FExpressions.Free;
      FExpressions := NewExpressions;
    except
      NewExpressions.Free;
      raise;
    end;

  end;

end;

procedure TRbwParser.DeleteExpression(const Index: integer);
begin
  Expressions[Index].Free;
  FExpressions.Delete(Index);
end;

procedure TRbwParser.RemoveExpression(const Expression: TExpression);
var
  Position: integer;
begin
  Position := FExpressions.IndexOfObject(Expression);
  if Position >= 0 then
  begin
    DeleteExpression(Position);
  end;
end;

procedure TRbwParser.RemoveVariable(const Variable: TCustomVariable);
var
  Index: integer;
  Expression: TExpression;
  Position: integer;
begin
  for Index := ExpressionCount - 1 downto 0 do
  begin
    Expression := Expressions[Index];
    if Expression.UsesVariable(Variable) then
    begin
      Expression.Free;
      FExpressions.Delete(Index);
    end;
  end;
  Position := FVariables.IndexOfObject(Variable);
  if Position >= 0 then
  begin
    FVariables.Delete(Position);
  end;
  FOwnedVariables.Remove(Variable)
end;

function TRbwParser.ExpressionCount: integer;
begin
  result := FExpressions.Count;
end;

{ TCustomValue }

constructor TCustomValue.Create(const VariableName: string;
  const DataType: TRbwDataType);
begin
  inherited Create(DataType);
  FUserName := Trim(VariableName);
  FName := UpperCase(FUserName);
end;

function TCustomValue.Decompile: string;
begin
  result := FUserName;
end;

{ TRealVariable }

constructor TRealVariable.Create(const VariableName: string);
begin
  inherited Create(VariableName, rdtDouble);
end;

function TRealVariable.GetValue: double;
begin
  result := PDouble(FResult)^;
end;

procedure TRealVariable.SetValue(const Value: double);
begin
  PDouble(FResult)^ := Value;
end;

{ TIntegerVariable }

constructor TIntegerVariable.Create(const VariableName: string);
begin
  inherited Create(VariableName, rdtInteger);
end;

function TIntegerVariable.GetValue: Integer;
begin
  result := PInteger(FResult)^;
end;

procedure TIntegerVariable.SetValue(const Value: Integer);
begin
  PInteger(FResult)^ := Value;
end;

{ TBooleanVariable }

constructor TBooleanVariable.Create(const VariableName: string);
begin
  inherited Create(VariableName, rdtBoolean);
end;

function TBooleanVariable.GetValue: Boolean;
begin
  result := PBoolean(FResult)^;
end;

procedure TBooleanVariable.SetValue(const Value: Boolean);
begin
  PBoolean(FResult)^ := Value;
end;

{ TStringVariable }

constructor TStringVariable.Create(const VariableName: string);
begin
  inherited Create(VariableName, rdtString);
end;

function TStringVariable.GetValue: string;
begin
  result := ResultString;
end;

procedure TStringVariable.SetValue(const Value: string);
begin
  ResultString := Value;
end;

{ TConstant }

constructor TConstant.Create(Value: string);
begin
  Create(rdtString);
  if (Length(Value) > 0) and (Value[1] = '"') and (Value[Length(Value)] = '"')
    then
  begin
    Value := Copy(Value, 2, Length(Value) - 2);
  end;
  ResultString := Value;
end;

constructor TConstant.Create(const DataType: TRbwDataType);
begin
  inherited Create;
  FResultType := DataType;
  case ResultType of
    rdtDouble:
      begin
        New(PDouble(FResult));
      end;
    rdtInteger:
      begin
        New(PInteger(FResult));
      end;
    rdtBoolean:
      begin
        New(PBoolean(FResult));
      end;
    rdtString:
      begin
        FResult := @ResultString;
      end;
  else
    Assert(False);
  end;
end;

constructor TConstant.Create(const Value: Boolean);
begin
  Create(rdtBoolean);
  PBoolean(FResult)^ := Value;
end;

function TConstant.BooleanResult: boolean;
begin
  if ResultType <> rdtBoolean then
  begin
    raise ErbwParserError.Create('Error: attempt to use '
      + DataTypeToString(ResultType)
      + ' as a boolean.');
  end;
  result := PBoolean(FResult)^;
end;

constructor TConstant.Create(const Value: double);
begin
  Create(rdtDouble);
  PDouble(FResult)^ := Value;
end;

constructor TConstant.Create(const Value: integer);
begin
  Create(rdtInteger);
  PInteger(FResult)^ := Value;
end;

destructor TConstant.Destroy;
begin
  if FResultType <> rdtString then
  begin
    Dispose(FResult);
  end;
  inherited;
end;

function TConstant.DoubleResult: double;
begin
  if not (ResultType in [rdtDouble, rdtInteger]) then
  begin
    raise ErbwParserError.Create('Error: attempt to use '
      + DataTypeToString(ResultType)
      + ' as a Real.');
  end;
  if ResultType = rdtInteger then
  begin
    result := PInteger(FResult)^;
  end
  else
  begin
    result := PDouble(FResult)^;
  end;

end;

function TConstant.IntegerResult: Integer;
begin
  if ResultType <> rdtInteger then
  begin
    raise ErbwParserError.Create('Error: attempt to use '
      + DataTypeToString(ResultType)
      + ' as an Integer.');
  end;
  result := PInteger(FResult)^;
end;

function TConstant.StringResult: string;
begin
  if ResultType <> rdtString then
  begin
    raise ErbwParserError.Create('Error: attempt to use '
      + DataTypeToString(ResultType)
      + ' as a String.');
  end;
  result := ResultString;
end;

function TConstant.Decompile: string;
begin
  result := ValueToString;
end;

function TConstant.ValueToString: string;
var
  OldDecimalSeparator: Char;
begin
  result := '';
  case ResultType of
    rdtDouble:
      begin
        OldDecimalSeparator := DecimalSeparator;
        try
          DecimalSeparator := '.';
          result := FloatToStr(DoubleResult);
        finally
          DecimalSeparator := OldDecimalSeparator;
        end;
        if (Pos('.', result) <= 0) and (Pos('e', result) <= 0) and (Pos('E',
          result) <= 0) then
        begin
          result := result + '.';
        end;

      end;
    rdtInteger:
      begin
        result := IntToStr(IntegerResult);
      end;
    rdtBoolean:
      begin
        if BooleanResult then
        begin
          result := 'True';
        end
        else
        begin
          result := 'False';
        end;
      end;
    rdtString:
      begin
        result := '"' + StringResult + '"';
      end;
  else
    Assert(False);
  end;
end;

procedure TConstant.SetResultString(const Value: string);
begin
  FResultString := Value;
  FResult := @FResultString
end;

{ TIntegerStack }

procedure TIntegerStack.Clear;
var
  Index: integer;
begin
  for Index := 0 to FList.Count - 1 do
  begin
    FreeMem(FList[Index]);
  end;
  FList.Clear;
end;

function TIntegerStack.Count: integer;
begin
  result := FList.Count;
end;

constructor TIntegerStack.Create;
begin
  inherited Create;
  FList := TList.Create;
end;

destructor TIntegerStack.Destroy;
begin
  Clear;
  FList.Free;
  inherited;

end;

function TIntegerStack.GetCapacity: integer;
begin
  result := FList.Capacity;
end;

function TIntegerStack.Peek: Integer;
begin
  result := PInteger(FList[FList.Count - 1])^
end;

function TIntegerStack.Pop: integer;
var
  LastPointer: PInteger;
begin
  if FList.Count = 0 then
  begin
    raise ERangeError.Create('Error: The Stack is empty.');
  end;
  LastPointer := FList[FList.Count - 1];
  result := LastPointer^;
  FreeMem(LastPointer);
  FList.Count := FList.Count - 1;
end;

procedure TIntegerStack.Push(const AnInteger: integer);
var
  LastPointer: PInteger;
begin
  GetMem(LastPointer, SizeOf(Integer));
  LastPointer^ := AnInteger;
  FList.Add(LastPointer);
end;

procedure TIntegerStack.SetCapacity(const Value: integer);
begin
  if Value <= 0 then
  begin
    Clear;
  end
  else
  begin
    while Capacity < Value do
    begin
      Pop;
    end;
    FList.Capacity := Value;
  end;
end;

{ TTokenStringList }

function TTokenStringList.Add(const S: string): Integer;
var
  Token: string;
begin
  Token := FixToken(S);
  if (Token = '') then
  begin
    result := -1;
  end
  else
  begin
    result := inherited Add(Token);
  end;
end;

function TTokenStringList.Compile(const Functions: TFunctionStringList;
  const Variables: TStringList): TExpression;
var
  Index: integer;
  Token: string;
  ABoolean: Boolean;
  AnInteger: integer;
  ADouble: double;
  Position: integer;
  ParenStack: TIntegerStack;
  StartParenPosition: integer;
  PriorExpression: TExpression;
  VariableList: TList;
  VarIndex, VarIndex2: integer;
  AnObject: TObject;
  isVariable: boolean;
  Constant: TConstant;
  ResultConstant: TConstant;
  resultVariable: TCustomValue;
  procedure CleanUp;
  var
    Index: integer;
    VarIndex: integer;
  begin
    begin
      for Index := 0 to Count - 1 do
      begin
        AnObject := Objects[Index];
        if AnObject <> nil then
        begin
          if AnObject is TList then
          begin
            VariableList := AnObject as TList;
            for VarIndex := 0 to VariableList.Count - 1 do
            begin
              Constant := VariableList[VarIndex];
              if not (Constant is TCustomValue) or (Constant is TExpression)
                then
              begin
                Constant.Free;
              end;
            end;
            VariableList.Free;
          end
          else
          begin
            Constant := AnObject as TConstant;
            if not (Constant is TCustomValue) or (Constant is TExpression) then
            begin
              Constant.Free;
            end;
          end;
          Objects[Index] := nil;
        end;
      end;
      //      raise;
    end;
  end;
begin
  result := nil;
  try
    for Index := 0 to Count - 1 do
    begin
      Token := Strings[Index];
      if (Token = '(') or (Token = ')') or (Token = ',') then
      begin
        Continue;
      end
      else if IsStringConstant(Token) then
      begin
        Objects[Index] := TConstant.Create(Token);
      end
      else if IsBooleanConstant(Token, ABoolean) then
      begin
        Objects[Index] := TConstant.Create(ABoolean);
      end
      else if IsIntegerConstant(Token, AnInteger) then
      begin
        Objects[Index] := TConstant.Create(AnInteger);
      end
      else if IsFloatConstant(Token, ADouble) then
      begin
        Objects[Index] := TConstant.Create(ADouble);
      end
      else if (Length(Token) = 1) and (Token[1] in Parenthesis) then
      begin
        Continue;
      end
      else if (Variables <> nil) and (Variables.Find(Token, Position)) then
      begin
        Objects[Index] := Variables.Objects[Position];
      end
      else if (Functions <> nil) and (Functions.Find(Token, Position)) then
      begin
        Objects[Index] := TExpression.New(Functions.FunctionClass[Position]);
      end;
    end;
    ParenStack := TIntegerStack.Create;
    try
      Index := 0;
      while Index < Count do
      begin
        Token := Strings[Index];
        if Token = '(' then
        begin
          ParenStack.Push(Index);
        end
        else if Token = ')' then
        begin
          StartParenPosition := ParenStack.Pop;
          MakeVariableList(StartParenPosition, Index);
          Index := StartParenPosition;
          isVariable := True;
          if Index > 0 then
          begin
            AnObject := Objects[Index - 1];
            if AnObject <> nil then
            begin
              PriorExpression := AnObject as TExpression;
              isVariable := False;
              VariableList := Objects[Index] as TList;
              if VariableList = nil then
              begin
                if ((Length(PriorExpression.Data)
                  - PriorExpression.FOptionalArguments) <> 0) then
                begin

                  if PriorExpression.FOptionalArguments = 0 then
                  begin
                    raise ErbwParserError.Create('Error: the '
                      + PriorExpression.FName
                      + ' function requires '
                      + IntToStr(Length(PriorExpression.Data))
                      + ' arguments but none were supplied.');
                  end
                  else
                  begin
                    raise ErbwParserError.Create('Error: the '
                      + PriorExpression.FName
                      + ' function requires between '
                      + IntToStr(Length(PriorExpression.Data) -
                      PriorExpression.FOptionalArguments)
                      + ' and '
                      + IntToStr(Length(PriorExpression.Data))
                      + ' arguments but none were supplied.');
                  end;

                end;
              end
              else
              begin
                if (PriorExpression.FOptionalArguments < 0) then
                begin
                  if (Length(PriorExpression.Data)
                    - 1 > VariableList.Count) then
                  begin
                    raise ErbwParserError.Create('Error: the '
                      + PriorExpression.FName
                      + ' function requires at least '
                      + IntToStr(Length(PriorExpression.Data) - 1)
                      + ' arguments but '
                      + IntToStr(VariableList.Count)
                      + ' were supplied.');
                  end;
                  PriorExpression.ResetDataLength(VariableList.Count);
                end
                else
                begin
                  if (Length(PriorExpression.Data) < VariableList.Count) then
                  begin
                    raise ErbwParserError.Create('Error: the '
                      + PriorExpression.FName
                      + ' function requires at most '
                      + IntToStr(Length(PriorExpression.Data))
                      + ' arguments but '
                      + IntToStr(VariableList.Count)
                      + ' were supplied.');
                  end;
                  if (Length(PriorExpression.Data)
                    - PriorExpression.FOptionalArguments > VariableList.Count)
                      then
                  begin
                    raise ErbwParserError.Create('Error: the '
                      + PriorExpression.FName
                      + ' function requires at least '
                      + IntToStr(Length(PriorExpression.Data) -
                      PriorExpression.FOptionalArguments)
                      + ' arguments but '
                      + IntToStr(VariableList.Count)
                      + ' were supplied.');
                  end;
                end;
                VarIndex2 := -1;
                try
                  for VarIndex := 0 to VariableList.Count - 1 do
                  begin
                    VarIndex2 := VarIndex;
                    PriorExpression.Variables[VarIndex] :=
                      VariableList[VarIndex];
                    VariableList[VarIndex] := nil;
                  end;
                except on E: ERbwParserError do
                  begin
                    raise ERbwParserError.Create('Error in argument number '
                      + IntToStr(VarIndex2 + 1)
                      + ' of '
                      + PriorExpression.Name
                      + Strings[Index]
                      + '; '
                      + E.Message);
                  end;
                end;
                VariableList.Free;
                Strings[Index - 1] := Strings[Index - 1] + Strings[Index];
                Delete(Index);
                Dec(Index);
              end;
            end;
          end;
          if isVariable then
          begin
            VariableList := Objects[Index] as TList;
            if VariableList.Count > 1 then
            begin
              Dec(Index);
              MakeVariableList(Index, Index + 1);
              VariableList := Objects[Index] as TList;
              if VariableList.Count <> 1 then
              begin
                for VarIndex := 0 to VariableList.Count - 1 do
                begin
                  Constant := VariableList[VarIndex];
                  if not (Constant is TCustomValue) or (Constant is TExpression)
                    then
                  begin
                    Constant.Free;
                  end;
                end;
                VariableList.Free;
                Objects[Index] := nil;
                raise ErbwParserError.Create('Parsing Error; Check that no '
                  + 'function or variable names have been misspelled.');
              end;
            end;

            // Assert(VariableList.Count=1);
            Objects[Index] := VariableList[0];
            VariableList.Free;
          end;
        end;
        Inc(Index);
      end;
    finally
      ParenStack.Free;
    end;
    MakeVariableList(0, Count - 1);
    VariableList := Objects[0] as TList;
    try
      if VariableList <> nil then
      begin
        if VariableList.Count <> 1 then
        begin
          for VarIndex := 0 to VariableList.Count - 1 do
          begin
            Constant := VariableList[VarIndex];
            if not (Constant is TCustomValue) or (Constant is TExpression) then
            begin
              Constant.Free;
            end;
          end;
          VariableList.Free;
          Objects[0] := nil;
          raise ErbwParserError.Create('Parsing Error; Check that no '
            + 'function or variable names have been misspelled.');
        end;
        ResultConstant := VariableList[0];
        if ResultConstant is TExpression then
        begin
          result := VariableList[0];
          result.FillVariables;
          //result.MakeLinkedList(nil);
          Constant := result.ConvertToConstant;
          result.FillVariables;
          if Constant <> nil then
          begin
            case Constant.FResultType of
              rdtDouble:
                begin
                  PDouble(result.FResult)^ := PDouble(Constant.FResult)^
                end;
              rdtInteger:
                begin
                  PInteger(result.FResult)^ := PInteger(Constant.FResult)^
                end;
              rdtBoolean:
                begin
                  PBoolean(result.FResult)^ := PBoolean(Constant.FResult)^
                end;
              rdtString:
                begin
                  result.ResultString := Constant.ResultString;
                end;
            else
              Assert(False);
            end;
            result.FunctionAddr := nil;
            result.ShouldEvaluate := False;
            result.FUserName := Constant.Decompile;
            Constant.Free;
          end;
        end
        else if ResultConstant is TCustomValue then
        begin
          resultVariable := VariableList[0];
          result := TVariableExpression.Create(resultVariable);
        end
        else
        begin
          result := TExpression.Create('Dummy',
            ResultConstant.ResultType);
          case result.ResultType of
            rdtDouble:
              begin
                PDouble(result.FResult)^ := PDouble(ResultConstant.FResult)^;
              end;
            rdtInteger:
              begin
                PInteger(result.FResult)^ := PInteger(ResultConstant.FResult)^;
              end;
            rdtBoolean:
              begin
                PBoolean(result.FResult)^ := PBoolean(ResultConstant.FResult)^;
              end;
            rdtString:
              begin
                result.ResultString := ResultConstant.ResultString;
              end;
          else
            Assert(False);
          end;
          result.FUserName := ResultConstant.Decompile;
          ResultConstant.Free;
        end;
      end;
    finally
      VariableList.Free;
      Objects[0] := nil;
    end;
  except
    on ERbwParserError do
    begin
      CleanUp;
      raise;
    end;
    on E: Exception do
    begin
      CleanUp;
      raise ERbwParserError.Create(E.Message);
    end;
  end
end;

procedure TTokenStringList.EvaluateBoolOperators(Start: integer;
  var Stop: integer);
var
  Index: integer;
  AnExpression: TExpression;
  PriorArgument: TConstant;
  SubsequentArgument: TConstant;
  Token: string;
begin
  Index := Start;
  while Index <= Stop do
  begin
    Token := UpperCase(Strings[Index]);
    if Token = '=' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "=" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "=" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType = rdtBoolean)
        and (SubsequentArgument.ResultType = rdtBoolean) then
      begin
        AnExpression := TOperator.Create(EqualBOperator);
      end
      else if (PriorArgument.ResultType = rdtString)
        and (SubsequentArgument.ResultType = rdtString) then
      begin
        AnExpression := TOperator.Create(EqualSOperator);
      end
      else if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(EqualIOperator);
      end
      else if (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        and (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        AnExpression := TOperator.Create(EqualROperator);
      end
      else
      begin
        raise ErbwParserError.Create('Error in parsing "=" operator; '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + ' can not be compared with the "=" operator.'
          );
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = '<>' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "<>" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "<>" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType = rdtBoolean)
        and (SubsequentArgument.ResultType = rdtBoolean) then
      begin
        AnExpression := TOperator.Create(NotEqualBOperator);
      end
      else if (PriorArgument.ResultType = rdtString)
        and (SubsequentArgument.ResultType = rdtString) then
      begin
        AnExpression := TOperator.Create(NotEqualSOperator);
      end
      else if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(NotEqualIOperator);
      end
      else if (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        and (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        AnExpression := TOperator.Create(NotEqualROperator);
      end
      else
      begin
        raise ErbwParserError.Create('Error in parsing "<>" operator; '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + ' can not be compared with the "<>" operator.'
          );
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = '<' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "<" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "<" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType = rdtBoolean)
        and (SubsequentArgument.ResultType = rdtBoolean) then
      begin
        AnExpression := TOperator.Create(LessThanBOperator);
      end
      else if (PriorArgument.ResultType = rdtString)
        and (SubsequentArgument.ResultType = rdtString) then
      begin
        AnExpression := TOperator.Create(LessThanSOperator);
      end
      else if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(LessThanIOperator);
      end
      else if (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        and (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        AnExpression := TOperator.Create(LessThanROperator);
      end
      else
      begin
        raise ErbwParserError.Create('Error in parsing "<" operator; '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + ' can not be compared with the "<" operator.'
          );
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = '>' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing ">" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing ">" operator.'
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType = rdtBoolean)
        and (SubsequentArgument.ResultType = rdtBoolean) then
      begin
        AnExpression := TOperator.Create(GreaterThanBOperator);
      end
      else if (PriorArgument.ResultType = rdtString)
        and (SubsequentArgument.ResultType = rdtString) then
      begin
        AnExpression := TOperator.Create(GreaterThanSOperator);
      end
      else if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(GreaterThanIOperator);
      end
      else if (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        and (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        AnExpression := TOperator.Create(GreaterThanROperator);
      end
      else
      begin
        raise ErbwParserError.Create('Error in parsing ">" operator; '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + ' can not be compared with the ">" operator.'
          );
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = '<=' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "<=" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "<=" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType = rdtBoolean)
        and (SubsequentArgument.ResultType = rdtBoolean) then
      begin
        AnExpression := TOperator.Create(LessThanOrEqualsBOperator);
      end
      else if (PriorArgument.ResultType = rdtString)
        and (SubsequentArgument.ResultType = rdtString) then
      begin
        AnExpression := TOperator.Create(LessThanOrEqualsSOperator);
      end
      else if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(LessThanOrEqualsIOperator);
      end
      else if (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        and (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        AnExpression := TOperator.Create(LessThanOrEqualsROperator);
      end
      else
      begin
        raise ErbwParserError.Create('Error in parsing "<=" operator; '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + ' can not be compared with the "<=" operator.'
          );
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = '>=' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing ">=" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing ">=" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType = rdtBoolean)
        and (SubsequentArgument.ResultType = rdtBoolean) then
      begin
        AnExpression := TOperator.Create(GreaterThanOrEqualsBOperator);
      end
      else if (PriorArgument.ResultType = rdtString)
        and (SubsequentArgument.ResultType = rdtString) then
      begin
        AnExpression := TOperator.Create(GreaterThanOrEqualsSOperator);
      end
      else if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(GreaterThanOrEqualsIOperator);
      end
      else if (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        and (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        AnExpression := TOperator.Create(GreaterThanOrEqualsROperator);
      end
      else
      begin
        raise ErbwParserError.Create('Error in parsing ">=" operator; '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + ' can not be compared with the ">=" operator.'
          );
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end;
    Inc(index);
  end;
end;

procedure TTokenStringList.EvaluateMultDivEtc(Start: integer;
  var Stop: integer);
var
  Index: integer;
  AnExpression: TExpression;
  PriorArgument: TConstant;
  SubsequentArgument: TConstant;
  Token: string;
begin
  Index := Start;
  while Index <= Stop do
  begin
    Token := UpperCase(Strings[Index]);
    if Token = '*' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "*" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "*" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if not (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        or not (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        raise ErbwParserError.Create('Error in parsing "*" operator;'
          + ' It is not possible to apply the "*" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(TimesIOperator);
      end
      else
      begin
        AnExpression := TOperator.Create(TimesROperator);
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = '/' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "/" operator  .'
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "/" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if not (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        or not (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        raise ErbwParserError.Create('Error in parsing "/" operator;'
          + ' It is not possible to apply the "/" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(DivideIOperator);
      end
      else
      begin
        AnExpression := TOperator.Create(DivideROperator);
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = 'DIV' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "DIV" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "DIV" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType <> rdtInteger)
        or (SubsequentArgument.ResultType <> rdtInteger) then
      begin
        raise ErbwParserError.Create('Error in parsing "DIV" operator;'
          + ' It is not possible to apply the "DIV" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      AnExpression := TOperator.Create(DivOperator);
      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = 'MOD' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "MOD" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "MOD" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType <> rdtInteger)
        or (SubsequentArgument.ResultType <> rdtInteger) then
      begin
        raise ErbwParserError.Create('Error in parsing "MOD" operator;'
          + ' It is not possible to apply the "MOD" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      AnExpression := TOperator.Create(ModOperator);
      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end
    else if Token = 'AND' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "AND" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "AND" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType <> rdtBoolean)
        or (SubsequentArgument.ResultType <> rdtBoolean) then
      begin
        raise ErbwParserError.Create('Error in parsing "AND" operator;'
          + ' It is not possible to apply the "AND" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      AnExpression := TOperator.New(AndOperator);
      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index);
    end;
    Inc(index);
  end;
end;

procedure TTokenStringList.EvaluateNotAndSigns(Start: integer;
  var Stop: integer);
var
  Index: integer;
  AnExpression: TExpression;
  AnArgument: TConstant;
  Token: string;
  IsSign: boolean;
begin
  Index := Stop;
  while Index >= Start do
  begin
    Token := Uppercase(Strings[Index]);
    if Token = 'NOT' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) then
      begin
        raise ErbwParserError.Create('Error in parsing "NOT" operator.');
      end;
      AnArgument := Objects[Index + 1] as TConstant;
      if (AnArgument = nil) then
      begin
        raise ErbwParserError.Create('Error in parsing "NOT" operator.');
      end;
      if (AnArgument.ResultType <> rdtBoolean) then
      begin
        raise ErbwParserError.Create('Error in parsing "NOT" operator;'
          + ' It is not possible to apply the "NOT" operator to '
          + DataTypeToString(AnArgument.ResultType)
          + '.'
          );
      end;
      Assert((AnArgument <> nil) and (AnArgument.ResultType = rdtBoolean));
      AnExpression := TOperator.Create(NotOperator);
      AnExpression.Variables[0] := AnArgument;
      Objects[Index] := AnExpression;
      Strings[Index] := Strings[Index] + ' ' + Strings[Index + 1];
      Delete(Index + 1);
      Dec(Stop);
    end;
    Dec(Index);
  end;

  Index := Start;
  while Index <= Stop do
  begin

    Token := Uppercase(Strings[Index]);
    if (Token = '+') or (Token = '-') then
    begin
      // see if this is a sign operator rather than a
      // plus or minus operator.
      if (Index + 1 >= Count) then
      begin
        raise ErbwParserError.Create('Error in parsing ' + Token + ' sign.');
      end;
      AnArgument := Objects[Index + 1] as TConstant;
      if AnArgument <> nil then
      begin
        if ((AnArgument.ResultType = rdtInteger)
          or (AnArgument.ResultType = rdtDouble))
          {and not (AnArgument is TExpression)} then
        begin
          // The next item is consistent with a sign operator.
          // Check the previous item.
          if (Index = 0) then
          begin
            IsSign := True;
          end
          else
          begin
            if (Index < 0) then
            begin
              raise ErbwParserError.Create('Error in parsing ' + Token +
                ' sign.');
            end;
            AnArgument := Objects[Index - 1] as TConstant;
            IsSign := (AnArgument = nil);
          end;
          if IsSign then
          begin
            AnArgument := Objects[Index + 1] as TConstant;
            if (Token = '+') then
            begin
              if AnArgument.ResultType = rdtInteger then
              begin
                AnExpression := TSignOperator.Create(PlusSignIOperator);
              end
              else
              begin
                AnExpression := TSignOperator.Create(PlusSignROperator);
              end;
            end
            else
            begin
              if AnArgument.ResultType = rdtInteger then
              begin
                AnExpression := TSignOperator.Create(MinusSignIOperator);
              end
              else
              begin
                AnExpression := TSignOperator.Create(MinusSignROperator);
              end;
            end;
            AnExpression.Variables[0] := AnArgument;
            Objects[Index] := AnExpression;
            Strings[Index] := Strings[Index] + Strings[Index + 1];
            Delete(Index + 1);
            Dec(Stop);
          end;

        end;
      end;
    end;
    Inc(Index);
  end;

end;

procedure TTokenStringList.EvaluatePlusMinusOrXor(Start: integer;
  var Stop: integer);
var
  Index: integer;
  AnExpression: TExpression;
  PriorArgument: TConstant;
  SubsequentArgument: TConstant;
  Token: string;
begin
  Index := Start;
  while Index <= Stop do
  begin
    Token := UpperCase(Strings[Index]);
    if Token = '+' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "+" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "+" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if not (PriorArgument.ResultType in [rdtDouble, rdtInteger, rdtString])
        or not (SubsequentArgument.ResultType in [rdtDouble, rdtInteger,
        rdtString]) then
      begin
        raise ErbwParserError.Create('Error in parsing "+" operator;'
          + ' It is not possible to apply the "+" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      if (PriorArgument.ResultType = rdtString)
        and (SubsequentArgument.ResultType = rdtString) then
      begin
        AnExpression := TOperator.Create(PlusSOperator);
      end
      else if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(PlusIOperator);
      end
      else if (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        and (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        AnExpression := TOperator.Create(PlusROperator);
      end
      else
      begin
        raise ErbwParserError.Create('Error in parsing "+" operator;'
          + ' It is not possible to apply the "+" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index)
    end
    else if Token = '-' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "-" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "-" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if not (PriorArgument.ResultType in [rdtDouble, rdtInteger])
        or not (SubsequentArgument.ResultType in [rdtDouble, rdtInteger]) then
      begin
        raise ErbwParserError.Create('Error in parsing "-" operator;'
          + ' It is not possible to apply the "-" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      if (PriorArgument.ResultType = rdtInteger)
        and (SubsequentArgument.ResultType = rdtInteger) then
      begin
        AnExpression := TOperator.Create(MinusIOperator);
      end
      else
      begin
        AnExpression := TOperator.Create(MinusROperator);
      end;

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index)
    end
    else if Token = 'OR' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "OR" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "OR" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType <> rdtBoolean)
        or (SubsequentArgument.ResultType <> rdtBoolean) then
      begin
        raise ErbwParserError.Create('Error in parsing "OR" operator;'
          + ' It is not possible to apply the "OR" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      AnExpression := TOperator.New(OrOperator);

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index)
    end
    else if Token = 'XOR' then
    begin
      if (Objects[Index] <> nil) or (Index + 1 >= Count) or (Index - 1 < 0) then
      begin
        raise ErbwParserError.Create('Error in parsing "XOR" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      PriorArgument := Objects[Index - 1] as TConstant;
      SubsequentArgument := Objects[Index + 1] as TConstant;
      if ((PriorArgument = nil) or (SubsequentArgument = nil)) then
      begin
        raise ErbwParserError.Create('Error in parsing "XOR" operator.  '
          + 'Check to make sure that all the variables in the formula have '
          + 'been defined.');
      end;
      if (PriorArgument.ResultType <> rdtBoolean)
        or (SubsequentArgument.ResultType <> rdtBoolean) then
      begin
        raise ErbwParserError.Create('Error in parsing "XOR" operator;'
          + ' It is not possible to apply the "XOR" operator to '
          + DataTypeToString(PriorArgument.ResultType)
          + ' and '
          + DataTypeToString(SubsequentArgument.ResultType)
          + '.'
          );
      end;
      AnExpression := TOperator.New(XorOperator);

      AnExpression.Variables[0] := PriorArgument;
      AnExpression.Variables[1] := SubsequentArgument;
      Objects[Index - 1] := AnExpression;
      Strings[Index - 1] := Strings[Index - 1] + ' ' + Strings[Index] + ' ' +
        Strings[Index + 1];
      Delete(Index + 1);
      Delete(Index);
      Dec(Stop, 2);
      Dec(Index)
    end;
    Inc(index);
  end;
end;

function TTokenStringList.FixToken(const Token: string): string;
begin
  result := Trim(Token);
  if (Length(result) > 1) and (Result[1] = '+') then
  begin
    result := Copy(result, 2, MAXINT);
  end;
end;

procedure TTokenStringList.Insert(Index: Integer; const S: string);
var
  Token: string;
begin
  Token := FixToken(S);
  if Token = '' then
  begin
    Exit;
  end
  else
  begin
    inherited Insert(Index, Token);
  end;
end;

procedure TTokenStringList.EvaluateSpecialImplementors(Start: integer;
  var Stop: integer);
var
  Functions, Arguments: TList;
  Index: integer;
  AnExpression: TExpression;
  AnArgument: TConstant;
  AnObject: TObject;
  FunctionIndex: integer;
  ArgumentIndex: integer;
  Token: string;
  FunctionClass: TFunctionClass;
  DeleteFunction: boolean;
  LastArgument: integer;
  NewString: string;
  function DataType: TRbwDataType;
  begin
    if ArgumentIndex >= FunctionClass.InputDataCount then
    begin
      Assert(FunctionClass.OptionalArguments <> 0);
      result := FunctionClass.InputDataTypes[FunctionClass.InputDataCount - 1];
    end
    else
    begin
      result := FunctionClass.InputDataTypes[ArgumentIndex];
    end;

  end;
  procedure FillArgumentList;
  var
    ArgumentIndex: integer;
    AList: TList;
    ListIndex: integer;
  begin
    Arguments.Clear;
    for ArgumentIndex := Index + 1 to Stop do
    begin
      LastArgument := ArgumentIndex;
      AnObject := Objects[ArgumentIndex];
      if AnObject is TList then
      begin
        AList := TList(AnObject);
        for ListIndex := 0 to AList.Count - 1 do
        begin
          Arguments.Add(AList[ListIndex]);
        end;
      end
      else
      begin
        AnArgument := AnObject as TConstant;
        if AnArgument <> nil then
        begin
          Arguments.Add(AnArgument);
        end;
      end;
      if Arguments.Count >= FunctionClass.InputDataCount then
      begin
        break;
      end;
    end;
  end;
begin
  Functions := TList.Create;
  Arguments := TList.Create;
  try
    Index := Start;
    while Index <= Stop do
    begin
      Token := UpperCase(Strings[Index]);
      AnObject := Objects[Index];
      if (AnObject = nil) and (Token <> '(')
        and (Token <> ')') and (Token <> ',') then
      begin
        Functions.Clear;
        for FunctionIndex := 0 to SpecialImplentorList.Count - 1 do
        begin
          FunctionClass := SpecialImplentorList[FunctionIndex].FunctionClass;
          if UpperCase(FunctionClass.Name) = Token then
          begin
            Functions.Add(FunctionClass);
          end;
        end;
        for FunctionIndex := Functions.Count - 1 downto 0 do
        begin
          FunctionClass := Functions[FunctionIndex];
          FillArgumentList;
          if FunctionClass.OptionalArguments < 0 then
          begin
            if (Arguments.Count < FunctionClass.InputDataCount - 1) then
            begin
              Functions.Delete(FunctionIndex);
              Continue;
            end;
          end
          else
          begin
            if (Arguments.Count < FunctionClass.InputDataCount) then
            begin
              Functions.Delete(FunctionIndex);
              Continue;
            end;
          end;
          DeleteFunction := False;
          for ArgumentIndex := 0 to Arguments.Count - 1 do
          begin
            AnArgument := Arguments[ArgumentIndex];
            if DataType <> AnArgument.ResultType then
            begin
              if (DataType = rdtDouble)
                and (AnArgument.ResultType = rdtInteger) then
              begin
                Continue;
              end
              else
              begin
                DeleteFunction := True;
                break;
              end;
            end;
          end;
          if DeleteFunction then
          begin
            Functions.Delete(FunctionIndex);
            Continue;
          end;
        end;
        if Functions.Count > 1 then
        begin
          for FunctionIndex := Functions.Count - 1 downto 0 do
          begin
            FunctionClass := Functions[FunctionIndex];
            FillArgumentList;
            DeleteFunction := False;
            for ArgumentIndex := 0 to Arguments.Count - 1 do
            begin
              AnArgument := Arguments[ArgumentIndex];
              if DataType <> AnArgument.ResultType then
              begin
                DeleteFunction := True;
                break;
              end;
            end;
            if DeleteFunction then
            begin
              Functions.Delete(FunctionIndex);
              Continue;
            end;
          end;
        end;
        if Functions.Count = 0 then
        begin
          Inc(Index);
          Continue;
        end;
        Assert(Functions.Count = 1);
        FunctionClass := Functions[0];
        AnExpression := TExpression.New(FunctionClass);
        if FunctionClass.OptionalArguments <> 0 then
        begin
          SetLength(AnExpression.Data, Arguments.Count);
          SetLength(AnExpression.VariablesForFunction, Arguments.Count);

        end;

        Objects[Index] := AnExpression;
        for ArgumentIndex := 0 to Arguments.Count - 1 do
        begin
          AnArgument := Arguments[ArgumentIndex];
          AnExpression.Data[ArgumentIndex].DataType := DataType;
          AnExpression.Variables[ArgumentIndex] := AnArgument;
        end;
        NewString := '';
        for ArgumentIndex := Index to Stop do
        begin
          NewString := NewString + Strings[ArgumentIndex];
          if (ArgumentIndex >= LastArgument)
            or (Strings[Index] = ')') or (Strings[Index] = ',') then
          begin
            LastArgument := ArgumentIndex;
            break;
          end;
        end;
        Strings[Index] := NewString;
        for ArgumentIndex := LastArgument downto Index + 1 do
        begin
          AnObject := Objects[ArgumentIndex];
          if AnObject is TList then
          begin
            AnObject.Free;
          end;
          Delete(ArgumentIndex);
        end;
        Stop := Stop - LastArgument + Index;

      end;
      Inc(Index);
    end;
  finally
    Functions.Free;
    Arguments.Free;
  end;
end;

procedure TTokenStringList.EvaluateOverloadedFunctions(Start: integer;
  var Stop: integer);
var
  Functions, Arguments: TList;
  Index: integer;
  AnExpression: TExpression;
  AnArgument: TConstant;
  AnObject: TObject;
  FunctionIndex: integer;
  ArgumentIndex: integer;
  Token: string;
  FunctionClass: TFunctionClass;
  DeleteFunction: boolean;
  LastArgument: integer;
  NewString: string;
  function DataType: TRbwDataType;
  begin
    if ArgumentIndex >= FunctionClass.InputDataCount then
    begin
      Assert(FunctionClass.OptionalArguments <> 0);
      result := FunctionClass.InputDataTypes[FunctionClass.InputDataCount - 1];
    end
    else
    begin
      result := FunctionClass.InputDataTypes[ArgumentIndex];
    end;

  end;
  procedure FillArgumentList;
  var
    ArgumentIndex: integer;
    AList: TList;
    ListIndex: integer;
  begin
    Arguments.Clear;
    for ArgumentIndex := Index + 1 to Stop do
    begin
      LastArgument := ArgumentIndex;
      AnObject := Objects[ArgumentIndex];
      if AnObject is TList then
      begin
        AList := TList(AnObject);
        for ListIndex := 0 to AList.Count - 1 do
        begin
          Arguments.Add(AList[ListIndex]);
        end;
      end
      else
      begin
        AnArgument := AnObject as TConstant;
        if AnArgument <> nil then
        begin
          Arguments.Add(AnArgument);
        end;
      end;
      if Arguments.Count >= FunctionClass.InputDataCount then
      begin
        break;
      end;
    end;
  end;
begin
  Functions := TList.Create;
  Arguments := TList.Create;
  try
    Index := Start;
    while Index <= Stop do
    begin
      Token := UpperCase(Strings[Index]);
      AnObject := Objects[Index];
      if (AnObject = nil) and (Token <> '(')
        and (Token <> ')') and (Token <> ',') then
      begin
        Functions.Clear;
        for FunctionIndex := 0 to OverloadedFunctionList.Count - 1 do
        begin
          FunctionClass := OverloadedFunctionList[FunctionIndex]
            as TFunctionClass;
          if UpperCase(FunctionClass.Name) = Token then
          begin
            Functions.Add(FunctionClass);
          end;
        end;
        for FunctionIndex := Functions.Count - 1 downto 0 do
        begin
          FunctionClass := Functions[FunctionIndex];
          FillArgumentList;
          if FunctionClass.OptionalArguments < 0 then
          begin
            if (Arguments.Count < FunctionClass.InputDataCount - 1) then
            begin
              Functions.Delete(FunctionIndex);
              Continue;
            end;
          end
          else
          begin
            if (Arguments.Count < FunctionClass.InputDataCount) then
            begin
              Functions.Delete(FunctionIndex);
              Continue;
            end;
          end;
          DeleteFunction := False;
          for ArgumentIndex := 0 to Arguments.Count - 1 do
          begin
            AnArgument := Arguments[ArgumentIndex];
            if DataType <> AnArgument.ResultType then
            begin
              if (DataType = rdtDouble)
                and (AnArgument.ResultType = rdtInteger) then
              begin
                Continue;
              end
              else
              begin
                DeleteFunction := True;
                break;
              end;
            end;
          end;
          if DeleteFunction then
          begin
            Functions.Delete(FunctionIndex);
            Continue;
          end;
        end;
        if Functions.Count > 1 then
        begin
          for FunctionIndex := Functions.Count - 1 downto 0 do
          begin
            FunctionClass := Functions[FunctionIndex];
            FillArgumentList;
            DeleteFunction := False;
            for ArgumentIndex := 0 to Arguments.Count - 1 do
            begin
              AnArgument := Arguments[ArgumentIndex];
              if DataType <> AnArgument.ResultType then
              begin
                DeleteFunction := True;
                break;
              end;
            end;
            if DeleteFunction then
            begin
              Functions.Delete(FunctionIndex);
              Continue;
            end;
          end;
        end;
        if Functions.Count = 0 then
        begin
          Inc(Index);
          Continue;
        end;
        Assert(Functions.Count = 1);
        FunctionClass := Functions[0];
        AnExpression := TExpression.New(FunctionClass);
        if FunctionClass.OptionalArguments <> 0 then
        begin
          SetLength(AnExpression.Data, Arguments.Count);
          SetLength(AnExpression.VariablesForFunction, Arguments.Count);
        end;

        Objects[Index] := AnExpression;
        for ArgumentIndex := 0 to Arguments.Count - 1 do
        begin
          AnArgument := Arguments[ArgumentIndex];
          AnExpression.Data[ArgumentIndex].DataType := DataType;
          AnExpression.Variables[ArgumentIndex] := AnArgument;
        end;
        NewString := '';
        for ArgumentIndex := Index to Stop do
        begin
          NewString := NewString + Strings[ArgumentIndex];
          if (ArgumentIndex >= LastArgument)
            or (Strings[Index] = ')') or (Strings[Index] = ',') then
          begin
            LastArgument := ArgumentIndex;
            break;
          end;
        end;
        Strings[Index] := NewString;
        for ArgumentIndex := LastArgument downto Index + 1 do
        begin
          AnObject := Objects[ArgumentIndex];
          if AnObject is TList then
          begin
            AnObject.Free;
          end;
          Delete(ArgumentIndex);
        end;
        Stop := Stop - LastArgument + Index;

      end;
      Inc(Index);
    end;
  finally
    Functions.Free;
    Arguments.Free;
  end;
end;

procedure TTokenStringList.MakeVariableList(Start, Stop: integer);
var
  ExpressionString: string;
  Index: integer;
  ResultList: TList;
  AnObject: TObject;
  Token: string;
begin
  EvaluateSpecialImplementors(Start, Stop);
  EvaluateOverloadedFunctions(Start, Stop);
  EvaluateNotAndSigns(Start, Stop);
  EvaluateMultDivEtc(Start, Stop);
  EvaluatePlusMinusOrXor(Start, Stop);
  EvaluateBoolOperators(Start, Stop);

  ResultList := TList.Create;
  try
    for Index := Start to Stop do
    begin
      AnObject := Objects[Index];
      if AnObject = nil then
      begin
        Token := Strings[Index];
        if (Token <> '(') and (Token <> ',') and (Token <> ')') then
        begin
          raise ErbwParserError.Create('Error: "' + Token + '" not recognized '
            + 'or the arguments for "' + Token
            + '" are not of the correct types.');
        end;
      end
      else
      begin
        ResultList.Add(AnObject);
      end;
    end;
    ExpressionString := '';
    for Index := Start to Stop do
    begin
      ExpressionString := ExpressionString + Strings[Index];
    end;
    Strings[Start] := ExpressionString;
    if ResultList.Count = 0 then
    begin
      ResultList.Free;
      ResultList := nil;
    end
    else
    begin
      Objects[Start] := ResultList;
      ResultList := nil;
    end;
    for Index := Stop downto Start + 1 do
    begin
      Delete(Index);
    end;
  except
    ResultList.Free;
    raise;
  end;
end;

{ TExpression }

procedure TExpression.Initalize(const FunctionAddress: Pointer;
  const DataTypes: array of TRbwDataType; const OptionalArguments: integer);
var
  Index: Integer;
begin
  FunctionAddr := FunctionAddress;
  ShouldEvaluate := Assigned(FunctionAddr);
  SetLength(Data, Length(DataTypes));
  SetLength(VariablesForFunction, Length(DataTypes));
  for Index := 0 to Length(DataTypes) - 1 do
  begin
    with Data[Index] do
    begin
      DataType := DataTypes[Index];
      Datum := nil;
    end;
  end;
  FOptionalArguments := OptionalArguments;
end;

procedure TExpression.ResetDataLength(const Count: integer);
var
  ArrayLength: integer;
  LastDataType: TRbwDataType;
  Index: integer;
begin
  Assert(FOptionalArguments < 0);
  ArrayLength := Length(Data);
  Assert(ArrayLength > 0);
  LastDataType := Data[ArrayLength - 1].DataType;
  SetLength(Data, Count);
  for Index := ArrayLength to Count - 1 do
  begin
    with Data[Index] do
    begin
      DataType := LastDataType;
      Datum := nil;
    end;
  end;
  SetLength(VariablesForFunction, Count);
end;

procedure TExpression.Evaluate;
var
//  AnExpression: TExpression;
  Index: integer;
  AVariable: TConstant;
  I: integer;
begin
  if ShouldEvaluate then
  begin
    for Index := 0 to Length(Data) - 1 do
    begin
      if Data[Index].Datum <> nil then
      begin
        AVariable := Data[Index].Datum;
        if AVariable is TExpression then
        begin
          TExpression(AVariable).Evaluate;
        end;
      end;
    end;
    for Index := 0 to StringVariableCount - 1 do
    begin
      I := StringVariableIndicies[Index];
      AVariable := Data[I].Datum;
      VariablesForFunction[I] := AVariable.FResult;
    end;
    case ResultType of
      rdtDouble:
        begin
          PDouble(FResult)^ := TRbwRealFunction(FunctionAddr)
            (VariablesForFunction);
        end;
      rdtInteger:
        begin
          PInteger(FResult)^ := TRbwIntegerFunction(FunctionAddr)
            (VariablesForFunction);
        end;
      rdtBoolean:
        begin
          PBoolean(FResult)^ := TRbwBooleanFunction(FunctionAddr)
            (VariablesForFunction);
        end;
      rdtString:
        begin
          ResultString := TRbwStringFunction(FunctionAddr)
            (VariablesForFunction);
        end;
    else
      Assert(False);
    end;
  end;
end;


constructor TExpression.Create(const FunctionClass: TFunctionClass);
begin
  Create(FunctionClass.FunctionRecord);
end;

function TExpression.GetVariables(const Index: integer): TConstant;
begin
  result := Data[Index].Datum;
end;

procedure TExpression.SetVariables(const Index: integer;
  const Value: TConstant);
var
  Converter: TExpression;
begin
  if ((Value = nil) or (Data[Index].DataType = Value.ResultType)
    or ((Data[Index].DataType = rdtDouble) and (Value.ResultType = rdtInteger)))
      then
  begin
    if (Data[Index].DataType = rdtDouble) and (Value.ResultType = rdtInteger)
      then
    begin
      Converter := TExpression.New(IntToDoubleFunction);
      Converter.Variables[0] := Value;
      Data[Index].Datum := Converter;
    end
    else
    begin
      Data[Index].Datum := Value;
    end;
  end
  else
  begin
    raise ErbwParserError.Create('Error: ' + DataTypeToString(Value.ResultType)
      + ' can not be assigned to '
      + DataTypeToString(Data[Index].DataType)
      + ' variable.');
  end;

end;

destructor TExpression.Destroy;
var
  Index: integer;
  AConstant: TConstant;
begin
  for Index := 0 to Length(Data) - 1 do
  begin
    AConstant := Variables[Index];
    if (AConstant is TExpression) or not (AConstant is TCustomValue) then
    begin
      AConstant.Free;
    end;
  end;
  FVariablesUsed.Free;
  inherited;
end;

constructor TExpression.Create(const FunctionRecord: TFunctionRecord);
begin
  Create(FunctionRecord.Name, FunctionRecord.ResultType,
    FunctionRecord.CanConvertToConstant);
  Initalize(@FunctionRecord.RFunctionAddr, FunctionRecord.InputDataTypes,
    FunctionRecord.OptionalArguments);
end;

function TExpression.ConvertToConstant: TConstant;
var
  Index: Integer;
  AConstant: TConstant;
  NewConstant: TConstant;
  CanConvert: boolean;
begin
  ShouldEvaluate := Assigned(FunctionAddr);
  result := nil;
  if not AllowConversionToConstant then
    Exit;
  CanConvert := True;
  for Index := 0 to Length(Data) - 1 do
  begin
    AConstant := Variables[Index];
    if AConstant <> nil then
    begin
      if AConstant is TExpression then
      begin
        NewConstant := TExpression(AConstant).ConvertToConstant;
        if NewConstant = nil then
        begin
          CanConvert := False;
        end
        else
        begin
          AConstant.Free;
          Variables[Index] := NewConstant;
        end;
      end
      else if AConstant is TCustomValue then
      begin
        CanConvert := False;
      end;
    end;
  end;
  if CanConvert then
  begin
    FillVariables;
    Evaluate;
    case ResultType of
      rdtDouble:
        begin
          result := TConstant.Create(DoubleResult);
        end;
      rdtInteger:
        begin
          result := TConstant.Create(IntegerResult);
        end;
      rdtBoolean:
        begin
          result := TConstant.Create(BooleanResult);
        end;
      rdtString:
        begin
          result := TConstant.Create(StringResult);
        end;
    else
      Assert(False);
    end;
  end;
  ShouldEvaluate := Assigned(FunctionAddr);
end;

function _Not(Values: array of pointer): Boolean;
begin
  result := not PBoolean(Values[0])^;
end;

function _Xor(Values: array of pointer): Boolean;
begin
  result := PBoolean(Values[0])^ xor PBoolean(Values[1])^;
end;

function _TimesI(Values: array of pointer): Integer;
begin
  result := PInteger(Values[0])^ * PInteger(Values[1])^;
end;

function _TimesR(Values: array of pointer): double;
begin
  result := PDouble(Values[0])^ * PDouble(Values[1])^;
end;

function _DivideR(Values: array of pointer): double;
begin
  result := PDouble(Values[0])^ / PDouble(Values[1])^;
end;

function _DivideI(Values: array of pointer): double;
begin
  result := PInteger(Values[0])^ / PInteger(Values[1])^;
end;

function _Div(Values: array of pointer): Integer;
begin
  result := PInteger(Values[0])^ div PInteger(Values[1])^;
end;

function _Mod(Values: array of pointer): Integer;
begin
  result := PInteger(Values[0])^ mod PInteger(Values[1])^;
end;

function _And(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ and PBoolean(Values[1])^;
end;

function _Or(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ or PBoolean(Values[1])^;
end;

function _PlusSignI(Values: array of pointer): integer;
begin
  result := PInteger(Values[0])^;
end;

function _PlusSignR(Values: array of pointer): double;
begin
  result := PDouble(Values[0])^;
end;

function _MinusSignI(Values: array of pointer): integer;
begin
  result := -PInteger(Values[0])^;
end;

function _MinusSignR(Values: array of pointer): double;
begin
  result := -PDouble(Values[0])^;
end;

function _PlusI(Values: array of pointer): integer;
begin
  result := PInteger(Values[0])^ + PInteger(Values[1])^;
end;

function _PlusR(Values: array of pointer): double;
begin
  result := PDouble(Values[0])^ + PDouble(Values[1])^;
end;

function _PlusS(Values: array of pointer): string;
var
  String1, String2: string;
begin
  String1 := PString(Values[0])^;
  String2 := PString(Values[1])^;
  result := String1 + String2;
end;

function _MinusI(Values: array of pointer): integer;
begin
  result := PInteger(Values[0])^ - PInteger(Values[1])^;
end;

function _MinusR(Values: array of pointer): double;
begin
  result := PDouble(Values[0])^ - PDouble(Values[1])^;
end;

function _EqualR(Values: array of pointer): boolean;
begin
  result := PDouble(Values[0])^ = PDouble(Values[1])^;
end;

function _EqualI(Values: array of pointer): boolean;
begin
  result := PInteger(Values[0])^ = PInteger(Values[1])^;
end;

function _EqualS(Values: array of pointer): boolean;
var
  String1, String2: String;
begin
  String1 := PString(Values[0])^;
  String2 := PString(Values[1])^;
  result := String1 = String2;
end;

function _EqualB(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ = PBoolean(Values[1])^;
end;

function _NotEqualB(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ <> PBoolean(Values[1])^;
end;

function _NotEqualI(Values: array of pointer): boolean;
begin
  result := PInteger(Values[0])^ <> PInteger(Values[1])^;
end;

function _NotEqualR(Values: array of pointer): boolean;
begin
  result := PDouble(Values[0])^ <> PDouble(Values[1])^;
end;

function _NotEqualS(Values: array of pointer): boolean;
var
  String1, String2: string;
begin
  String1 := PString(Values[0])^;
  String2 := PString(Values[1])^;
  result := String1 <> String2;
end;

function _LessThanB(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ < PBoolean(Values[1])^;
end;

function _LessThanI(Values: array of pointer): boolean;
begin
  result := PInteger(Values[0])^ < PInteger(Values[1])^;
end;

function _LessThanR(Values: array of pointer): boolean;
begin
  result := PDouble(Values[0])^ < PDouble(Values[1])^;
end;

function _LessThanS(Values: array of pointer): boolean;
var
  String1, String2: String;
begin
  String1 := PString(Values[0])^;
  String2 := PString(Values[1])^;
  result := AnsiCompareStr(String1, String2) < 0;
end;

function _GreaterThanB(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ > PBoolean(Values[1])^;
end;

function _GreaterThanI(Values: array of pointer): boolean;
begin
  result := PInteger(Values[0])^ > PInteger(Values[1])^;
end;

function _GreaterThanR(Values: array of pointer): boolean;
begin
  result := PDouble(Values[0])^ > PDouble(Values[1])^;
end;

function _GreaterThanS(Values: array of pointer): boolean;
var
  String1, String2: String;
begin
  String1 := PString(Values[0])^;
  String2 := PString(Values[1])^;
  result := AnsiCompareStr(String1, String2) > 0;
end;

function _LessThanOrEqualsB(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ <= PBoolean(Values[1])^;
end;

function _LessThanOrEqualsI(Values: array of pointer): boolean;
begin
  result := PInteger(Values[0])^ <= PInteger(Values[1])^;
end;

function _LessThanOrEqualsR(Values: array of pointer): boolean;
begin
  result := PDouble(Values[0])^ <= PDouble(Values[1])^;
end;

function _LessThanOrEqualsS(Values: array of pointer): boolean;
var
  String1, String2: string;
begin
  String1 := PString(Values[0])^;
  String2 := PString(Values[1])^;
  result := AnsiCompareStr(String1, String2) <= 0;
end;

function _GreaterThanOrEqualsB(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[0])^ >= PBoolean(Values[1])^;
end;

function _GreaterThanOrEqualsI(Values: array of pointer): boolean;
begin
  result := PInteger(Values[0])^ >= PInteger(Values[1])^;
end;

function _GreaterThanOrEqualsR(Values: array of pointer): boolean;
begin
  result := PDouble(Values[0])^ >= PDouble(Values[1])^;
end;

function _GreaterThanOrEqualsS(Values: array of pointer): boolean;
var
  String1, String2: String;
begin
  String1 := PString(Values[0])^;
  String2 := PString(Values[1])^;
  result := AnsiCompareStr(String1, String2) >= 0;
end;

function _IntToDouble(Values: array of pointer): double;
begin
  result := PInteger(Values[0])^;
end;

function _AbsR(Values: array of pointer): double;
begin
  result := Abs(PDouble(Values[0])^);
end;

function _AbsI(Values: array of pointer): integer;
begin
  result := Abs(PInteger(Values[0])^);
end;

function _Arccos(Values: array of pointer): double;
begin
  result := ArcCos(PDouble(Values[0])^);
end;

function _Arccosh(Values: array of pointer): double;
begin
  result := ArcCosh(PDouble(Values[0])^);
end;

function _Arcsin(Values: array of pointer): double;
begin
  result := ArcSin(PDouble(Values[0])^);
end;

function _Arcsinh(Values: array of pointer): double;
begin
  result := ArcSinh(PDouble(Values[0])^);
end;

function _Arctan2(Values: array of pointer): double;
begin
  result := ArcTan2(PDouble(Values[0])^, PDouble(Values[1])^);
end;

function _Arctanh(Values: array of pointer): double;
begin
  result := ArcTanh(PDouble(Values[0])^);
end;

function _Cos(Values: array of pointer): double;
begin
  result := Cos(PDouble(Values[0])^);
end;

function _Cosh(Values: array of pointer): double;
begin
  result := Cosh(PDouble(Values[0])^);
end;

function _Copy(Values: array of pointer): string;
begin
  result := Copy(PString(Values[0])^, PInteger(Values[1])^,
    PInteger(Values[2])^);
end;

function _DegToRad(Values: array of pointer): double;
begin
  result := DegToRad(PDouble(Values[0])^);
end;

function _Factorial(Values: array of pointer): integer;
var
  StartValue: integer;
begin
  StartValue := PInteger(Values[0])^;
  if StartValue <= 1 then
  begin
    result := 1;
  end
  else if StartValue > 12 then
  begin
    raise EIntOverflow.Create('Factorial for values greater than 12 '
      + 'can not be computed by the Factorial function');
  end
  else
  begin
    result := StartValue;
    while (StartValue > 1) do
    begin
      dec(StartValue);
      result := result * StartValue;
    end;
  end;
end;

function _FactorialR(Values: array of pointer): double;
var
  StartValue: integer;
begin
  StartValue := PInteger(Values[0])^;
  if StartValue <= 1 then
  begin
    result := 1;
  end
  else if StartValue > 170 then
  begin
    raise EOverflow.Create('Factorial for values greater than 170 '
      + 'can not be computed by the FactorialR function');
  end
  else
  begin
    result := StartValue;
    while (StartValue > 1) do
    begin
      dec(StartValue);
      result := result * StartValue;
    end;
  end;
end;

function _Frac(Values: array of pointer): double;
begin
  result := Frac(PDouble(Values[0])^);
end;

function _FloatToStr(Values: array of pointer): string;
var
  OldDecimalSeparator: Char;
begin
  OldDecimalSeparator := DecimalSeparator;
  try
    DecimalSeparator := '.';
    result := FloatToStr(PDouble(Values[0])^);
  finally
    DecimalSeparator := OldDecimalSeparator;
  end;
end;

function _CaseBoolean(Values: array of pointer): boolean;
begin
  result := PBoolean(Values[PInteger(Values[0])^])^;
end;

function _CaseInteger(Values: array of pointer): integer;
begin
  result := PInteger(Values[PInteger(Values[0])^])^;
end;

function _CaseDouble(Values: array of pointer): double;
begin
  result := PDouble(Values[PInteger(Values[0])^])^;
end;

function _CaseString(Values: array of pointer): string;
begin
  result := PString(Values[PInteger(Values[0])^])^;
end;

function _IfBoolean(Values: array of pointer): boolean;
begin
  if PBoolean(Values[0])^ then
  begin
    result := PBoolean(Values[1])^;
  end
  else
  begin
    result := PBoolean(Values[2])^;
  end;
end;

function _IfInteger(Values: array of pointer): integer;
begin
  if PBoolean(Values[0])^ then
  begin
    result := PInteger(Values[1])^;
  end
  else
  begin
    result := PInteger(Values[2])^;
  end;
end;

function _IfDouble(Values: array of pointer): double;
begin
  if PBoolean(Values[0])^ then
  begin
    result := PDouble(Values[1])^;
  end
  else
  begin
    result := PDouble(Values[2])^;
  end;
end;

function _IfString(Values: array of pointer): string;
begin
  if PBoolean(Values[0])^ then
  begin
    result := PString(Values[1])^;
  end
  else
  begin
    result := PString(Values[2])^;
  end;
end;

function _IntPower(Values: array of pointer): double;
begin
  result := IntPower(PDouble(Values[0])^, PInteger(Values[1])^);
end;

function _IntToStr(Values: array of pointer): string;
begin
  result := IntToStr(PInteger(Values[0])^);
end;

function _Length(Values: array of pointer): integer;
begin
  result := Length(PString(Values[0])^);
end;

function _ln(Values: array of pointer): double;
begin
  result := ln(PDouble(Values[0])^);
end;

function _log10(Values: array of pointer): double;
begin
  result := log10(PDouble(Values[0])^);
end;

function _logN(Values: array of pointer): double;
begin
  result := logN(PDouble(Values[0])^, PDouble(Values[1])^);
end;

function _LowerCase(Values: array of pointer): string;
begin
  result := LowerCase(PString(Values[0])^);
end;

function _MaxI(Values: array of pointer): integer;
var
  Index: integer;
begin
  result := Max(PInteger(Values[0])^, PInteger(Values[1])^);
  for Index := 2 to Length(Values) - 1 do
  begin
    result := Max(result, PInteger(Values[Index])^);
  end;
end;

function _MaxR(Values: array of pointer): double;
var
  Index: integer;
begin
  result := Max(PDouble(Values[0])^, PDouble(Values[1])^);
  for Index := 2 to Length(Values) - 1 do
  begin
    result := Max(result, PDouble(Values[Index])^);
  end;
end;

function _MinI(Values: array of pointer): integer;
var
  Index: integer;
begin
  result := Min(PInteger(Values[0])^, PInteger(Values[1])^);
  for Index := 2 to Length(Values) - 1 do
  begin
    result := Min(result, PInteger(Values[Index])^);
  end;
end;

function _MinR(Values: array of pointer): double;
var
  Index: integer;
begin
  result := Min(PDouble(Values[0])^, PDouble(Values[1])^);
  for Index := 2 to Length(Values) - 1 do
  begin
    result := Min(result, PDouble(Values[Index])^);
  end;
end;

function _MultiInterpolate(Values: array of pointer): double;
begin
  result := 0;
  Assert(False);
end;

function _Odd(Values: array of pointer): boolean;
begin
  result := Odd(PInteger(Values[0])^);
end;

function _Pos(Values: array of pointer): integer;
begin
  result := Pos(PString(Values[0])^, PString(Values[1])^);
end;

function _PosEx(Values: array of pointer): integer;
begin
  result := PosEx(PString(Values[0])^, PString(Values[1])^,
    PInteger(Values[2])^);
end;

function _Power(Values: array of pointer): double;
begin
  result := Power(PDouble(Values[0])^, PDouble(Values[1])^);
end;

function _Pi(Values: array of pointer): double;
begin
  result := Pi;
end;

function _RadToDeg(Values: array of pointer): double;
begin
  result := RadToDeg(PDouble(Values[0])^);
end;

function _Round(Values: array of pointer): integer;
begin
  result := Round(PDouble(Values[0])^);
end;

function _Sin(Values: array of pointer): double;
begin
  result := Sin(PDouble(Values[0])^);
end;

function _Sinh(Values: array of pointer): double;
begin
  result := Sinh(PDouble(Values[0])^);
end;

function _SqrR(Values: array of pointer): double;
begin
  result := sqr(PDouble(Values[0])^);
end;

function _SqrI(Values: array of pointer): integer;
begin
  result := sqr(PInteger(Values[0])^);
end;

function _Sqrt(Values: array of pointer): double;
begin
  result := sqrt(PDouble(Values[0])^);
end;

function _StrToInt(Values: array of pointer): integer;
begin
  result := StrToInt(PString(Values[0])^);
end;

function _StrToIntDef(Values: array of pointer): integer;
begin
  result := StrToIntDef(PString(Values[0])^, PInteger(Values[1])^);
end;

function _StrToFloat(Values: array of pointer): double;
var
  AString: string;
  Code: Integer;
begin
  AString := PString(Values[0])^;
  Val(AString, result, Code);
  if Code <> 0 then
  begin
    raise EConvertError.Create(AString +
      ' can not be converted to a real number.');
  end;
end;

function _StrToFloatDef(Values: array of pointer): double;
var
  AString: string;
  Code: Integer;
begin
  AString := PString(Values[0])^;
  Val(AString, result, Code);
  if Code <> 0 then
  begin
    result := PDouble(Values[1])^;
  end;
end;

function _Tan(Values: array of pointer): double;
begin
  result := tan(PDouble(Values[0])^);
end;

function _Tanh(Values: array of pointer): double;
begin
  result := tanh(PDouble(Values[0])^);
end;

function _Trunc(Values: array of pointer): integer;
begin
  result := trunc(PDouble(Values[0])^);
end;

function _UpperCase(Values: array of pointer): string;
begin
  result := UpperCase(PString(Values[0])^);
end;

function _Interpolate(Values: array of pointer): double;
var
  Position: double;
  Value1: double;
  Distance1: double;
  Value2: double;
  Distance2: double;
begin
  Position := PDouble(Values[0])^;
  Value1 := PDouble(Values[1])^;
  Distance1 := PDouble(Values[2])^;
  Value2 := PDouble(Values[3])^;
  Distance2 := PDouble(Values[4])^;
  result := (Position - Distance1) / (Distance2 - Distance1)
    * (Value2 - Value1) + Value1;
end;

function _Distance(Values: array of pointer): double;
var
  X1: double;
  Y1: double;
  X2: double;
  Y2: double;
begin
  X1 := PDouble(Values[0])^;
  Y1 := PDouble(Values[1])^;
  X2 := PDouble(Values[2])^;
  Y2 := PDouble(Values[3])^;
  result := Sqrt(Sqr(X1 - X2) + Sqr(Y1 - Y2));
end;

class function TExpression.New(
  const FunctionClass: TFunctionClass): TExpression;
var
  Index: integer;
  SpecialImplementor: TSpecialImplementor;
begin
  if (FunctionClass = IntToDoubleFunction)
    or (FunctionClass.FunctionRecord.Name = IntToDoubleFunction.Name) then
  begin
    result := TIntToDoubleExpression.Create(FunctionClass);
  end
  else if FunctionClass = AndOperator then
  begin
    result := TAndExpression.Create(FunctionClass);
  end
  else if FunctionClass = OrOperator then
  begin
    result := TOrExpression.Create(FunctionClass);
  end
  else if FunctionClass = XorOperator then
  begin
    result := TOperator.Create(FunctionClass);
  end
  else if (FunctionClass = CaseBOverloadedFunction)
    or (FunctionClass = CaseIOverloadedFunction)
    or (FunctionClass = CaseROverloadedFunction)
    or (FunctionClass = CaseSOverloadedFunction)
    or (FunctionClass = IfBOverloadedFunction)
    or (FunctionClass = IfIOverloadedFunction)
    or (FunctionClass = IfROverloadedFunction)
    or (FunctionClass = IfSOverloadedFunction)
    or (FunctionClass.FunctionRecord.Name = IfBooleanFunction.Name)
    or (FunctionClass.FunctionRecord.Name = IfIntegerFunction.Name)
    or (FunctionClass.FunctionRecord.Name = IfRealFunction.Name)
    or (FunctionClass.FunctionRecord.Name = IfStringFunction.Name)
    or (FunctionClass.FunctionRecord.Name = CaseBooleanFunction.Name)
    or (FunctionClass.FunctionRecord.Name = CaseIntegerFunction.Name)
    or (FunctionClass.FunctionRecord.Name = CaseDoubleFunction.Name)
    or (FunctionClass.FunctionRecord.Name = CaseStringFunction.Name) then
  begin
    result := TSelectExpression.Create(FunctionClass);
  end
  else if (FunctionClass = MultiInterpolateFunction) then
  begin
    result := TMultiInterpolateExpression.Create(FunctionClass);
  end
  else
  begin
    for Index := 0 to SpecialImplentorList.Count - 1 do
    begin
      SpecialImplementor := SpecialImplentorList[Index];
      if (SpecialImplementor.FunctionClass = FunctionClass) then
      begin
        result := SpecialImplementor.Implementor.Create(
          SpecialImplementor.FunctionClass);
        Exit;
      end;
    end;
    result := TExpression.Create(FunctionClass);
  end;
end;

constructor TExpression.Create(const VariableName: string;
  const DataType: TRbwDataType; const CanConvertToConstant: boolean);
begin
  Create(VariableName, DataType);
  FunctionAddr := nil;
  ShouldEvaluate := False;
  AllowConversionToConstant := CanConvertToConstant;
end;

procedure TExpression.SetAllowConversionToConstant(const Value: boolean);
begin
  FAllowConversionToConstant := Value;
end;

function TExpression.Decompile: string;
var
  Index: integer;
  ArrayLength: integer;
  AVariable: TConstant;
  NeedsParenthesis: boolean;
  ParametersPresent: boolean;
begin
  result := FUserName;
  ArrayLength := Length(Data);
  if (ArrayLength > 0) and Assigned(FunctionAddr) then
  begin
    NeedsParenthesis := True;
    if ArrayLength = 1 then
    begin
      AVariable := Data[0].Datum;
      if AVariable is TOperator then
      begin
        if Length(TOperator(AVariable).Data) = 2 then
        begin
          NeedsParenthesis := False;
        end;
      end;
    end;
    ParametersPresent := False;
    for Index := 0 to ArrayLength - 1 do
    begin
      if Data[Index].Datum <> nil then
      begin
        ParametersPresent := True;
        Break;
      end;
    end;

    if NeedsParenthesis and ParametersPresent then
    begin
      result := result + '(';
    end;
    for Index := 0 to ArrayLength - 1 do
    begin
      if Data[Index].Datum <> nil then
      begin
        if Index > 0 then
        begin
          result := result + ', ';
        end;
        AVariable := Data[Index].Datum;
        if AVariable is TExpression then
        begin
          TExpression(AVariable).FTopLevel := False;
        end;

        result := result + AVariable.Decompile;
      end;
    end;
    if NeedsParenthesis and ParametersPresent then
    begin
      result := result + ')';
    end;
  end;
end;

function TExpression.UsesVariable(const Variable: TCustomVariable): boolean;
var
  Index: integer;
  ArrayLength: integer;
begin
  Assert(Variable <> nil);
  result := False;
  ArrayLength := Length(Data);
  if ArrayLength > 0 then
  begin
    for Index := 0 to ArrayLength - 1 do
    begin
      result := Data[Index].Datum = Variable;
      if result then
      begin
        Exit;
      end;
    end;
  end;
end;

procedure TExpression.FillVariables;
var
  Index: integer;
  AVariable: TConstant;
begin
  StringVariableCount := 0;
  Assert(Length(VariablesForFunction) = Length(Data));
  SetLength(StringVariableIndicies, Length(Data));

  for Index := 0 to Length(Data) - 1 do
  begin
    if Data[Index].Datum = nil then
    begin
      VariablesForFunction[Index] := nil;
    end
    else
    begin
      AVariable := Data[Index].Datum;
      VariablesForFunction[Index] := AVariable.FResult;
      if (AVariable.FResultType = rdtString)
        and (AVariable is TCustomValue) then
      begin
        StringVariableIndicies[StringVariableCount] := Index;
        Inc(StringVariableCount);
      end;
      if AVariable is TExpression then
      begin
        TExpression(AVariable).FillVariables;
      end;
    end;
  end;
  SetLength(StringVariableIndicies, StringVariableCount);
end;

function TExpression.GetVariablesUsed: TStringList;
var
  Index: integer;
  ArrayLength: integer;
  AVariable: TConstant;
  InnerIndex: integer;
  TempList: TStringList;
begin
  FVariablesUsed.Clear;
  result := FVariablesUsed;
  ArrayLength := Length(Data);
  if (ArrayLength > 0) {and Assigned(FunctionAddr)} then
  begin
    for Index := 0 to ArrayLength - 1 do
    begin
      if Data[Index].Datum <> nil then
      begin
        AVariable := Data[Index].Datum;
        if AVariable is TExpression then
        begin
          TempList := TExpression(AVariable).VariablesUsed;
          for InnerIndex := 0 to TempList.Count - 1 do
          begin
            result.Add(TempList[InnerIndex]);
          end;
        end
        else if AVariable is TCustomValue then
        begin
          result.Add(AVariable.Decompile);
        end;
      end;
    end;
  end;
end;

constructor TExpression.Create(const VariableName: string;
  const DataType: TRbwDataType);
begin
  inherited Create(VariableName, DataType);
  FVariablesUsed := TStringList.Create;
  FVariablesUsed.Sorted := True;
  FVariablesUsed.Duplicates := dupIgnore;
  FTopLevel := True;
end;

{ TFunctionClass }

constructor TFunctionClass.Create;
begin
  AllowConversionToConstant := True;
end;

function TFunctionClass.GetAllowConversionToConstant: boolean;
begin
  result := FunctionRecord.CanConvertToConstant;
end;

function TFunctionClass.GetBFunctionAddr: TrbwBooleanFunction;
begin
  if FunctionRecord.ResultType <> rdtBoolean then
  begin
    raise ErbwParserError.Create(
      'Error: The function is of the incorrect type.');
  end;
  result := FunctionRecord.BFunctionAddr;
end;

function TFunctionClass.GetHidden: boolean;
begin
  result := FunctionRecord.Hidden;
end;

function TFunctionClass.GetIFunctionAddr: TrbwIntegerFunction;
begin
  if FunctionRecord.ResultType <> rdtInteger then
  begin
    raise ErbwParserError.Create(
      'Error: The function is of the incorrect type.');
  end;
  result := FunctionRecord.IFunctionAddr;
end;

function TFunctionClass.GetInputDataCount: integer;
begin
  result := Length(FunctionRecord.InputDataTypes);
end;

function TFunctionClass.GetInputDataTypes(const Index: integer): TRbwDataType;
begin
  if (Index < 0) or (Index >= InputDataCount) then
  begin
    raise ERangeError.Create(IntToStr(Index) + ' is out of range.');
  end;
  result := FunctionRecord.InputDataTypes[Index];
end;

function TFunctionClass.GetName: string;
begin
  result := FunctionRecord.Name
end;

function TFunctionClass.GetOptionalArguments: integer;
begin
  result := FunctionRecord.OptionalArguments;
end;

function TFunctionClass.GetResultType: TRbwDataType;
begin
  result := FunctionRecord.ResultType;
end;

function TFunctionClass.GetRFunctionAddr: TrbwRealFunction;
begin
  if FunctionRecord.ResultType <> rdtDouble then
  begin
    raise ErbwParserError.Create(
      'Error: The function is of the incorrect type.');
  end;
  result := FunctionRecord.RFunctionAddr;
end;

function TFunctionClass.GetSFunctionAddr: TrbwStringFunction;
begin
  if FunctionRecord.ResultType <> rdtString then
  begin
    raise ErbwParserError.Create(
      'Error: The function is of the incorrect type.');
  end;
  result := FunctionRecord.SFunctionAddr;
end;

procedure TFunctionClass.SetAllowConversionToConstant(
  const Value: boolean);
begin
  FunctionRecord.CanConvertToConstant := Value;
end;

procedure TFunctionClass.SetBFunctionAddr(
  const Value: TrbwBooleanFunction);
begin
  FunctionRecord.BFunctionAddr := Value;
  FunctionRecord.ResultType := rdtBoolean;
end;

procedure TFunctionClass.SetHidden(const Value: boolean);
begin
  FunctionRecord.Hidden := Value
end;

procedure TFunctionClass.SetIFunctionAddr(
  const Value: TrbwIntegerFunction);
begin
  FunctionRecord.IFunctionAddr := Value;
  FunctionRecord.ResultType := rdtInteger;
end;

procedure TFunctionClass.SetInputDataCount(const Value: integer);
begin
  SetLength(FunctionRecord.InputDataTypes, Value);
end;

procedure TFunctionClass.SetInputDataTypes(const Index: integer;
  const Value: TRbwDataType);
begin
  if (Index < 0) or (Index >= InputDataCount) then
  begin
    raise ERangeError.Create(IntToStr(Index) + ' is out of range.');
  end;
  FunctionRecord.InputDataTypes[Index] := Value;
end;

procedure TFunctionClass.SetName(const Value: string);
begin
  FunctionRecord.Name := Value;
end;

procedure TFunctionClass.SetOptionalArguments(const Value: integer);
begin
  FunctionRecord.OptionalArguments := Value;
end;

procedure TFunctionClass.SetRFunctionAddr(const Value: TrbwRealFunction);
begin
  FunctionRecord.RFunctionAddr := Value;
  FunctionRecord.ResultType := rdtDouble;
end;

procedure TFunctionClass.SetSFunctionAddr(const Value: TrbwStringFunction);
begin
  FunctionRecord.SFunctionAddr := Value;
  FunctionRecord.ResultType := rdtString;
end;

{ TFunctionStringList }

function TFunctionStringList.Add(
  const FunctionRecord: TFunctionRecord): Integer;
var
  AFunctionClass: TFunctionClass;
  IIndex: integer;
  ArrayLength: integer;
begin
  AFunctionClass := TFunctionClass.Create;
  try
    ArrayLength := Length(FunctionRecord.InputDataTypes);
    SetLength(AFunctionClass.FunctionRecord.InputDataTypes, ArrayLength);
    for IIndex := 0 to ArrayLength - 1 do
    begin
      AFunctionClass.FunctionRecord.InputDataTypes[IIndex] :=
        FunctionRecord.InputDataTypes[IIndex];
    end;
    AFunctionClass.FunctionRecord.Synonyms := FunctionRecord.Synonyms;
    SetLength(AFunctionClass.FunctionRecord.Synonyms,
      Length(AFunctionClass.FunctionRecord.Synonyms));
    AFunctionClass.Hidden := FunctionRecord.Hidden;
    AFunctionClass.Prototype := FunctionRecord.Prototype;
    AFunctionClass.FunctionRecord.ResultType := FunctionRecord.ResultType;
    AFunctionClass.FunctionRecord.Name := FunctionRecord.Name;
    AFunctionClass.FunctionRecord.Prototype := FunctionRecord.Prototype;
    AFunctionClass.FunctionRecord.OptionalArguments :=
      FunctionRecord.OptionalArguments;
    AFunctionClass.FunctionRecord.CanConvertToConstant :=
      FunctionRecord.CanConvertToConstant;
    AFunctionClass.FunctionRecord.RFunctionAddr := FunctionRecord.RFunctionAddr;
    Assert(Assigned(AFunctionClass.FunctionRecord.RFunctionAddr));
    result := AddObject(FunctionRecord.Name, AFunctionClass);
    for IIndex := 0 to Length(AFunctionClass.FunctionRecord.Synonyms) -1 do
    begin
      AddObject(AFunctionClass.FunctionRecord.Synonyms[IIndex], AFunctionClass);
    end;

  except on EStringListError do
    begin
      AFunctionClass.Free;
      raise ErbwParserError.Create('Error: A function named ' +
        FunctionRecord.Name
        + ' already exists.');
    end;
  end;
end;

procedure TFunctionStringList.Clear;
var
  Index: integer;
  AFunctionClass: TFunctionClass;
begin
  // delete synonyms
  for Index := Count - 1 downto 0 do
  begin
    AFunctionClass := Objects[Index] as TFunctionClass;
    if AFunctionClass.FunctionRecord.Name <> Strings[Index] then
    begin
      inherited Delete(Index);
    end;
  end;

  // Free all non-synonyms.
  for Index := 0 to Count - 1 do
  begin
    AFunctionClass := Objects[Index] as TFunctionClass;
    AFunctionClass.Free;
  end;
  inherited;
end;

constructor TFunctionStringList.Create;
begin
  inherited;
  CaseSensitive := False;
  Duplicates := dupError;
  Sorted := True;

  AbsFunctionI.ResultType := rdtInteger;
  AbsFunctionI.Name := 'AbsI';
  AbsFunctionI.Prototype := 'Math|AbsI(Value)';
  SetLength(AbsFunctionI.InputDataTypes, 1);
  AbsFunctionI.InputDataTypes[0] := rdtInteger;
  AbsFunctionI.OptionalArguments := 0;
  AbsFunctionI.CanConvertToConstant := True;
  AbsFunctionI.IFunctionAddr := _AbsI;
  Add(AbsFunctionI);

  AbsFunctionR.ResultType := rdtDouble;
  AbsFunctionR.Name := 'AbsR';
  AbsFunctionR.Prototype := 'Math|AbsR(Value)';
  SetLength(AbsFunctionR.InputDataTypes, 1);
  AbsFunctionR.InputDataTypes[0] := rdtDouble;
  AbsFunctionR.OptionalArguments := 0;
  AbsFunctionR.CanConvertToConstant := True;
  AbsFunctionR.RFunctionAddr := _AbsR;
  Add(AbsFunctionR);

  ArcCosFunction.ResultType := rdtDouble;
  ArcCosFunction.Name := 'ArcCos';
  ArcCosFunction.Prototype := 'Trig|ArcCos(Value)';
  SetLength(ArcCosFunction.InputDataTypes, 1);
  ArcCosFunction.InputDataTypes[0] := rdtDouble;
  ArcCosFunction.OptionalArguments := 0;
  ArcCosFunction.CanConvertToConstant := True;
  ArcCosFunction.RFunctionAddr := _arccos;
  Add(ArcCosFunction);

  ArcCoshFunction.ResultType := rdtDouble;
  ArcCoshFunction.Name := 'ArcCosh';
  ArcCoshFunction.Prototype := 'Trig|ArcCosh(Value)';
  SetLength(ArcCoshFunction.InputDataTypes, 1);
  ArcCoshFunction.InputDataTypes[0] := rdtDouble;
  ArcCoshFunction.OptionalArguments := 0;
  ArcCoshFunction.CanConvertToConstant := True;
  ArcCoshFunction.RFunctionAddr := _arccosh;
  Add(ArcCoshFunction);

  ArcSinFunction.ResultType := rdtDouble;
  ArcSinFunction.Name := 'ArcSin';
  ArcSinFunction.Prototype := 'Trig|ArcSin(Value)';
  SetLength(ArcSinFunction.InputDataTypes, 1);
  ArcSinFunction.InputDataTypes[0] := rdtDouble;
  ArcSinFunction.OptionalArguments := 0;
  ArcSinFunction.CanConvertToConstant := True;
  ArcSinFunction.RFunctionAddr := _arcsin;
  Add(ArcSinFunction);

  ArcSinhFunction.ResultType := rdtDouble;
  ArcSinhFunction.Name := 'ArcSinh';
  ArcSinhFunction.Prototype := 'Trig|ArcSinh(Value)';
  SetLength(ArcSinhFunction.InputDataTypes, 1);
  ArcSinhFunction.InputDataTypes[0] := rdtDouble;
  ArcSinhFunction.OptionalArguments := 0;
  ArcSinhFunction.CanConvertToConstant := True;
  ArcSinhFunction.RFunctionAddr := _arcsinh;
  Add(ArcSinhFunction);

  ArcTan2Function.ResultType := rdtDouble;
  ArcTan2Function.Name := 'ArcTan2';
  ArcTan2Function.Prototype := 'Trig|ArcTan2(Y, X)';
  SetLength(ArcTan2Function.InputDataTypes, 2);
  ArcTan2Function.InputDataTypes[0] := rdtDouble;
  ArcTan2Function.InputDataTypes[1] := rdtDouble;
  ArcTan2Function.OptionalArguments := 0;
  ArcTan2Function.CanConvertToConstant := True;
  ArcTan2Function.RFunctionAddr := _arctan2;
  Add(ArcTan2Function);

  ArcTanhFunction.ResultType := rdtDouble;
  ArcTanhFunction.Name := 'ArcTanh';
  ArcTanhFunction.Prototype := 'Trig|ArcTanh(Value)';
  SetLength(ArcTanhFunction.InputDataTypes, 1);
  ArcTanhFunction.InputDataTypes[0] := rdtDouble;
  ArcTanhFunction.OptionalArguments := 0;
  ArcTanhFunction.CanConvertToConstant := True;
  ArcTanhFunction.RFunctionAddr := _arctanh;
  Add(ArcTanhFunction);

  CopyFunction.ResultType := rdtString;
  CopyFunction.Name := 'Copy';
  CopyFunction.Prototype := 'Text|Copy(text_Value, StartIndex, Count)';
  SetLength(CopyFunction.InputDataTypes, 3);
  CopyFunction.InputDataTypes[0] := rdtString;
  CopyFunction.InputDataTypes[1] := rdtInteger;
  CopyFunction.InputDataTypes[2] := rdtInteger;
  CopyFunction.OptionalArguments := 0;
  CopyFunction.CanConvertToConstant := True;
  CopyFunction.SFunctionAddr := _Copy;
  Add(CopyFunction);

  CosFunction.ResultType := rdtDouble;
  CosFunction.Name := 'Cos';
  CosFunction.Prototype := 'Trig|Cos(Value)';
  SetLength(CosFunction.InputDataTypes, 1);
  CosFunction.InputDataTypes[0] := rdtDouble;
  CosFunction.OptionalArguments := 0;
  CosFunction.CanConvertToConstant := True;
  CosFunction.RFunctionAddr := _cos;
  Add(CosFunction);

  CoshFunction.ResultType := rdtDouble;
  CoshFunction.Name := 'Cosh';
  CoshFunction.Prototype := 'Trig|Cosh(Value)';
  SetLength(CoshFunction.InputDataTypes, 1);
  CoshFunction.InputDataTypes[0] := rdtDouble;
  CoshFunction.OptionalArguments := 0;
  CoshFunction.CanConvertToConstant := True;
  CoshFunction.RFunctionAddr := _cosh;
  Add(CoshFunction);

  CaseBooleanFunction.ResultType := rdtBoolean;
  CaseBooleanFunction.Name := 'CaseB';
  CaseBooleanFunction.Prototype :=
    'Logical|CaseB(Index, boolean_Result1, boolean_Result2, ...)';
  SetLength(CaseBooleanFunction.InputDataTypes, 4);
  CaseBooleanFunction.InputDataTypes[0] := rdtInteger;
  CaseBooleanFunction.InputDataTypes[1] := rdtBoolean;
  CaseBooleanFunction.InputDataTypes[2] := rdtBoolean;
  CaseBooleanFunction.InputDataTypes[3] := rdtBoolean;
  CaseBooleanFunction.CanConvertToConstant := True;
  CaseBooleanFunction.OptionalArguments := -1;
  CaseBooleanFunction.BFunctionAddr := _CaseBoolean;
  Add(CaseBooleanFunction);

  CaseIntegerFunction.ResultType := rdtInteger;
  CaseIntegerFunction.Name := 'CaseI';
  CaseIntegerFunction.Prototype :=
    'Logical|CaseI(Index, integer_Result1, integer_Result2, ...)';
  SetLength(CaseIntegerFunction.InputDataTypes, 4);
  CaseIntegerFunction.InputDataTypes[0] := rdtInteger;
  CaseIntegerFunction.InputDataTypes[1] := rdtInteger;
  CaseIntegerFunction.InputDataTypes[2] := rdtInteger;
  CaseIntegerFunction.InputDataTypes[3] := rdtInteger;
  CaseIntegerFunction.OptionalArguments := -1;
  CaseIntegerFunction.CanConvertToConstant := True;
  CaseIntegerFunction.IFunctionAddr := _CaseInteger;
  Add(CaseIntegerFunction);

  CaseDoubleFunction.ResultType := rdtDouble;
  CaseDoubleFunction.Name := 'CaseR';
  CaseDoubleFunction.Prototype :=
    'Logical|CaseR(Index, real_Result1, real_Result2, ...)';
  SetLength(CaseDoubleFunction.InputDataTypes, 4);
  CaseDoubleFunction.InputDataTypes[0] := rdtInteger;
  CaseDoubleFunction.InputDataTypes[1] := rdtDouble;
  CaseDoubleFunction.InputDataTypes[2] := rdtDouble;
  CaseDoubleFunction.InputDataTypes[3] := rdtDouble;
  CaseDoubleFunction.OptionalArguments := -1;
  CaseDoubleFunction.CanConvertToConstant := True;
  CaseDoubleFunction.RFunctionAddr := _CaseDouble;
  Add(CaseDoubleFunction);

  CaseStringFunction.ResultType := rdtString;
  CaseStringFunction.Name := 'CaseT';
  CaseStringFunction.Prototype :=
    'Logical|CaseT(Index, text_Result1, text_Result2, ...)';
  SetLength(CaseStringFunction.InputDataTypes, 4);
  CaseStringFunction.InputDataTypes[0] := rdtInteger;
  CaseStringFunction.InputDataTypes[1] := rdtString;
  CaseStringFunction.InputDataTypes[2] := rdtString;
  CaseStringFunction.InputDataTypes[3] := rdtString;
  CaseStringFunction.OptionalArguments := -1;
  CaseStringFunction.CanConvertToConstant := True;
  CaseStringFunction.SFunctionAddr := _CaseString;
  SetLength(CaseStringFunction.Synonyms, 1);
  CaseStringFunction.Synonyms[0] := 'CaseS';
  Add(CaseStringFunction);

  DegToRadFunction.ResultType := rdtDouble;
  DegToRadFunction.Name := 'DegToRad';
  DegToRadFunction.Prototype := 'Trig|DegToRad(Value)';
  SetLength(DegToRadFunction.InputDataTypes, 1);
  DegToRadFunction.InputDataTypes[0] := rdtDouble;
  DegToRadFunction.OptionalArguments := 0;
  DegToRadFunction.CanConvertToConstant := True;
  DegToRadFunction.RFunctionAddr := _DegToRad;
  Add(DegToRadFunction);

  FactorialFunction.ResultType := rdtInteger;
  FactorialFunction.Name := 'FactorialI';
  FactorialFunction.Prototype := 'Math|FactorialI(Value_Less_than_13)';
  SetLength(FactorialFunction.InputDataTypes, 1);
  FactorialFunction.InputDataTypes[0] := rdtInteger;
  FactorialFunction.OptionalArguments := 0;
  FactorialFunction.CanConvertToConstant := True;
  FactorialFunction.IFunctionAddr := _Factorial;
  Add(FactorialFunction);

  FactorialFFunction.ResultType := rdtDouble;
  FactorialFFunction.Name := 'FactorialR';
  FactorialFFunction.Prototype := 'Math|FactorialR(Value_Less_than_171)';
  SetLength(FactorialFFunction.InputDataTypes, 1);
  FactorialFFunction.InputDataTypes[0] := rdtInteger;
  FactorialFFunction.OptionalArguments := 0;
  FactorialFFunction.CanConvertToConstant := True;
  FactorialFFunction.RFunctionAddr := _FactorialR;
  Add(FactorialFFunction);

  FracFunction.ResultType := rdtDouble;
  FracFunction.Name := 'Frac';
  FracFunction.Prototype := 'Math|Frac(Value)';
  SetLength(FracFunction.InputDataTypes, 1);
  FracFunction.InputDataTypes[0] := rdtDouble;
  FracFunction.OptionalArguments := 0;
  FracFunction.CanConvertToConstant := True;
  FracFunction.RFunctionAddr := _Frac;
  Add(FracFunction);

  FloatToStrFunction.ResultType := rdtString;
  FloatToStrFunction.Name := 'FloatToText';
  FloatToStrFunction.Prototype := 'Text|FloatToText(Value)';
  SetLength(FloatToStrFunction.InputDataTypes, 1);
  FloatToStrFunction.InputDataTypes[0] := rdtDouble;
  FloatToStrFunction.OptionalArguments := 0;
  FloatToStrFunction.CanConvertToConstant := True;
  FloatToStrFunction.SFunctionAddr := _FloatToStr;
  SetLength(FloatToStrFunction.Synonyms, 1);
  FloatToStrFunction.Synonyms[0] := 'FloatToStr';
  Add(FloatToStrFunction);

  IfBooleanFunction.ResultType := rdtBoolean;
  IfBooleanFunction.Name := 'IfB';
  IfBooleanFunction.Prototype :=
    'Logical|IfB(Boolean_Value, If_True_boolean_Result, If_False_boolean_Result)';
  SetLength(IfBooleanFunction.InputDataTypes, 3);
  IfBooleanFunction.InputDataTypes[0] := rdtBoolean;
  IfBooleanFunction.InputDataTypes[1] := rdtBoolean;
  IfBooleanFunction.InputDataTypes[2] := rdtBoolean;
  IfBooleanFunction.OptionalArguments := 0;
  IfBooleanFunction.CanConvertToConstant := True;
  IfBooleanFunction.BFunctionAddr := _IfBoolean;
  Add(IfBooleanFunction);

  IfIntegerFunction.ResultType := rdtInteger;
  IfIntegerFunction.Name := 'IfI';
  IfIntegerFunction.Prototype :=
    'Logical|IfI(Boolean_Value, If_True_integer_Result, If_False_integer_Result)';
  SetLength(IfIntegerFunction.InputDataTypes, 3);
  IfIntegerFunction.InputDataTypes[0] := rdtBoolean;
  IfIntegerFunction.InputDataTypes[1] := rdtInteger;
  IfIntegerFunction.InputDataTypes[2] := rdtInteger;
  IfIntegerFunction.OptionalArguments := 0;
  IfIntegerFunction.CanConvertToConstant := True;
  IfIntegerFunction.IFunctionAddr := _IfInteger;
  Add(IfIntegerFunction);

  IfRealFunction.ResultType := rdtDouble;
  IfRealFunction.Name := 'IfR';
  IfRealFunction.Prototype :=
    'Logical|IfR(Boolean_Value, If_True_real_Result, If_False_real_Result)';
  SetLength(IfRealFunction.InputDataTypes, 3);
  IfRealFunction.InputDataTypes[0] := rdtBoolean;
  IfRealFunction.InputDataTypes[1] := rdtDouble;
  IfRealFunction.InputDataTypes[2] := rdtDouble;
  IfRealFunction.OptionalArguments := 0;
  IfRealFunction.CanConvertToConstant := True;
  IfRealFunction.RFunctionAddr := _IfDouble;
  Add(IfRealFunction);

  IfStringFunction.ResultType := rdtString;
  IfStringFunction.Name := 'IfT';
  IfStringFunction.Prototype :=
    'Logical|IfT(Boolean_Value, If_True_text_Result, If_False_text_Result)';
  SetLength(IfStringFunction.InputDataTypes, 3);
  IfStringFunction.InputDataTypes[0] := rdtBoolean;
  IfStringFunction.InputDataTypes[1] := rdtString;
  IfStringFunction.InputDataTypes[2] := rdtString;
  IfStringFunction.OptionalArguments := 0;
  IfStringFunction.CanConvertToConstant := True;
  IfStringFunction.SFunctionAddr := _IfString;
  SetLength(IfStringFunction.Synonyms, 1);
  IfStringFunction.Synonyms[0] := 'IfS';
  Add(IfStringFunction);

  IntPowerFunction.ResultType := rdtDouble;
  IntPowerFunction.Name := 'IntPower';
  IntPowerFunction.Prototype := 'Math|IntPower(Base, Exponent)';
  SetLength(IntPowerFunction.InputDataTypes, 2);
  IntPowerFunction.InputDataTypes[0] := rdtDouble;
  IntPowerFunction.InputDataTypes[1] := rdtInteger;
  IntPowerFunction.OptionalArguments := 0;
  IntPowerFunction.CanConvertToConstant := True;
  IntPowerFunction.RFunctionAddr := _IntPower;
  Add(IntPowerFunction);

  IntToStrFunction.ResultType := rdtString;
  IntToStrFunction.Name := 'IntToText';
  IntToStrFunction.Prototype := 'Text|IntToText(Value)';
  SetLength(IntToStrFunction.InputDataTypes, 1);
  IntToStrFunction.InputDataTypes[0] := rdtInteger;
  IntToStrFunction.OptionalArguments := 0;
  IntToStrFunction.CanConvertToConstant := True;
  IntToStrFunction.SFunctionAddr := _IntToStr;
  SetLength(IntToStrFunction.Synonyms, 1);
  IntToStrFunction.Synonyms[0] := 'IntToStr';
  Add(IntToStrFunction);

  LengthFunction.ResultType := rdtInteger;
  LengthFunction.Name := 'Length';
  LengthFunction.Prototype := 'Text|Length(text_Value)';
  SetLength(LengthFunction.InputDataTypes, 1);
  LengthFunction.InputDataTypes[0] := rdtString;
  LengthFunction.OptionalArguments := 0;
  LengthFunction.CanConvertToConstant := True;
  LengthFunction.IFunctionAddr := _length;
  Add(LengthFunction);

  LnFunction.ResultType := rdtDouble;
  LnFunction.Name := 'ln';
  LnFunction.Prototype := 'Math|ln(Value)';
  SetLength(LnFunction.InputDataTypes, 1);
  LnFunction.InputDataTypes[0] := rdtDouble;
  LnFunction.OptionalArguments := 0;
  LnFunction.CanConvertToConstant := True;
  LnFunction.RFunctionAddr := _ln;
  Add(LnFunction);

  Log10Function.ResultType := rdtDouble;
  Log10Function.Name := 'log10';
  Log10Function.Prototype := 'Math|log10(Value)';
  SetLength(Log10Function.InputDataTypes, 1);
  Log10Function.InputDataTypes[0] := rdtDouble;
  Log10Function.OptionalArguments := 0;
  Log10Function.CanConvertToConstant := True;
  Log10Function.RFunctionAddr := _log10;
  Add(Log10Function);

  LogNFunction.ResultType := rdtDouble;
  LogNFunction.Name := 'logN';
  LogNFunction.Prototype := 'Math|logN(Base, Value)';
  SetLength(LogNFunction.InputDataTypes, 2);
  LogNFunction.InputDataTypes[0] := rdtDouble;
  LogNFunction.InputDataTypes[1] := rdtDouble;
  LogNFunction.OptionalArguments := 0;
  LogNFunction.CanConvertToConstant := True;
  LogNFunction.RFunctionAddr := _logN;
  Add(LogNFunction);

  LowerCaseFunction.ResultType := rdtString;
  LowerCaseFunction.Name := 'LowerCase';
  LowerCaseFunction.Prototype := 'Text|LowerCase(text_Value)';
  SetLength(LowerCaseFunction.InputDataTypes, 1);
  LowerCaseFunction.InputDataTypes[0] := rdtString;
  LowerCaseFunction.OptionalArguments := 0;
  LowerCaseFunction.CanConvertToConstant := True;
  LowerCaseFunction.SFunctionAddr := _LowerCase;
  Add(LowerCaseFunction);

  MaxIFunction.ResultType := rdtInteger;
  MaxIFunction.Name := 'MaxI';
  MaxIFunction.Prototype := 'Math|MaxI(integer_Value1, integer_Value2, ...)';
  SetLength(MaxIFunction.InputDataTypes, 3);
  MaxIFunction.InputDataTypes[0] := rdtInteger;
  MaxIFunction.InputDataTypes[1] := rdtInteger;
  MaxIFunction.InputDataTypes[2] := rdtInteger;
  MaxIFunction.OptionalArguments := -1;
  MaxIFunction.CanConvertToConstant := True;
  MaxIFunction.IFunctionAddr := _maxI;
  Add(MaxIFunction);

  MaxRFunction.ResultType := rdtDouble;
  MaxRFunction.Name := 'MaxR';
  MaxRFunction.Prototype := 'Math|MaxR(real_Value1, real_Value2, ...)';
  SetLength(MaxRFunction.InputDataTypes, 3);
  MaxRFunction.InputDataTypes[0] := rdtDouble;
  MaxRFunction.InputDataTypes[1] := rdtDouble;
  MaxRFunction.InputDataTypes[2] := rdtDouble;
  MaxRFunction.OptionalArguments := -1;
  MaxRFunction.CanConvertToConstant := True;
  MaxRFunction.RFunctionAddr := _maxR;
  Add(MaxRFunction);

  MinIFunction.ResultType := rdtInteger;
  MinIFunction.Name := 'MinI';
  MinIFunction.Prototype := 'Math|MinI(integer_Value1, integer_Value2, ...)';
  SetLength(MinIFunction.InputDataTypes, 3);
  MinIFunction.InputDataTypes[0] := rdtInteger;
  MinIFunction.InputDataTypes[1] := rdtInteger;
  MinIFunction.InputDataTypes[2] := rdtInteger;
  MinIFunction.OptionalArguments := -1;
  MinIFunction.CanConvertToConstant := True;
  MinIFunction.IFunctionAddr := _minI;
  Add(MinIFunction);

  MinRFunction.ResultType := rdtDouble;
  MinRFunction.Name := 'MinR';
  MinRFunction.Prototype := 'Math|MinR(real_Value1, real_Value2, ...)';
  SetLength(MinRFunction.InputDataTypes, 3);
  MinRFunction.InputDataTypes[0] := rdtDouble;
  MinRFunction.InputDataTypes[1] := rdtDouble;
  MinRFunction.InputDataTypes[2] := rdtDouble;
  MinRFunction.OptionalArguments := -1;
  MinRFunction.CanConvertToConstant := True;
  MinRFunction.RFunctionAddr := _minR;
  Add(MinRFunction);

  OddFunction.ResultType := rdtBoolean;
  OddFunction.Name := 'Odd';
  OddFunction.Prototype := 'Math|Odd(Value)';
  SetLength(OddFunction.InputDataTypes, 1);
  OddFunction.InputDataTypes[0] := rdtInteger;
  OddFunction.CanConvertToConstant := True;
  OddFunction.BFunctionAddr := _Odd;
  Add(OddFunction);

  PiFunction.ResultType := rdtDouble;
  PiFunction.Name := 'Pi';
  PiFunction.Prototype := 'Math|Pi';
  SetLength(PiFunction.InputDataTypes, 0);
  PiFunction.OptionalArguments := 0;
  PiFunction.CanConvertToConstant := True;
  PiFunction.RFunctionAddr := _Pi;
  Add(PiFunction);

  PosFunction.ResultType := rdtInteger;
  PosFunction.Name := 'Pos';
  PosFunction.Prototype := 'Text|Pos(SubText, TextValue)';
  SetLength(PosFunction.InputDataTypes, 2);
  PosFunction.InputDataTypes[0] := rdtString;
  PosFunction.InputDataTypes[1] := rdtString;
  PosFunction.OptionalArguments := 0;
  PosFunction.CanConvertToConstant := True;
  PosFunction.IFunctionAddr := _Pos;
  Add(PosFunction);

  PosExFunction.ResultType := rdtInteger;
  PosExFunction.Name := 'PosEx';
  PosExFunction.Prototype := 'Text|PosEx(SubText, TextValue, Offset)';
  SetLength(PosExFunction.InputDataTypes, 3);
  PosExFunction.InputDataTypes[0] := rdtString;
  PosExFunction.InputDataTypes[1] := rdtString;
  PosExFunction.InputDataTypes[2] := rdtInteger;
  PosExFunction.OptionalArguments := 0;
  PosExFunction.CanConvertToConstant := True;
  PosExFunction.IFunctionAddr := _PosEx;
  Add(PosExFunction);

  PowerFunction.ResultType := rdtDouble;
  PowerFunction.Name := 'Power';
  PowerFunction.Prototype := 'Math|Power(Base, Exponent)';
  SetLength(PowerFunction.InputDataTypes, 2);
  PowerFunction.InputDataTypes[0] := rdtDouble;
  PowerFunction.InputDataTypes[1] := rdtDouble;
  PowerFunction.OptionalArguments := 0;
  PowerFunction.CanConvertToConstant := True;
  PowerFunction.RFunctionAddr := _Power;
  Add(PowerFunction);

  RadToDegFunction.ResultType := rdtDouble;
  RadToDegFunction.Name := 'RadToDeg';
  RadToDegFunction.Prototype := 'Trig|RadToDeg(Value)';
  SetLength(RadToDegFunction.InputDataTypes, 1);
  RadToDegFunction.InputDataTypes[0] := rdtDouble;
  RadToDegFunction.OptionalArguments := 0;
  RadToDegFunction.CanConvertToConstant := True;
  RadToDegFunction.RFunctionAddr := _RadToDeg;
  Add(RadToDegFunction);

  RoundFunction.ResultType := rdtInteger;
  RoundFunction.Name := 'Round';
  RoundFunction.Prototype := 'Math|Round(Value)';
  SetLength(RoundFunction.InputDataTypes, 1);
  RoundFunction.InputDataTypes[0] := rdtDouble;
  RoundFunction.OptionalArguments := 0;
  RoundFunction.CanConvertToConstant := True;
  RoundFunction.IFunctionAddr := _Round;
  Add(RoundFunction);

  SinFunction.ResultType := rdtDouble;
  SinFunction.Name := 'Sin';
  SinFunction.Prototype := 'Trig|Sin(Value)';
  SetLength(SinFunction.InputDataTypes, 1);
  SinFunction.InputDataTypes[0] := rdtDouble;
  SinFunction.OptionalArguments := 0;
  SinFunction.CanConvertToConstant := True;
  SinFunction.RFunctionAddr := _sin;
  Add(SinFunction);

  SinhFunction.ResultType := rdtDouble;
  SinhFunction.Name := 'Sinh';
  SinhFunction.Prototype := 'Trig|Sinh(Value)';
  SetLength(SinhFunction.InputDataTypes, 1);
  SinhFunction.InputDataTypes[0] := rdtDouble;
  SinhFunction.OptionalArguments := 0;
  SinhFunction.CanConvertToConstant := True;
  SinhFunction.RFunctionAddr := _sinh;
  Add(SinhFunction);

  SqrIFunction.ResultType := rdtInteger;
  SqrIFunction.Name := 'SqrI';
  SqrIFunction.Prototype := 'Math|SqrI(integer_Value)';
  SetLength(SqrIFunction.InputDataTypes, 1);
  SqrIFunction.InputDataTypes[0] := rdtInteger;
  SqrIFunction.OptionalArguments := 0;
  SqrIFunction.CanConvertToConstant := True;
  SqrIFunction.IFunctionAddr := _SqrI;
  Add(SqrIFunction);

  SqrRFunction.ResultType := rdtDouble;
  SqrRFunction.Name := 'SqrR';
  SqrRFunction.Prototype := 'Math|SqrR(real_Value)';
  SetLength(SqrRFunction.InputDataTypes, 1);
  SqrRFunction.InputDataTypes[0] := rdtDouble;
  SqrRFunction.OptionalArguments := 0;
  SqrRFunction.CanConvertToConstant := True;
  SqrRFunction.RFunctionAddr := _SqrR;
  Add(SqrRFunction);

  SqrtFunction.ResultType := rdtDouble;
  SqrtFunction.Name := 'Sqrt';
  SqrtFunction.Prototype := 'Math|Sqrt(Value)';
  SetLength(SqrtFunction.InputDataTypes, 1);
  SqrtFunction.InputDataTypes[0] := rdtDouble;
  SqrtFunction.OptionalArguments := 0;
  SqrtFunction.CanConvertToConstant := True;
  SqrtFunction.RFunctionAddr := _sqrt;
  Add(SqrtFunction);

  StrToIntFunction.ResultType := rdtInteger;
  StrToIntFunction.Name := 'TextToInt';
  StrToIntFunction.Prototype := 'Text|TextToInt(Value)';
  SetLength(StrToIntFunction.InputDataTypes, 1);
  StrToIntFunction.InputDataTypes[0] := rdtString;
  StrToIntFunction.OptionalArguments := 0;
  StrToIntFunction.CanConvertToConstant := True;
  StrToIntFunction.IFunctionAddr := _StrToInt;
  SetLength(StrToIntFunction.Synonyms, 1);
  StrToIntFunction.Synonyms[0] := 'StrToInt';
  Add(StrToIntFunction);

  StrToIntDefFunction.ResultType := rdtInteger;
  StrToIntDefFunction.Name := 'TextToIntDef';
  StrToIntDefFunction.Prototype := 'Text|TextToIntDef(Value, DefaultResult)';
  SetLength(StrToIntDefFunction.InputDataTypes, 2);
  StrToIntDefFunction.InputDataTypes[0] := rdtString;
  StrToIntDefFunction.InputDataTypes[1] := rdtInteger;
  StrToIntDefFunction.OptionalArguments := 0;
  StrToIntDefFunction.CanConvertToConstant := True;
  StrToIntDefFunction.IFunctionAddr := _StrToIntDef;
  SetLength(StrToIntDefFunction.Synonyms, 1);
  StrToIntDefFunction.Synonyms[0] := 'StrToIntDef';
  Add(StrToIntDefFunction);

  StrToFloatFunction.ResultType := rdtDouble;
  StrToFloatFunction.Name := 'TextToFloat';
  StrToFloatFunction.Prototype := 'Text|TextToFloat(Value)';
  SetLength(StrToFloatFunction.InputDataTypes, 1);
  StrToFloatFunction.InputDataTypes[0] := rdtString;
  StrToFloatFunction.OptionalArguments := 0;
  StrToFloatFunction.CanConvertToConstant := True;
  StrToFloatFunction.RFunctionAddr := _StrToFloat;
  SetLength(StrToFloatFunction.Synonyms, 1);
  StrToFloatFunction.Synonyms[0] := 'StrToFloat';
  Add(StrToFloatFunction);

  StrToFloatDefFunction.ResultType := rdtDouble;
  StrToFloatDefFunction.Name := 'TextToFloatDef';
  StrToFloatDefFunction.Prototype :=
    'Text|TextToFloatDef(Value, DefaultResult)';
  SetLength(StrToFloatDefFunction.InputDataTypes, 2);
  StrToFloatDefFunction.InputDataTypes[0] := rdtString;
  StrToFloatDefFunction.InputDataTypes[1] := rdtDouble;
  StrToFloatDefFunction.OptionalArguments := 0;
  StrToFloatDefFunction.CanConvertToConstant := True;
  StrToFloatDefFunction.RFunctionAddr := _StrToFloatDef;
  SetLength(StrToFloatDefFunction.Synonyms, 1);
  StrToFloatDefFunction.Synonyms[0] := 'StrToFloatDef';
  Add(StrToFloatDefFunction);

  TanFunction.ResultType := rdtDouble;
  TanFunction.Name := 'Tan';
  TanFunction.Prototype := 'Trig|Tan(Value)';
  SetLength(TanFunction.InputDataTypes, 1);
  TanFunction.InputDataTypes[0] := rdtDouble;
  TanFunction.OptionalArguments := 0;
  TanFunction.CanConvertToConstant := True;
  TanFunction.RFunctionAddr := _tan;
  Add(TanFunction);

  TanhFunction.ResultType := rdtDouble;
  TanhFunction.Name := 'Tanh';
  TanhFunction.Prototype := 'Trig|Tanh(Value)';
  SetLength(TanhFunction.InputDataTypes, 1);
  TanhFunction.InputDataTypes[0] := rdtDouble;
  TanhFunction.OptionalArguments := 0;
  TanhFunction.CanConvertToConstant := True;
  TanhFunction.RFunctionAddr := _tanh;
  Add(TanhFunction);

  TruncFunction.ResultType := rdtInteger;
  TruncFunction.Name := 'Trunc';
  TruncFunction.Prototype := 'Math|Trunc(Value)';
  SetLength(TruncFunction.InputDataTypes, 1);
  TruncFunction.InputDataTypes[0] := rdtDouble;
  TruncFunction.OptionalArguments := 0;
  TruncFunction.CanConvertToConstant := True;
  TruncFunction.IFunctionAddr := _trunc;
  Add(TruncFunction);

  UpperCaseFunction.ResultType := rdtString;
  UpperCaseFunction.Name := 'UpperCase';
  UpperCaseFunction.Prototype := 'Text|UpperCase(text_Value)';
  SetLength(UpperCaseFunction.InputDataTypes, 1);
  UpperCaseFunction.InputDataTypes[0] := rdtString;
  UpperCaseFunction.OptionalArguments := 0;
  UpperCaseFunction.CanConvertToConstant := True;
  UpperCaseFunction.SFunctionAddr := _UpperCase;
  Add(UpperCaseFunction);

  InterpolateFunction.ResultType := rdtDouble;
  InterpolateFunction.Name := 'Interpolate';
  InterpolateFunction.Prototype :=
    'Math|Interpolate(Position, Value1, Distance1, Value2, Distance2)';
  SetLength(InterpolateFunction.InputDataTypes, 5);
  InterpolateFunction.InputDataTypes[0] := rdtDouble;
  InterpolateFunction.InputDataTypes[1] := rdtDouble;
  InterpolateFunction.InputDataTypes[2] := rdtDouble;
  InterpolateFunction.InputDataTypes[3] := rdtDouble;
  InterpolateFunction.InputDataTypes[4] := rdtDouble;
  InterpolateFunction.OptionalArguments := 0;
  InterpolateFunction.CanConvertToConstant := True;
  InterpolateFunction.RFunctionAddr := _Interpolate;
  Add(InterpolateFunction);

  DistanceFunction.ResultType := rdtDouble;
  DistanceFunction.Name := 'Distance';
  DistanceFunction.Prototype := 'Math|Distance(X1, Y1, X2, Y2)';
  SetLength(DistanceFunction.InputDataTypes, 4);
  DistanceFunction.InputDataTypes[0] := rdtDouble;
  DistanceFunction.InputDataTypes[1] := rdtDouble;
  DistanceFunction.InputDataTypes[2] := rdtDouble;
  DistanceFunction.InputDataTypes[3] := rdtDouble;
  DistanceFunction.OptionalArguments := 0;
  DistanceFunction.CanConvertToConstant := True;
  DistanceFunction.RFunctionAddr := _Distance;
  Add(DistanceFunction);
end;

procedure TFunctionStringList.Delete(Index: Integer);
var
  AFunctionClass: TFunctionClass;
begin
  AFunctionClass := Objects[Index] as TFunctionClass;
  AFunctionClass.Free;
  inherited;
end;

destructor TFunctionStringList.Destroy;
begin
  Clear;
  inherited;
end;

function TFunctionStringList.GetFunctionClass(
  const Index: integer): TFunctionClass;
begin
  Result := Objects[Index] as TFunctionClass;
end;

{ TAndExpression }

procedure TAndExpression.Evaluate;
var
  Index: integer;
  ArrayLength: integer;
  AVariable: TConstant;
begin
  if ShouldEvaluate then
  begin
    ArrayLength := Length(Data);
    Assert(ArrayLength = 2);
    Assert(ResultType = rdtBoolean);
    for Index := 0 to ArrayLength - 1 do
    begin
      if Data[Index].Datum = nil then
      begin
        Assert(False);
      end
      else
      begin
        AVariable := Data[Index].Datum;
        if AVariable is TExpression then
        begin
          TExpression(AVariable).Evaluate;
        end;
        PBoolean(FResult)^ := PBoolean(AVariable.FResult)^;
        if not PBoolean(FResult)^ then
          Exit;
      end;
    end;
  end;
end;

{ TOrExpression }

procedure TOrExpression.Evaluate;
var
  Index: integer;
  ArrayLength: integer;
  AVariable: TConstant;
begin
  if ShouldEvaluate then
  begin
    ArrayLength := Length(Data);
    Assert(ArrayLength = 2);
    Assert(ResultType = rdtBoolean);
    for Index := 0 to ArrayLength - 1 do
    begin
      if Data[Index].Datum = nil then
      begin
        Assert(False);
      end
      else
      begin
        AVariable := Data[Index].Datum;
        if AVariable is TExpression then
        begin
          TExpression(AVariable).Evaluate;
        end;
        PBoolean(FResult)^ := PBoolean(AVariable.FResult)^;
        if PBoolean(FResult)^ then
          Exit;
      end;
    end;
  end;
end;

{ TSelectExpression }

procedure TSelectExpression.Evaluate;
var
  ArrayLength: integer;
  AVariable: TConstant;
  SelectIndex: integer;
begin
  if ShouldEvaluate then
  begin
    ArrayLength := Length(Data);
    Assert(ArrayLength >= 3);
    SelectIndex := 0;
    case Data[0].DataType of
      rdtInteger, rdtBoolean:
        begin
          AVariable := Data[0].Datum;
          if AVariable is TExpression then
          begin
            TExpression(AVariable).Evaluate;
          end;
          if Data[0].DataType = rdtInteger then
          begin
            SelectIndex := PInteger(AVariable.FResult)^;
          end
          else
          begin
            SelectIndex := 2 - Ord(pBoolean(AVariable.FResult)^);
          end;
        end;
    else
      Assert(False);
    end;
    if (SelectIndex <= 0) then
    begin
      raise ERbwParserError.Create('Error: Unable to evaluate ' + Name
        + ' function when the first argument evaluates to '
        + IntToStr(SelectIndex)
        + ' because the first argument must always be positive.');
    end;

    if (SelectIndex >= ArrayLength) then
    begin
      raise ERbwParserError.Create('Error: Unable to evaluate ' + Name
        + ' function when the first argument evaluates to '
        + IntToStr(SelectIndex)
        + ' because the number of arguments supplied to the function was '
        + IntToStr(ArrayLength) + ' instead of at least '
        + InttoStr(SelectIndex + 1) + '.');
    end;

    if Data[SelectIndex].Datum = nil then
    begin
      Assert(False);
    end
    else
    begin
      AVariable := Data[SelectIndex].Datum;
      Assert(AVariable.FResultType = FResultType);
      if AVariable is TExpression then
      begin
        TExpression(AVariable).Evaluate;
      end;
      case ResultType of
        rdtDouble:
          begin
            PDouble(FResult)^ := PDouble(AVariable.FResult)^;
          end;
        rdtInteger:
          begin
            PInteger(FResult)^ := PInteger(AVariable.FResult)^;
          end;
        rdtBoolean:
          begin
            PBoolean(FResult)^ := PBoolean(AVariable.FResult)^;
          end;
        rdtString:
          begin
            ResultString := AVariable.ResultString;
          end;
      else
        Assert(False);
      end;
    end;
  end;
end;


{ TCustomVariable }

constructor TCustomVariable.Create(const VariableName: string;
  const DataType: TRbwDataType);
begin
  inherited Create(VariableName, DataType);
  ValidateVariableName(VariableName);
end;

{ TOperator }

function TOperator.Decompile: string;
var
  ArrayLength: integer;
  AVariable: TConstant;
  Paren: string;
begin
  if Assigned(FunctionAddr) then
  begin
    result := Name;
    ArrayLength := Length(Data);
    Assert(ArrayLength in [1, 2]);
    if ArrayLength = 1 then
    begin
      AVariable := Data[0].Datum;
      Assert(AVariable <> nil);
      result := result + ' ' + AVariable.Decompile;
    end
    else
    begin
      AVariable := Data[0].Datum;
      Assert(AVariable <> nil);
      if FTopLevel then
      begin
        Paren := '';
      end
      else
      begin
        Paren := '(';
      end;
      if AVariable is TExpression then
      begin
        TExpression(AVariable).FTopLevel := False;
      end;

      result := Paren + AVariable.Decompile + ' ' + result + ' ';
      AVariable := Data[1].Datum;
      Assert(AVariable <> nil);
      if AVariable is TExpression then
      begin
        TExpression(AVariable).FTopLevel := False;
      end;
      result := result + AVariable.Decompile;
      if not FTopLevel then
      begin
        result := result + ')';
      end;
    end;
  end
  else
  begin
    result := ValueToString;
  end;
end;

{ TSignOperator }

function TSignOperator.Decompile: string;
var
  ArrayLength: integer;
  AVariable: TConstant;
begin
  if Assigned(FunctionAddr) then
  begin
    result := Name;
    ArrayLength := Length(Data);
    Assert(ArrayLength = 1);
    AVariable := Data[0].Datum;
    Assert(AVariable <> nil);
    if AVariable is TExpression then
    begin
      TExpression(AVariable).FTopLevel := False;
    end;
    result := result + AVariable.Decompile;
  end
  else
  begin
    result := ValueToString;
  end;
end;

{ TIntToDoubleExpression }

function TIntToDoubleExpression.Decompile: string;
var
  ArrayLength: integer;
  AVariable: TConstant;
begin
  ArrayLength := Length(Data);
  Assert(ArrayLength = 1);
  AVariable := Data[0].Datum;
  Assert(AVariable <> nil);
  result := AVariable.Decompile;
end;

procedure InitializeVariables;
var
  Index: integer;
  AFunction: TFunctionClass;
begin
  SpecialImplentorList := TSpecialImplentorList.Create;

  IntToDoubleFunction := TFunctionClass.Create;
  IntToDoubleFunction.InputDataCount := 1;
  IntToDoubleFunction.RFunctionAddr := _IntToDouble;
  IntToDoubleFunction.Name := 'IntToDouble';
  IntToDoubleFunction.InputDataTypes[0] := rdtInteger;
  IntToDoubleFunction.OptionalArguments := 0;

  OperatorList := TObjectList.Create;

  NotOperator := TFunctionClass.Create;
  OperatorList.Add(NotOperator);
  NotOperator.InputDataCount := 1;
  NotOperator.BFunctionAddr := _not;
  NotOperator.Name := 'NOT';
  NotOperator.Prototype := 'Logical|not';
  NotOperator.InputDataTypes[0] := rdtBoolean;
  NotOperator.OptionalArguments := 0;

  XorOperator := TFunctionClass.Create;
  OperatorList.Add(XorOperator);
  XorOperator.InputDataCount := 2;
  XorOperator.BFunctionAddr := _Xor;
  XorOperator.Name := 'XOR';
  XorOperator.Prototype := 'Logical|xor';
  XorOperator.InputDataTypes[0] := rdtBoolean;
  XorOperator.InputDataTypes[1] := rdtBoolean;
  XorOperator.OptionalArguments := 0;

  TimesIOperator := TFunctionClass.Create;
  OperatorList.Add(TimesIOperator);
  TimesIOperator.InputDataCount := 2;
  TimesIOperator.IFunctionAddr := _TimesI;
  TimesIOperator.Name := '*';
  TimesIOperator.Prototype := 'Math|*';
  TimesIOperator.InputDataTypes[0] := rdtInteger;
  TimesIOperator.InputDataTypes[1] := rdtInteger;
  TimesIOperator.OptionalArguments := 0;

  TimesROperator := TFunctionClass.Create;
  OperatorList.Add(TimesROperator);
  TimesROperator.InputDataCount := 2;
  TimesROperator.RFunctionAddr := _TimesR;
  TimesROperator.Name := '*';
  TimesROperator.Prototype := 'Math|*';
  TimesROperator.InputDataTypes[0] := rdtDouble;
  TimesROperator.InputDataTypes[1] := rdtDouble;
  TimesROperator.OptionalArguments := 0;

  DivideROperator := TFunctionClass.Create;
  OperatorList.Add(DivideROperator);
  DivideROperator.InputDataCount := 2;
  DivideROperator.RFunctionAddr := _DivideR;
  DivideROperator.Name := '/';
  DivideROperator.Prototype := 'Math|/';
  DivideROperator.InputDataTypes[0] := rdtDouble;
  DivideROperator.InputDataTypes[1] := rdtDouble;
  DivideROperator.OptionalArguments := 0;

  DivideIOperator := TFunctionClass.Create;
  OperatorList.Add(DivideIOperator);
  DivideIOperator.InputDataCount := 2;
  DivideIOperator.RFunctionAddr := _DivideI;
  DivideIOperator.Name := '/';
  DivideIOperator.Prototype := 'Math|/';
  DivideIOperator.InputDataTypes[0] := rdtInteger;
  DivideIOperator.InputDataTypes[1] := rdtInteger;
  DivideIOperator.OptionalArguments := 0;

  DivOperator := TFunctionClass.Create;
  OperatorList.Add(DivOperator);
  DivOperator.InputDataCount := 2;
  DivOperator.IFunctionAddr := _Div;
  DivOperator.Name := 'div';
  DivOperator.Prototype := 'Math|div';
  DivOperator.InputDataTypes[0] := rdtInteger;
  DivOperator.InputDataTypes[1] := rdtInteger;
  DivOperator.OptionalArguments := 0;

  ModOperator := TFunctionClass.Create;
  OperatorList.Add(ModOperator);
  ModOperator.InputDataCount := 2;
  ModOperator.IFunctionAddr := _Mod;
  ModOperator.Name := 'mod';
  ModOperator.Prototype := 'Math|mod';
  ModOperator.InputDataTypes[0] := rdtInteger;
  ModOperator.InputDataTypes[1] := rdtInteger;
  ModOperator.OptionalArguments := 0;

  AndOperator := TFunctionClass.Create;
  OperatorList.Add(AndOperator);
  AndOperator.InputDataCount := 2;
  AndOperator.BFunctionAddr := _And;
  AndOperator.Name := 'and';
  AndOperator.Prototype := 'Logical|and';
  AndOperator.InputDataTypes[0] := rdtBoolean;
  AndOperator.InputDataTypes[1] := rdtBoolean;
  AndOperator.OptionalArguments := 0;

  OrOperator := TFunctionClass.Create;
  OperatorList.Add(OrOperator);
  OrOperator.InputDataCount := 2;
  OrOperator.BFunctionAddr := _Or;
  OrOperator.Name := 'or';
  OrOperator.Prototype := 'Logical|or';
  OrOperator.InputDataTypes[0] := rdtBoolean;
  OrOperator.InputDataTypes[1] := rdtBoolean;
  OrOperator.OptionalArguments := 0;

  PlusSignIOperator := TFunctionClass.Create;
  OperatorList.Add(PlusSignIOperator);
  PlusSignIOperator.InputDataCount := 1;
  PlusSignIOperator.IFunctionAddr := _PlusSignI;
  PlusSignIOperator.Name := '+';
  PlusSignIOperator.Prototype := 'Math|+';
  PlusSignIOperator.InputDataTypes[0] := rdtInteger;
  PlusSignIOperator.OptionalArguments := 0;

  PlusSignROperator := TFunctionClass.Create;
  OperatorList.Add(PlusSignROperator);
  PlusSignROperator.InputDataCount := 1;
  PlusSignROperator.RFunctionAddr := _PlusSignR;
  PlusSignROperator.Name := '+';
  PlusSignROperator.Prototype := 'Math|+';
  PlusSignROperator.InputDataTypes[0] := rdtDouble;
  PlusSignROperator.OptionalArguments := 0;

  MinusSignIOperator := TFunctionClass.Create;
  OperatorList.Add(MinusSignIOperator);
  MinusSignIOperator.InputDataCount := 1;
  MinusSignIOperator.IFunctionAddr := _MinusSignI;
  MinusSignIOperator.Name := '-';
  MinusSignIOperator.Prototype := 'Math|-';
  MinusSignIOperator.InputDataTypes[0] := rdtInteger;
  MinusSignIOperator.OptionalArguments := 0;

  MinusSignROperator := TFunctionClass.Create;
  OperatorList.Add(MinusSignROperator);
  MinusSignROperator.InputDataCount := 1;
  MinusSignROperator.RFunctionAddr := _MinusSignR;
  MinusSignROperator.Name := '-';
  MinusSignROperator.Prototype := 'Math|-';
  MinusSignROperator.InputDataTypes[0] := rdtDouble;
  MinusSignROperator.OptionalArguments := 0;

  PlusIOperator := TFunctionClass.Create;
  OperatorList.Add(PlusIOperator);
  PlusIOperator.InputDataCount := 2;
  PlusIOperator.IFunctionAddr := _PlusI;
  PlusIOperator.Name := '+';
  PlusIOperator.Prototype := 'Math|+';
  PlusIOperator.InputDataTypes[0] := rdtInteger;
  PlusIOperator.InputDataTypes[1] := rdtInteger;
  PlusIOperator.OptionalArguments := 0;

  PlusROperator := TFunctionClass.Create;
  OperatorList.Add(PlusROperator);
  PlusROperator.InputDataCount := 2;
  PlusROperator.RFunctionAddr := _PlusR;
  PlusROperator.Name := '+';
  PlusROperator.Prototype := 'Math|+';
  PlusROperator.InputDataTypes[0] := rdtDouble;
  PlusROperator.InputDataTypes[1] := rdtDouble;
  PlusROperator.OptionalArguments := 0;

  PlusSOperator := TFunctionClass.Create;
  OperatorList.Add(PlusSOperator);
  PlusSOperator.InputDataCount := 2;
  PlusSOperator.SFunctionAddr := _PlusS;
  PlusSOperator.Name := '+';
  PlusSOperator.Prototype := 'Text|+';
  PlusSOperator.InputDataTypes[0] := rdtString;
  PlusSOperator.InputDataTypes[1] := rdtString;
  PlusSOperator.OptionalArguments := 0;

  MinusIOperator := TFunctionClass.Create;
  OperatorList.Add(MinusIOperator);
  MinusIOperator.InputDataCount := 2;
  MinusIOperator.IFunctionAddr := _MinusI;
  MinusIOperator.Name := '-';
  MinusIOperator.Prototype := 'Math|-';
  MinusIOperator.InputDataTypes[0] := rdtInteger;
  MinusIOperator.InputDataTypes[1] := rdtInteger;
  MinusIOperator.OptionalArguments := 0;

  MinusROperator := TFunctionClass.Create;
  OperatorList.Add(MinusROperator);
  MinusROperator.InputDataCount := 2;
  MinusROperator.RFunctionAddr := _MinusR;
  MinusROperator.Name := '-';
  MinusROperator.Prototype := 'Math|-';
  MinusROperator.InputDataTypes[0] := rdtDouble;
  MinusROperator.InputDataTypes[1] := rdtDouble;
  MinusROperator.OptionalArguments := 0;

  EqualIOperator := TFunctionClass.Create;
  OperatorList.Add(EqualIOperator);
  EqualIOperator.InputDataCount := 2;
  EqualIOperator.BFunctionAddr := _EqualI;
  EqualIOperator.Name := '=';
  EqualIOperator.Prototype := 'Logical|=';
  EqualIOperator.InputDataTypes[0] := rdtInteger;
  EqualIOperator.InputDataTypes[1] := rdtInteger;
  EqualIOperator.OptionalArguments := 0;

  EqualROperator := TFunctionClass.Create;
  OperatorList.Add(EqualROperator);
  EqualROperator.InputDataCount := 2;
  EqualROperator.BFunctionAddr := _EqualR;
  EqualROperator.Name := '=';
  EqualROperator.Prototype := 'Logical|=';
  EqualROperator.InputDataTypes[0] := rdtDouble;
  EqualROperator.InputDataTypes[1] := rdtDouble;
  EqualROperator.OptionalArguments := 0;

  EqualSOperator := TFunctionClass.Create;
  OperatorList.Add(EqualSOperator);
  EqualSOperator.InputDataCount := 2;
  EqualSOperator.BFunctionAddr := _EqualS;
  EqualSOperator.Name := '=';
  EqualSOperator.Prototype := 'Logical|=';
  EqualSOperator.InputDataTypes[0] := rdtString;
  EqualSOperator.InputDataTypes[1] := rdtString;
  EqualSOperator.OptionalArguments := 0;

  EqualBOperator := TFunctionClass.Create;
  OperatorList.Add(EqualBOperator);
  EqualBOperator.InputDataCount := 2;
  EqualBOperator.BFunctionAddr := _EqualB;
  EqualBOperator.Name := '=';
  EqualBOperator.Prototype := 'Logical|=';
  EqualBOperator.InputDataTypes[0] := rdtBoolean;
  EqualBOperator.InputDataTypes[1] := rdtBoolean;
  EqualBOperator.OptionalArguments := 0;

  NotEqualIOperator := TFunctionClass.Create;
  OperatorList.Add(NotEqualIOperator);
  NotEqualIOperator.InputDataCount := 2;
  NotEqualIOperator.BFunctionAddr := _NotEqualI;
  NotEqualIOperator.Name := '<>';
  NotEqualIOperator.Prototype := 'Logical|<>';
  NotEqualIOperator.InputDataTypes[0] := rdtInteger;
  NotEqualIOperator.InputDataTypes[1] := rdtInteger;
  NotEqualIOperator.OptionalArguments := 0;

  NotEqualROperator := TFunctionClass.Create;
  OperatorList.Add(NotEqualROperator);
  NotEqualROperator.InputDataCount := 2;
  NotEqualROperator.BFunctionAddr := _NotEqualR;
  NotEqualROperator.Name := '<>';
  NotEqualROperator.Prototype := 'Logical|<>';
  NotEqualROperator.InputDataTypes[0] := rdtDouble;
  NotEqualROperator.InputDataTypes[1] := rdtDouble;
  NotEqualROperator.OptionalArguments := 0;

  NotEqualSOperator := TFunctionClass.Create;
  OperatorList.Add(NotEqualSOperator);
  NotEqualSOperator.InputDataCount := 2;
  NotEqualSOperator.BFunctionAddr := _NotEqualS;
  NotEqualSOperator.Name := '<>';
  NotEqualSOperator.Prototype := 'Logical|<>';
  NotEqualSOperator.InputDataTypes[0] := rdtString;
  NotEqualSOperator.InputDataTypes[1] := rdtString;
  NotEqualSOperator.OptionalArguments := 0;

  NotEqualBOperator := TFunctionClass.Create;
  OperatorList.Add(NotEqualBOperator);
  NotEqualBOperator.InputDataCount := 2;
  NotEqualBOperator.BFunctionAddr := _NotEqualB;
  NotEqualBOperator.Name := '<>';
  NotEqualBOperator.Prototype := 'Logical|<>';
  NotEqualBOperator.InputDataTypes[0] := rdtBoolean;
  NotEqualBOperator.InputDataTypes[1] := rdtBoolean;
  NotEqualBOperator.OptionalArguments := 0;

  LessThanIOperator := TFunctionClass.Create;
  OperatorList.Add(LessThanIOperator);
  LessThanIOperator.InputDataCount := 2;
  LessThanIOperator.BFunctionAddr := _LessThanI;
  LessThanIOperator.Name := '<';
  LessThanIOperator.Prototype := 'Logical|<';
  LessThanIOperator.InputDataTypes[0] := rdtInteger;
  LessThanIOperator.InputDataTypes[1] := rdtInteger;
  LessThanIOperator.OptionalArguments := 0;

  LessThanROperator := TFunctionClass.Create;
  OperatorList.Add(LessThanROperator);
  LessThanROperator.InputDataCount := 2;
  LessThanROperator.BFunctionAddr := _LessThanR;
  LessThanROperator.Name := '<';
  LessThanROperator.Prototype := 'Logical|<';
  LessThanROperator.InputDataTypes[0] := rdtDouble;
  LessThanROperator.InputDataTypes[1] := rdtDouble;
  LessThanROperator.OptionalArguments := 0;

  LessThanSOperator := TFunctionClass.Create;
  OperatorList.Add(LessThanSOperator);
  LessThanSOperator.InputDataCount := 2;
  LessThanSOperator.BFunctionAddr := _LessThanS;
  LessThanSOperator.Name := '<';
  LessThanSOperator.Prototype := 'Logical|<';
  LessThanSOperator.InputDataTypes[0] := rdtString;
  LessThanSOperator.InputDataTypes[1] := rdtString;
  LessThanSOperator.OptionalArguments := 0;

  LessThanBOperator := TFunctionClass.Create;
  OperatorList.Add(LessThanBOperator);
  LessThanBOperator.InputDataCount := 2;
  LessThanBOperator.BFunctionAddr := _LessThanB;
  LessThanBOperator.Name := '<';
  LessThanBOperator.Prototype := 'Logical|<';
  LessThanBOperator.InputDataTypes[0] := rdtBoolean;
  LessThanBOperator.InputDataTypes[1] := rdtBoolean;
  LessThanBOperator.OptionalArguments := 0;

  GreaterThanIOperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanIOperator);
  GreaterThanIOperator.InputDataCount := 2;
  GreaterThanIOperator.BFunctionAddr := _GreaterThanI;
  GreaterThanIOperator.Name := '>';
  GreaterThanIOperator.Prototype := 'Logical|>';
  GreaterThanIOperator.InputDataTypes[0] := rdtInteger;
  GreaterThanIOperator.InputDataTypes[1] := rdtInteger;
  GreaterThanIOperator.OptionalArguments := 0;

  GreaterThanROperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanROperator);
  GreaterThanROperator.InputDataCount := 2;
  GreaterThanROperator.BFunctionAddr := _GreaterThanR;
  GreaterThanROperator.Name := '>';
  GreaterThanROperator.Prototype := 'Logical|>';
  GreaterThanROperator.InputDataTypes[0] := rdtDouble;
  GreaterThanROperator.InputDataTypes[1] := rdtDouble;
  GreaterThanROperator.OptionalArguments := 0;

  GreaterThanSOperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanSOperator);
  GreaterThanSOperator.InputDataCount := 2;
  GreaterThanSOperator.BFunctionAddr := _GreaterThanS;
  GreaterThanSOperator.Name := '>';
  GreaterThanSOperator.Prototype := 'Logical|>';
  GreaterThanSOperator.InputDataTypes[0] := rdtString;
  GreaterThanSOperator.InputDataTypes[1] := rdtString;
  GreaterThanSOperator.OptionalArguments := 0;

  GreaterThanBOperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanBOperator);
  GreaterThanBOperator.InputDataCount := 2;
  GreaterThanBOperator.BFunctionAddr := _GreaterThanB;
  GreaterThanBOperator.Name := '>';
  GreaterThanBOperator.Prototype := 'Logical|>';
  GreaterThanBOperator.InputDataTypes[0] := rdtBoolean;
  GreaterThanBOperator.InputDataTypes[1] := rdtBoolean;
  GreaterThanBOperator.OptionalArguments := 0;

  LessThanOrEqualsIOperator := TFunctionClass.Create;
  OperatorList.Add(LessThanOrEqualsIOperator);
  LessThanOrEqualsIOperator.InputDataCount := 2;
  LessThanOrEqualsIOperator.BFunctionAddr := _LessThanOrEqualsI;
  LessThanOrEqualsIOperator.Name := '<=';
  LessThanOrEqualsIOperator.Prototype := 'Logical|<=';
  LessThanOrEqualsIOperator.InputDataTypes[0] := rdtInteger;
  LessThanOrEqualsIOperator.InputDataTypes[1] := rdtInteger;
  LessThanOrEqualsIOperator.OptionalArguments := 0;

  LessThanOrEqualsROperator := TFunctionClass.Create;
  OperatorList.Add(LessThanOrEqualsROperator);
  LessThanOrEqualsROperator.InputDataCount := 2;
  LessThanOrEqualsROperator.BFunctionAddr := _LessThanOrEqualsR;
  LessThanOrEqualsROperator.Name := '<=';
  LessThanOrEqualsROperator.Prototype := 'Logical|<=';
  LessThanOrEqualsROperator.InputDataTypes[0] := rdtDouble;
  LessThanOrEqualsROperator.InputDataTypes[1] := rdtDouble;
  LessThanOrEqualsROperator.OptionalArguments := 0;

  LessThanOrEqualsSOperator := TFunctionClass.Create;
  OperatorList.Add(LessThanOrEqualsSOperator);
  LessThanOrEqualsSOperator.InputDataCount := 2;
  LessThanOrEqualsSOperator.BFunctionAddr := _LessThanOrEqualsS;
  LessThanOrEqualsSOperator.Name := '<=';
  LessThanOrEqualsSOperator.Prototype := 'Logical|<=';
  LessThanOrEqualsSOperator.InputDataTypes[0] := rdtString;
  LessThanOrEqualsSOperator.InputDataTypes[1] := rdtString;
  LessThanOrEqualsSOperator.OptionalArguments := 0;

  LessThanOrEqualsBOperator := TFunctionClass.Create;
  OperatorList.Add(LessThanOrEqualsBOperator);
  LessThanOrEqualsBOperator.InputDataCount := 2;
  LessThanOrEqualsBOperator.BFunctionAddr := _LessThanOrEqualsB;
  LessThanOrEqualsBOperator.Name := '<=';
  LessThanOrEqualsBOperator.Prototype := 'Logical|<=';
  LessThanOrEqualsBOperator.InputDataTypes[0] := rdtBoolean;
  LessThanOrEqualsBOperator.InputDataTypes[1] := rdtBoolean;
  LessThanOrEqualsBOperator.OptionalArguments := 0;

  GreaterThanOrEqualsIOperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanOrEqualsIOperator);
  GreaterThanOrEqualsIOperator.InputDataCount := 2;
  GreaterThanOrEqualsIOperator.BFunctionAddr := _GreaterThanOrEqualsI;
  GreaterThanOrEqualsIOperator.Name := '>=';
  GreaterThanOrEqualsIOperator.Prototype := 'Logical|>=';
  GreaterThanOrEqualsIOperator.InputDataTypes[0] := rdtInteger;
  GreaterThanOrEqualsIOperator.InputDataTypes[1] := rdtInteger;
  GreaterThanOrEqualsIOperator.OptionalArguments := 0;

  GreaterThanOrEqualsROperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanOrEqualsROperator);
  GreaterThanOrEqualsROperator.InputDataCount := 2;
  GreaterThanOrEqualsROperator.BFunctionAddr := _GreaterThanOrEqualsR;
  GreaterThanOrEqualsROperator.Name := '>=';
  GreaterThanOrEqualsROperator.Prototype := 'Logical|>=';
  GreaterThanOrEqualsROperator.InputDataTypes[0] := rdtDouble;
  GreaterThanOrEqualsROperator.InputDataTypes[1] := rdtDouble;
  GreaterThanOrEqualsROperator.OptionalArguments := 0;

  GreaterThanOrEqualsSOperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanOrEqualsSOperator);
  GreaterThanOrEqualsSOperator.InputDataCount := 2;
  GreaterThanOrEqualsSOperator.BFunctionAddr := _GreaterThanOrEqualsS;
  GreaterThanOrEqualsSOperator.Name := '>=';
  GreaterThanOrEqualsSOperator.Prototype := 'Logical|>=';
  GreaterThanOrEqualsSOperator.InputDataTypes[0] := rdtString;
  GreaterThanOrEqualsSOperator.InputDataTypes[1] := rdtString;
  GreaterThanOrEqualsSOperator.OptionalArguments := 0;

  GreaterThanOrEqualsBOperator := TFunctionClass.Create;
  OperatorList.Add(GreaterThanOrEqualsBOperator);
  GreaterThanOrEqualsBOperator.InputDataCount := 2;
  GreaterThanOrEqualsBOperator.BFunctionAddr := _GreaterThanOrEqualsB;
  GreaterThanOrEqualsBOperator.Name := '>=';
  GreaterThanOrEqualsBOperator.Prototype := 'Logical|>=';
  GreaterThanOrEqualsBOperator.InputDataTypes[0] := rdtBoolean;
  GreaterThanOrEqualsBOperator.InputDataTypes[1] := rdtBoolean;
  GreaterThanOrEqualsBOperator.OptionalArguments := 0;

  OverloadedFunctionList := TObjectList.Create;

  AbsIOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(AbsIOverloadedFunction);
  AbsIOverloadedFunction.InputDataCount := 1;
  AbsIOverloadedFunction.IFunctionAddr := _AbsI;
  AbsIOverloadedFunction.Name := 'Abs';
  AbsIOverloadedFunction.Prototype := 'Math|Abs(Value)';
  AbsIOverloadedFunction.InputDataTypes[0] := rdtInteger;
  AbsIOverloadedFunction.OptionalArguments := 0;

  AbsROverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(AbsROverloadedFunction);
  AbsROverloadedFunction.InputDataCount := 1;
  AbsROverloadedFunction.RFunctionAddr := _AbsR;
  AbsROverloadedFunction.Name := 'Abs';
  AbsROverloadedFunction.Prototype := 'Math|Abs(Value)';
  AbsROverloadedFunction.InputDataTypes[0] := rdtDouble;
  AbsROverloadedFunction.OptionalArguments := 0;

  MaxIOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(MaxIOverloadedFunction);
  MaxIOverloadedFunction.InputDataCount := 3;
  MaxIOverloadedFunction.OptionalArguments := -1;
  MaxIOverloadedFunction.IFunctionAddr := _MaxI;
  MaxIOverloadedFunction.Name := 'Max';
  MaxIOverloadedFunction.Prototype := 'Math|Max(Value1, Value2, ...)';
  MaxIOverloadedFunction.InputDataTypes[0] := rdtInteger;
  MaxIOverloadedFunction.InputDataTypes[1] := rdtInteger;
  MaxIOverloadedFunction.InputDataTypes[2] := rdtInteger;

  MaxROverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(MaxROverloadedFunction);
  MaxROverloadedFunction.InputDataCount := 3;
  MaxROverloadedFunction.OptionalArguments := -1;
  MaxROverloadedFunction.RFunctionAddr := _MaxR;
  MaxROverloadedFunction.Name := 'Max';
  MaxROverloadedFunction.Prototype := 'Math|Max(Value1, Value2, ...)';
  MaxROverloadedFunction.InputDataTypes[0] := rdtDouble;
  MaxROverloadedFunction.InputDataTypes[1] := rdtDouble;
  MaxROverloadedFunction.InputDataTypes[2] := rdtDouble;

  MinIOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(MinIOverloadedFunction);
  MinIOverloadedFunction.InputDataCount := 3;
  MinIOverloadedFunction.OptionalArguments := -1;
  MinIOverloadedFunction.IFunctionAddr := _MinI;
  MinIOverloadedFunction.Name := 'Min';
  MinIOverloadedFunction.Prototype := 'Math|Min(Value1, Value2, ...)';
  MinIOverloadedFunction.InputDataTypes[0] := rdtInteger;
  MinIOverloadedFunction.InputDataTypes[1] := rdtInteger;
  MinIOverloadedFunction.InputDataTypes[2] := rdtInteger;

  MinROverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(MinROverloadedFunction);
  MinROverloadedFunction.InputDataCount := 3;
  MinROverloadedFunction.OptionalArguments := -1;
  MinROverloadedFunction.RFunctionAddr := _MinR;
  MinROverloadedFunction.Name := 'Min';
  MinROverloadedFunction.Prototype := 'Math|Min(Value1, Value2, ...)';
  MinROverloadedFunction.InputDataTypes[0] := rdtDouble;
  MinROverloadedFunction.InputDataTypes[1] := rdtDouble;
  MinROverloadedFunction.InputDataTypes[2] := rdtDouble;

  SqrIOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(SqrIOverloadedFunction);
  SqrIOverloadedFunction.InputDataCount := 1;
  SqrIOverloadedFunction.OptionalArguments := 0;
  SqrIOverloadedFunction.IFunctionAddr := _SqrI;
  SqrIOverloadedFunction.Name := 'Sqr';
  SqrIOverloadedFunction.Prototype := 'Math|Sqr(Value)';
  SqrIOverloadedFunction.InputDataTypes[0] := rdtInteger;

  SqrROverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(SqrROverloadedFunction);
  SqrROverloadedFunction.InputDataCount := 1;
  SqrROverloadedFunction.OptionalArguments := 0;
  SqrROverloadedFunction.RFunctionAddr := _SqrR;
  SqrROverloadedFunction.Name := 'Sqr';
  SqrROverloadedFunction.Prototype := 'Math|Sqr(Value)';
  SqrROverloadedFunction.InputDataTypes[0] := rdtDouble;

  CaseBOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(CaseBOverloadedFunction);
  CaseBOverloadedFunction.InputDataCount := 4;
  CaseBOverloadedFunction.OptionalArguments := -1;
  CaseBOverloadedFunction.BFunctionAddr := _CaseBoolean;
  CaseBOverloadedFunction.Name := 'Case';
  CaseBOverloadedFunction.Prototype :=
    'Logical|Case(Index, Result1, Result2, ...)';
  CaseBOverloadedFunction.InputDataTypes[0] := rdtInteger;
  CaseBOverloadedFunction.InputDataTypes[1] := rdtBoolean;
  CaseBOverloadedFunction.InputDataTypes[2] := rdtBoolean;
  CaseBOverloadedFunction.InputDataTypes[3] := rdtBoolean;

  CaseIOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(CaseIOverloadedFunction);
  CaseIOverloadedFunction.InputDataCount := 4;
  CaseIOverloadedFunction.OptionalArguments := -1;
  CaseIOverloadedFunction.IFunctionAddr := _CaseInteger;
  CaseIOverloadedFunction.Name := 'Case';
  CaseIOverloadedFunction.Prototype :=
    'Logical|Case(Index, Result1, Result2, ...)';
  CaseIOverloadedFunction.InputDataTypes[0] := rdtInteger;
  CaseIOverloadedFunction.InputDataTypes[1] := rdtInteger;
  CaseIOverloadedFunction.InputDataTypes[2] := rdtInteger;
  CaseIOverloadedFunction.InputDataTypes[3] := rdtInteger;

  CaseROverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(CaseROverloadedFunction);
  CaseROverloadedFunction.InputDataCount := 4;
  CaseROverloadedFunction.OptionalArguments := -1;
  CaseROverloadedFunction.RFunctionAddr := _CaseDouble;
  CaseROverloadedFunction.Name := 'Case';
  CaseROverloadedFunction.Prototype :=
    'Logical|Case(Index, Result1, Result2, ...)';
  CaseROverloadedFunction.InputDataTypes[0] := rdtInteger;
  CaseROverloadedFunction.InputDataTypes[1] := rdtDouble;
  CaseROverloadedFunction.InputDataTypes[2] := rdtDouble;
  CaseROverloadedFunction.InputDataTypes[3] := rdtDouble;

  CaseSOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(CaseSOverloadedFunction);
  CaseSOverloadedFunction.InputDataCount := 4;
  CaseSOverloadedFunction.OptionalArguments := -1;
  CaseSOverloadedFunction.SFunctionAddr := _CaseString;
  CaseSOverloadedFunction.Name := 'Case';
  CaseSOverloadedFunction.Prototype :=
    'Logical|Case(Index, Result1, Result2, ...)';
  CaseSOverloadedFunction.InputDataTypes[0] := rdtInteger;
  CaseSOverloadedFunction.InputDataTypes[1] := rdtString;
  CaseSOverloadedFunction.InputDataTypes[2] := rdtString;
  CaseSOverloadedFunction.InputDataTypes[3] := rdtString;

  IfBOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(IfBOverloadedFunction);
  IfBOverloadedFunction.InputDataCount := 3;
  IfBOverloadedFunction.OptionalArguments := 0;
  IfBOverloadedFunction.BFunctionAddr := _IfBoolean;
  IfBOverloadedFunction.Name := 'If';
  IfBOverloadedFunction.Prototype :=
    'Logical|If(Boolean_Value, If_True_Result, If_False_Result)';
  IfBOverloadedFunction.InputDataTypes[0] := rdtBoolean;
  IfBOverloadedFunction.InputDataTypes[1] := rdtBoolean;
  IfBOverloadedFunction.InputDataTypes[2] := rdtBoolean;

  IfIOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(IfIOverloadedFunction);
  IfIOverloadedFunction.InputDataCount := 3;
  IfIOverloadedFunction.OptionalArguments := 0;
  IfIOverloadedFunction.IFunctionAddr := _IfInteger;
  IfIOverloadedFunction.Name := 'If';
  IfIOverloadedFunction.Prototype :=
    'Logical|If(Boolean_Value, If_True_Result, If_False_Result)';
  IfIOverloadedFunction.InputDataTypes[0] := rdtBoolean;
  IfIOverloadedFunction.InputDataTypes[1] := rdtInteger;
  IfIOverloadedFunction.InputDataTypes[2] := rdtInteger;

  IfROverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(IfROverloadedFunction);
  IfROverloadedFunction.InputDataCount := 3;
  IfROverloadedFunction.OptionalArguments := 0;
  IfROverloadedFunction.RFunctionAddr := _IfDouble;
  IfROverloadedFunction.Name := 'If';
  IfROverloadedFunction.Prototype :=
    'Logical|If(Boolean_Value, If_True_Result, If_False_Result)';
  IfROverloadedFunction.InputDataTypes[0] := rdtBoolean;
  IfROverloadedFunction.InputDataTypes[1] := rdtDouble;
  IfROverloadedFunction.InputDataTypes[2] := rdtDouble;

  IfSOverloadedFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(IfSOverloadedFunction);
  IfSOverloadedFunction.InputDataCount := 3;
  IfSOverloadedFunction.OptionalArguments := 0;
  IfSOverloadedFunction.SFunctionAddr := _IfString;
  IfSOverloadedFunction.Name := 'If';
  IfSOverloadedFunction.Prototype :=
    'Logical|If(Boolean_Value, If_True_Result, If_False_Result)';
  IfSOverloadedFunction.InputDataTypes[0] := rdtBoolean;
  IfSOverloadedFunction.InputDataTypes[1] := rdtString;
  IfSOverloadedFunction.InputDataTypes[2] := rdtString;

  MultiInterpolateFunction := TFunctionClass.Create;
  OverloadedFunctionList.Add(MultiInterpolateFunction);
  MultiInterpolateFunction.InputDataCount := 4;
  MultiInterpolateFunction.OptionalArguments := -1;
  MultiInterpolateFunction.RFunctionAddr := _MultiInterpolate;
  MultiInterpolateFunction.Name := 'MultiInterpolate';
  MultiInterpolateFunction.Prototype :=
    'Math|MultiInterpolate(Position, Value1, Distance1, [Value2, Distance2,] ...)';
  MultiInterpolateFunction.InputDataTypes[0] := rdtDouble;
  MultiInterpolateFunction.InputDataTypes[1] := rdtDouble;
  MultiInterpolateFunction.InputDataTypes[2] := rdtDouble;
  MultiInterpolateFunction.InputDataTypes[3] := rdtDouble;
end;

{ TVariableExpression }

constructor TVariableExpression.Create(const Variable: TCustomValue);
var
  VariableConverter: TFunctionRecord;
begin
  VariableConverter.ResultType := Variable.ResultType;
  VariableConverter.OptionalArguments := 0;
  VariableConverter.CanConvertToConstant := False;
  VariableConverter.Prototype := 'Dummy';
  VariableConverter.RFunctionAddr := nil;
  VariableConverter.Name := Variable.FUserName;
  SetLength(VariableConverter.InputDataTypes, 1);
  VariableConverter.InputDataTypes[0] := Variable.ResultType;
  inherited Create(VariableConverter);
  Variables[0] := Variable;
end;

function TVariableExpression.Decompile: string;
var
  AVariable: TConstant;
begin
  Assert(Length(Data) = 1);
  AVariable := Data[0].Datum;
  Assert(AVariable <> nil);
  result := AVariable.Decompile;
end;

procedure TVariableExpression.Evaluate;
var
  resultVariable: TConstant;
begin
  resultVariable := Variables[0];
  assert(resultVariable.ResultType = ResultType);
  case ResultType of
    rdtDouble:
      begin
        PDouble(FResult)^ := PDouble(resultVariable.FResult)^;
      end;
    rdtInteger:
      begin
        PInteger(FResult)^ := PInteger(resultVariable.FResult)^;
      end;
    rdtBoolean:
      begin
        PBoolean(FResult)^ := PBoolean(resultVariable.FResult)^;
      end;
    rdtString:
      begin
        ResultString := resultVariable.ResultString;
      end;
  else
    Assert(False);
  end;

end;

{ TMultiInterpolateExpression }

procedure TMultiInterpolateExpression.Evaluate;
var
  ArrayLength: integer;
  AVariable: TConstant;
  PositionVariable: TConstant;
  Position: double;
  PriorDistance: double;
  DistanceIndex: integer;
  Distance: double;
  PriorValue: double;
  Value: double;
begin
  if ShouldEvaluate then
  begin
    ArrayLength := Length(Data);
    Assert(ArrayLength >= 3);
    Assert((ArrayLength mod 2) = 1);

    if ArrayLength = 3 then
    begin
      AVariable := Data[1].Datum;
      if AVariable is TExpression then
      begin
        TExpression(AVariable).Evaluate;
      end;
      PDouble(FResult)^ := AVariable.DoubleResult;
      Exit;
    end;

    Assert(Data[0].DataType in [rdtDouble, rdtInteger]);
    PositionVariable := Data[0].Datum;
    if PositionVariable is TExpression then
    begin
      TExpression(PositionVariable).Evaluate;
    end;
    Position := PositionVariable.DoubleResult;

    PriorDistance := 0;
    for DistanceIndex := 0 to ArrayLength div 2 - 1 do
    begin
      Assert(Data[DistanceIndex * 2 + 2].DataType in [rdtDouble, rdtInteger]);
      AVariable := Data[DistanceIndex * 2 + 2].Datum;
      if AVariable is TExpression then
      begin
        TExpression(AVariable).Evaluate;
      end;
      Distance := AVariable.DoubleResult;
      if Position < Distance then
      begin
        Assert(Data[DistanceIndex * 2 + 1].DataType in [rdtDouble, rdtInteger]);
        AVariable := Data[DistanceIndex * 2 + 1].Datum;
        if AVariable is TExpression then
        begin
          TExpression(AVariable).Evaluate;
        end;
        Value := AVariable.DoubleResult;

        if (DistanceIndex = 0) or (Distance = PriorDistance) then
        begin
          PDouble(FResult)^ := Value;
        end
        else
        begin
          Assert(Data[DistanceIndex * 2 - 1].DataType in [rdtDouble,
            rdtInteger]);
          AVariable := Data[DistanceIndex * 2 - 1].Datum;
          if AVariable is TExpression then
          begin
            TExpression(AVariable).Evaluate;
          end;
          PriorValue := AVariable.DoubleResult;

          PDouble(FResult)^ := (Position - PriorDistance)
            / (Distance - PriorDistance) * (Value - PriorValue) + PriorValue;
        end;
        Exit;
      end
      else if Distance = Position then
      begin
        Assert(Data[DistanceIndex * 2 + 1].DataType in [rdtDouble, rdtInteger]);
        AVariable := Data[DistanceIndex * 2 + 1].Datum;
        if AVariable is TExpression then
        begin
          TExpression(AVariable).Evaluate;
        end;
        PDouble(FResult)^ := AVariable.DoubleResult;
        Exit;
      end
      else
      begin
        if DistanceIndex = (ArrayLength div 2) then
        begin
          Assert(Data[DistanceIndex * 2 + 1].DataType in [rdtDouble,
            rdtInteger]);
          AVariable := Data[DistanceIndex * 2 + 1].Datum;
          if AVariable is TExpression then
          begin
            TExpression(AVariable).Evaluate;
          end;
          PDouble(FResult)^ := AVariable.DoubleResult;
          Exit;
        end;
        PriorDistance := Distance;
      end;
    end;
  end;
end;

{ TSpecialImplentorList }

function TSpecialImplentorList.Add(
  const Item: TSpecialImplementor): Integer;
begin
  result := FList.Add(Item);
end;

procedure TSpecialImplentorList.Clear;
begin
  FList.Clear
end;

constructor TSpecialImplentorList.Create;
begin
  inherited;
  FList := TList.Create;
end;

procedure TSpecialImplentorList.Delete(const Index: Integer);
begin
  FList.Delete(Index);
end;

destructor TSpecialImplentorList.Destroy;
begin
  FList.Free;
  inherited;
end;

function TSpecialImplentorList.GetCapacity: integer;
begin
  result := FList.Capacity;
end;

function TSpecialImplentorList.GetCount: integer;
begin
  result := FList.Count;
end;

function TSpecialImplentorList.GetItems(
  const Index: integer): TSpecialImplementor;
begin
  result := FList[Index];
end;

function TSpecialImplentorList.Remove(
  const Item: TSpecialImplementor): Integer;
begin
  result := FList.Remove(Item)
end;

procedure TSpecialImplentorList.SetCapacity(const Value: integer);
begin
  FList.Capacity := Value;
end;

procedure TSpecialImplentorList.SetItems(const Index: integer;
  const Value: TSpecialImplementor);
begin
  FList[Index] := Value;
end;

{ ERbwParserError }

constructor ERbwParserError.CreateMode(const Msg: string;
  const ErrorMode: integer);
begin
  inherited Create(Msg);
  ErrorType := ErrorMode
end;

initialization
  begin
    InitializeVariables
  end;

finalization
  OperatorList.Free;
  OverloadedFunctionList.Free;
  IntToDoubleFunction.Free;
  SpecialImplentorList.Free;
end.

